/*
   3DNOW and 3DNOWEX optimized IMDCT
   Licence: GPL v2
   Copyrights: Nick Kurshev
*/

#undef PSWAP_MM
#ifdef HAVE_3DNOWEX
#define PSWAP_MM(mm_base,mm_hlp) "pswapd	"mm_base","mm_base"\n\t"
#else
#define PSWAP_MM(mm_base,mm_hlp)\
	"movq	"mm_base","mm_hlp"\n\t"\
	"psrlq $32, "mm_base"\n\t"\
	"punpckldq "mm_hlp","mm_base"\n\t"
#endif

static void
#ifdef HAVE_3DNOWEX
imdct_do_512_3dnowex
#else
imdct_do_512_3dnow
#endif
(sample_t data[],sample_t delay[], sample_t bias)
{
    int m;
    int two_m;
    int two_m_plus_one;

    sample_t tmp[4];

    sample_t *data_ptr;
    sample_t *delay_ptr;
    sample_t *window_ptr;
	
    /* 512 IMDCT with source and dest data in 'data' */
    /* see the c version (dct_do_512()), its allmost identical, just in C */ 

    /* Pre IFFT complex multiply plus IFFT cmplx conjugate */
    /* Bit reversed shuffling */
	asm volatile(
		"xorl	%%esi, %%esi\n\t"
		"leal	"MANGLE(bit_reverse_512)", %%eax\n\t"
		"movl	$1008, %%edi\n\t"
		"pushl	%%ebp\n\t" /* use ebp without telling gcc */
		".balign 16\n\t"
		"1:\n\t"
		"movd	12(%0, %%edi), %%mm0\n\t"	// XR
		"punpckldq (%0, %%esi), %%mm0\n\t"	// IR
		"movd	4(%0, %%edi), %%mm1\n\t"	// Xr
		"punpckldq 8(%0, %%esi), %%mm1\n\t"	// ir
		"movq	"MANGLE(sseSinCos1c)"(%%esi), %%mm4\n\t"
		"movq	8+"MANGLE(sseSinCos1c)"(%%esi), %%mm5\n\t"
		"pfmul	%%mm0, %%mm4\n\t"
		"pfmul	%%mm1, %%mm5\n\t"
		PSWAP_MM("%%mm0", "%%mm6")
		PSWAP_MM("%%mm1", "%%mm7")
		"pfmul	"MANGLE(sseSinCos1d)"(%%esi), %%mm0\n\t"
		"pfmul	8+"MANGLE(sseSinCos1d)"(%%esi), %%mm1\n\t"
		"pfsub	%%mm0, %%mm4\n\t"
		"pfsub	%%mm1, %%mm5\n\t"
		"movzbl	(%%eax), %%edx\n\t"
		"movzbl	1(%%eax), %%ebp\n\t"
		"movq	%%mm4, (%1, %%edx,8)\n\t"
		"movq	%%mm5, (%1, %%ebp,8)\n\t"
		"addl	$16, %%esi\n\t"
		"addl	$2, %%eax\n\t"
		"subl	$16, %%edi\n\t"
		" jnc	1b\n\t"
		"popl	%%ebp\n\t"/*no we didnt touch ebp *g* */
		:: "b" (data), "c" (buf)
		: "%esi", "%edi", "%eax", "%edx"
	);

    /* 1. iteration */
    /* Note w[0][0]={1,0} */
	asm volatile(
		"pxor	%%mm3, %%mm3\n\t"
		"pxor	%%mm4, %%mm4\n\t"
		"movl	%0, %%esi\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"movq	(%%esi), %%mm0\n\t"
		"movq	8(%%esi), %%mm2\n\t"
		"movq	(%%esi), %%mm1\n\t"
		"movq	8(%%esi), %%mm5\n\t"
		"pfadd	%%mm2, %%mm0\n\t"
		"pfsub	%%mm4, %%mm0\n\t"
		"pfsub	%%mm5, %%mm1\n\t"
		"movq	%%mm0, (%%esi)\n\t"
		"movq	%%mm1, 8(%%esi)\n\t"
		"addl	$16, %%esi\n\t"
		"cmpl	%1, %%esi\n\t"
		" jb	1b\n\t"
		:: "g" (buf), "r" (buf + 128)
		: "%esi"
	);
    /* 2. iteration */
    /* Note w[1]={{1,0}, {0,-1}} */
	asm volatile(
		"movq	"MANGLE(ps111_1)", %%mm6\n\t"
		"movq	8+"MANGLE(ps111_1)", %%mm7\n\t"
		"movl	%0, %%esi\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"movq	16(%%esi), %%mm4\n\t"
		"movq	24(%%esi), %%mm5\n\t"
		PSWAP_MM("%%mm5", "%%mm0")
		"pfmul	%%mm6, %%mm4\n\t"
		"pfmul	%%mm7, %%mm5\n\t"
		"movq	(%%esi), %%mm0\n\t"
		"movq	8(%%esi), %%mm1\n\t"
		"movq	%%mm0, %%mm2\n\t"
		"movq	%%mm1, %%mm3\n\t"
		"pfadd	%%mm4, %%mm0\n\t"
		"pfadd	%%mm5, %%mm1\n\t"
		"pfsub	%%mm4, %%mm2\n\t"
		"pfsub	%%mm5, %%mm3\n\t"
		"movq	%%mm0, (%%esi)\n\t"
		"movq	%%mm1, 8(%%esi)\n\t"
		"movq	%%mm2, 16(%%esi)\n\t"
		"movq	%%mm3, 24(%%esi)\n\t"
		"addl	$32, %%esi\n\t"
		"cmpl	%1, %%esi\n\t"
		" jb	1b\n\t"
		:: "g" (buf), "r" (buf + 128)
		: "%esi"
	);
    /* 3. iteration */
/*
 Note sseW2+0={1,1,sqrt(2),sqrt(2))
 Note sseW2+16={0,0,sqrt(2),-sqrt(2))
 Note sseW2+32={0,0,-sqrt(2),-sqrt(2))
 Note sseW2+48={1,-1,sqrt(2),-sqrt(2))
*/
/* 
TODO :	make it easy for 3dnow! 
	for 3dnowex! can be optimized even more
*/
#ifdef HAVE_3DNOWEX
	asm volatile(
		"movl	%0, %%esi\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"movq	32(%%esi), %%mm4\n\t"
		"movq	40(%%esi), %%mm5\n\t"
		"movq	48(%%esi), %%mm6\n\t"
		"movq	56(%%esi), %%mm7\n\t"
		"movq	"MANGLE(sseW2)", %%mm0\n\t" /*xmm4*/
		"movq	8+"MANGLE(sseW2)", %%mm1\n\t"
		"movq	32+"MANGLE(sseW2)", %%mm2\n\t"/*xmm5*/
		"movq	40+"MANGLE(sseW2)", %%mm3\n\t"
		"pfmul	%%mm4, %%mm0\n\t"
		"pfmul	%%mm5, %%mm1\n\t"
		"pfmul	%%mm6, %%mm2\n\t"
		"pfmul	%%mm7, %%mm3\n\t"
		"pswapd	%%mm4, %%mm4\n\t"
		"pswapd	%%mm5, %%mm5\n\t"
		"pswapd	%%mm6, %%mm6\n\t"
		"pswapd	%%mm7, %%mm7\n\t"
		"pfmul	16+"MANGLE(sseW2)", %%mm4\n\t"
		"pfmul	24+"MANGLE(sseW2)", %%mm5\n\t"
		"pfmul	48+"MANGLE(sseW2)", %%mm6\n\t"
		"pfmul	56+"MANGLE(sseW2)", %%mm7\n\t"
		"pfadd	%%mm0, %%mm4\n\t"
		"pfadd	%%mm1, %%mm5\n\t"
		"pfadd	%%mm2, %%mm6\n\t"
		"pfadd	%%mm3, %%mm7\n\t"
		"movq	%%mm4, %2\n\t"
		"movq	%%mm5, 8%2\n\t"
		"movq	%%mm6, 16%2\n\t"
		"movq	%%mm7, 24%2\n\t"
		"movq	(%%esi), %%mm0\n\t"
		"movq	8(%%esi), %%mm1\n\t"
		"movq	16(%%esi), %%mm2\n\t"
		"movq	24(%%esi), %%mm3\n\t"
		"pfadd	%%mm0, %%mm4\n\t"
		"pfadd	%%mm1, %%mm5\n\t"
		"pfadd	%%mm2, %%mm6\n\t"
		"pfadd	%%mm3, %%mm7\n\t"
		"pfsub	%2, %%mm0\n\t"
		"pfsub	8%2, %%mm1\n\t"
		"pfsub	16%2, %%mm2\n\t"
		"pfsub	24%2, %%mm3\n\t"
		"movq	%%mm4, (%%esi)\n\t" 
		"movq	%%mm5, 8(%%esi)\n\t" 
		"movq	%%mm6, 16(%%esi)\n\t" 
		"movq	%%mm7, 24(%%esi)\n\t" 
		"movq	%%mm0, 32(%%esi)\n\t" 
		"movq	%%mm1, 40(%%esi)\n\t" 
		"movq	%%mm2, 48(%%esi)\n\t" 
		"movq	%%mm3, 56(%%esi)\n\t" 
		"addl	$64, %%esi\n\t"
		"cmpl	%1, %%esi\n\t"
		" jb 1b\n\t"
		:: "g" (buf), "r" (buf + 128), "m"(tmp[0])
		: "%esi"
	);
#else
	asm volatile(
		"movl	%0, %%esi\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"movq	32(%%esi), %%mm4\n\t"
		"movq	40(%%esi), %%mm5\n\t"
		"movq	48(%%esi), %%mm6\n\t"
		"movq	56(%%esi), %%mm7\n\t"
		"movq	"MANGLE(sseW2)", %%mm0\n\t" /*xmm4*/
		"movq	8+"MANGLE(sseW2)", %%mm1\n\t"
		"movq	32+"MANGLE(sseW2)", %%mm2\n\t"/*xmm5*/
		"movq	40+"MANGLE(sseW2)", %%mm3\n\t"
		"pfmul	%%mm4, %%mm0\n\t"
		"pfmul	%%mm5, %%mm1\n\t"
		"pfmul	%%mm6, %%mm2\n\t"
		"pfmul	%%mm7, %%mm3\n\t"
		"movq	%%mm0, %2\n\t"
		"movq	%%mm1, 8%2\n\t"
		"movq	%%mm2, 16%2\n\t"
		"movq	%%mm3, 24%2\n\t"
		"movq	%%mm4, %%mm0\n\t"
		"movq	%%mm5, %%mm1\n\t"
		"movq	%%mm6, %%mm2\n\t"
		"movq	%%mm7, %%mm3\n\t"
		"psrlq	$32, %%mm4\n\t"
		"psrlq	$32, %%mm5\n\t"
		"psrlq	$32, %%mm6\n\t"
		"psrlq	$32, %%mm7\n\t"
		"punpckldq %%mm0, %%mm4\n\t"
		"punpckldq %%mm1, %%mm5\n\t"
		"punpckldq %%mm2, %%mm6\n\t"
		"punpckldq %%mm3, %%mm7\n\t"
		"pfmul	48+"MANGLE(sseW2)", %%mm6\n\t"
		"pfmul	56+"MANGLE(sseW2)", %%mm7\n\t"
		"pfmul	16+"MANGLE(sseW2)", %%mm4\n\t"
		"pfmul	24+"MANGLE(sseW2)", %%mm5\n\t"
		"movq	(%%esi), %%mm0\n\t"
		"movq	8(%%esi), %%mm1\n\t"
		"movq	16(%%esi), %%mm2\n\t"
		"movq	24(%%esi), %%mm3\n\t"
		"pfadd	%2, %%mm4\n\t"
		"pfadd	8%2, %%mm5\n\t"
		"pfadd	16%2, %%mm6\n\t"
		"pfadd	24%2, %%mm7\n\t"
		"movq	%%mm4, %2\n\t"
		"movq	%%mm5, 8%2\n\t"
		"movq	%%mm6, 16%2\n\t"
		"movq	%%mm7, 24%2\n\t"
		"pfadd	%%mm0, %%mm4\n\t"
		"pfadd	%%mm1, %%mm5\n\t"
		"pfadd	%%mm2, %%mm6\n\t"
		"pfadd	%%mm3, %%mm7\n\t"
		"pfsub	%2, %%mm0\n\t"
		"pfsub	8%2, %%mm1\n\t"
		"pfsub	16%2, %%mm2\n\t"
		"pfsub	24%2, %%mm3\n\t"
		"movq	%%mm4, (%%esi)\n\t" 
		"movq	%%mm5, 8(%%esi)\n\t" 
		"movq	%%mm6, 16(%%esi)\n\t" 
		"movq	%%mm7, 24(%%esi)\n\t" 
		"movq	%%mm0, 32(%%esi)\n\t" 
		"movq	%%mm1, 40(%%esi)\n\t" 
		"movq	%%mm2, 48(%%esi)\n\t" 
		"movq	%%mm3, 56(%%esi)\n\t" 
		"addl	$64, %%esi\n\t"
		"cmpl	%1, %%esi\n\t"
		" jb 1b\n\t"
		:: "g" (buf), "r" (buf + 128), "m"(tmp[0])
		: "%esi"
	);
#endif
    /* 4-7. iterations */
    for (m=3; m < 7; m++) {
	two_m = (1 << m);
	two_m_plus_one = two_m<<1;
	asm volatile(
		"movl	%0, %%esi\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"xorl	%%edi, %%edi\n\t"
		"leal	(%%esi, %3), %%edx\n\t"
		"2:\n\t"
		"movq	(%%edx, %%edi), %%mm2\n\t"
		"movq	8(%%edx, %%edi), %%mm3\n\t"
		"movq	(%4, %%edi, 2), %%mm4\n\t"
		"movq	8(%4, %%edi, 2), %%mm5\n\t"
		"pfmul	%%mm2, %%mm4\n\t"
		"pfmul	%%mm3, %%mm5\n\t"
		PSWAP_MM("%%mm2", "%%mm0")
		PSWAP_MM("%%mm3", "%%mm1")
		"pfmul	16(%4, %%edi, 2), %%mm2\n\t"
		"pfmul	24(%4, %%edi, 2), %%mm3\n\t"
		"movq	(%%esi, %%edi), %%mm0\n\t"
		"movq	8(%%esi, %%edi), %%mm1\n\t"
		"pfadd	%%mm4, %%mm2\n\t"
		"pfadd	%%mm5, %%mm3\n\t"
		"movq	%%mm2, %%mm4\n\t"
		"movq	%%mm3, %%mm5\n\t"
		"pfadd	%%mm0, %%mm2\n\t"
		"pfadd	%%mm1, %%mm3\n\t"
		"pfsub	%%mm4, %%mm0\n\t"
		"pfsub	%%mm5, %%mm1\n\t"
		"movq	%%mm2, (%%esi, %%edi)\n\t"
		"movq	%%mm3, 8(%%esi, %%edi)\n\t"
		"movq	%%mm0, (%%edx, %%edi)\n\t"
		"movq	%%mm1, 8(%%edx, %%edi)\n\t"
		"addl	$16, %%edi\n\t"
		"cmpl	%3, %%edi\n\t"
		" jb	2b\n\t"
		"addl	%2, %%esi\n\t"
		"cmpl	%1, %%esi\n\t"
		" jb	1b\n\t"
		:: "g" (buf), "m" (buf+128), "m" (two_m_plus_one<<3), "r" (two_m<<3),
		   "r" (sseW[m])
		: "%esi", "%edi", "%edx"
	);
    }

    /* Post IFFT complex multiply  plus IFFT complex conjugate*/
	asm volatile(
		"movl	$-1024, %%esi\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"movq	(%0, %%esi), %%mm0\n\t"
		"movq	8(%0, %%esi), %%mm1\n\t"
		"movq	%%mm0, %%mm2\n\t"
		"movq	%%mm1, %%mm3\n\t"
		PSWAP_MM("%%mm0", "%%mm6")
		PSWAP_MM("%%mm1", "%%mm7")
		"pfmul	1024+"MANGLE(sseSinCos1c)"(%%esi), %%mm2\n\t"
		"pfmul	1032+"MANGLE(sseSinCos1c)"(%%esi), %%mm3\n\t"
		"pfmul	1024+"MANGLE(sseSinCos1d)"(%%esi), %%mm0\n\t"
		"pfmul	1032+"MANGLE(sseSinCos1d)"(%%esi), %%mm1\n\t"
		"pfadd	%%mm2, %%mm0\n\t"
		"pfadd	%%mm3, %%mm1\n\t"
		"movq	%%mm0, (%0, %%esi)\n\t"
		"movq	%%mm1, 8(%0, %%esi)\n\t"
		"addl	$16, %%esi\n\t"
		" jnz	1b\n\t"
		:: "r" (buf+128)
		: "%esi"
	);   
	
    data_ptr = data;
    delay_ptr = delay;
    window_ptr = imdct_window;

    /* Window and convert to real valued signal */
	asm volatile(
		"xorl	%%edi, %%edi\n\t"
		"xorl	%%esi, %%esi\n\t"
		"movd	%3, %%mm4\n\t"
		"punpckldq %%mm4, %%mm4\n\t"
		"movq	%%mm4, %%mm5\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"movd	4(%0, %%esi), %%mm0\n\t"	// ? A
		"punpckldq -8(%0, %%edi), %%mm0\n\t"	// B A
		"movd	12(%0, %%esi), %%mm1\n\t"	// ? C
		"punpckldq -16(%0, %%edi), %%mm1\n\t"	// D C
		"pfmul	"MANGLE(sseWindow)"(%%esi), %%mm0\n\t"
		"pfmul	8+"MANGLE(sseWindow)"(%%esi), %%mm1\n\t"
		"pfadd	(%2, %%esi), %%mm0\n\t"
		"pfadd	8(%2, %%esi), %%mm1\n\t"
		"pfadd	%%mm4, %%mm0\n\t"
		"pfadd	%%mm5, %%mm1\n\t"
		"movq	%%mm0, (%1, %%esi)\n\t"
		"movq	%%mm1, 8(%1, %%esi)\n\t"
		"addl	$16, %%esi\n\t"
		"subl	$16, %%edi\n\t"
		"cmpl	$512, %%esi\n\t" 
		" jb	1b\n\t"
		:: "r" (buf+64), "r" (data_ptr), "r" (delay_ptr), "m" (bias)
		: "%esi", "%edi"
	);
	data_ptr+=128;
	delay_ptr+=128;
//	window_ptr+=128;

	asm volatile(
		"movl	$1024, %%edi\n\t"
		"xorl	%%esi, %%esi\n\t"
		"movd	%3, %%mm4\n\t"
		"punpckldq %%mm4, %%mm4\n\t"
		"movq	%%mm4, %%mm5\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"movd	(%0, %%esi), %%mm0\n\t"		// ? A
		"punpckldq -4(%0, %%edi), %%mm0\n\t"	// B A
		"movd	8(%0, %%esi), %%mm1\n\t"	// ? C
		"punpckldq -12(%0, %%edi), %%mm1\n\t"	// D C
		"pfmul	512+"MANGLE(sseWindow)"(%%esi), %%mm0\n\t"
		"pfmul	520+"MANGLE(sseWindow)"(%%esi), %%mm1\n\t"
		"pfadd	(%2, %%esi), %%mm0\n\t"
		"pfadd	8(%2, %%esi), %%mm1\n\t"
		"pfadd	%%mm4, %%mm0\n\t"
		"pfadd	%%mm5, %%mm1\n\t"
		"movq	%%mm0, (%1, %%esi)\n\t"
		"movq	%%mm1, 8(%1, %%esi)\n\t"
		"addl	$16, %%esi\n\t"
		"subl	$16, %%edi\n\t"
		"cmpl	$512, %%esi\n\t" 
		" jb	1b\n\t"
		:: "r" (buf), "r" (data_ptr), "r" (delay_ptr), "m" (bias)
		: "%esi", "%edi"
	);
	data_ptr+=128;
//	window_ptr+=128;

    /* The trailing edge of the window goes into the delay line */
    delay_ptr = delay;

	asm volatile(
		"xorl	%%edi, %%edi\n\t"
		"xorl	%%esi, %%esi\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"movd	(%0, %%esi), %%mm0\n\t"		// ? A
		"punpckldq -4(%0, %%edi), %%mm0\n\t"	// B A 
		"movd	8(%0, %%esi), %%mm1\n\t"	// ? C
		"punpckldq -12(%0, %%edi), %%mm1\n\t"	// D C 
		"pfmul	1024+"MANGLE(sseWindow)"(%%esi), %%mm0\n\t"
		"pfmul	1032+"MANGLE(sseWindow)"(%%esi), %%mm1\n\t"
		"movq	%%mm0, (%1, %%esi)\n\t"
		"movq	%%mm1, 8(%1, %%esi)\n\t"
		"addl	$16, %%esi\n\t"
		"subl	$16, %%edi\n\t"
		"cmpl	$512, %%esi\n\t" 
		" jb	1b\n\t"
		:: "r" (buf+64), "r" (delay_ptr)
		: "%esi", "%edi"
	);

	delay_ptr+=128;
//	window_ptr-=128;

	asm volatile(
		"movl	$1024, %%edi\n\t"
		"xorl	%%esi, %%esi\n\t"
		".balign 16\n\t"
		"1:\n\t"
		"movd	4(%0, %%esi), %%mm0\n\t"	// ? A
		"punpckldq -8(%0, %%edi), %%mm0\n\t"	// B A
		"movd	12(%0, %%esi), %%mm1\n\t"	// ? C
		"punpckldq -16(%0, %%edi), %%mm1\n\t"	// D C
		"pfmul	1536+"MANGLE(sseWindow)"(%%esi), %%mm0\n\t"
		"pfmul	1544+"MANGLE(sseWindow)"(%%esi), %%mm1\n\t"
		"movq	%%mm0, (%1, %%esi)\n\t"
		"movq	%%mm1, 8(%1, %%esi)\n\t"
		"addl	$16, %%esi\n\t"
		"subl	$16, %%edi\n\t"
		"cmpl	$512, %%esi\n\t" 
		" jb	1b\n\t"
		:: "r" (buf), "r" (delay_ptr)
		: "%esi", "%edi"
	);
	asm volatile("femms":::"memory");
}
