/* This is a modified version of simple_idct.c from ffmpeg.
 * Modifications by Mns Rullgrd <mru@users.sf.net>
 */

/*
    Copyright (C) 2001 Michael Niedermayer (michaelni@gmx.at)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
  based upon some outcommented c code from mpeg2dec (idct_mmx.c written by Aaron Holtzman <aholtzma@ess.engr.uvic.ca>)
*/

#include <inttypes.h>

#include "mpeg2.h"
#include "mpeg2_internal.h"
#include "asm_alpha.h"

extern void put_pixels_clamped_mvi_asm(const int16_t *block, uint8_t *pixels, 
				       int line_size);
extern void add_pixels_clamped_mvi_asm(const int16_t *block, uint8_t *pixels, 
				       int line_size);

void (*put_pixels_clamped_alpha_fn)(const int16_t *block, uint8_t *pixels, 
				    int line_size);
void (*add_pixels_clamped_alpha_fn)(const int16_t *block, uint8_t *pixels, 
				    int line_size);

#if 0
#define W1 2841 /* 2048*sqrt (2)*cos (1*pi/16) */
#define W2 2676 /* 2048*sqrt (2)*cos (2*pi/16) */
#define W3 2408 /* 2048*sqrt (2)*cos (3*pi/16) */
#define W4 2048 /* 2048*sqrt (2)*cos (4*pi/16) */
#define W5 1609 /* 2048*sqrt (2)*cos (5*pi/16) */
#define W6 1108 /* 2048*sqrt (2)*cos (6*pi/16) */
#define W7 565  /* 2048*sqrt (2)*cos (7*pi/16) */
#define ROW_SHIFT 8
#define COL_SHIFT 17
#else
#define W1  22725  //cos(i*M_PI/16)*sqrt(2)*(1<<14) + 0.5
#define W2  21407  //cos(i*M_PI/16)*sqrt(2)*(1<<14) + 0.5
#define W3  19266  //cos(i*M_PI/16)*sqrt(2)*(1<<14) + 0.5
#define W4  16384  //cos(i*M_PI/16)*sqrt(2)*(1<<14) + 0.5
#define W5  12873  //cos(i*M_PI/16)*sqrt(2)*(1<<14) + 0.5
#define W6  8867   //cos(i*M_PI/16)*sqrt(2)*(1<<14) + 0.5
#define W7  4520   //cos(i*M_PI/16)*sqrt(2)*(1<<14) + 0.5
#define ROW_SHIFT 11
#define COL_SHIFT 20 // 6
#endif

/* 0: all entries 0, 1: only first entry nonzero, 2: otherwise  */
static inline int idctRowCondDC(int16_t *row)
{
    int_fast32_t a0, a1, a2, a3, b0, b1, b2, b3;
    uint64_t l, r, t;
    uint64_t *lrow = (uint64_t *) row;

    l = lrow[0];
    r = lrow[1];

    if (r == 0) {
	if (l == 0)
	    return 0;
	if ((l & ~0xffffULL) == 0) {
	    uint64_t v;
	    v = (extwl(l, 0) << 3) & 0xffff;
	    v += v << 16;
	    v += v << 32;
	    lrow[0] = v;
	    lrow[1] = v;

	    return 1;
	}
    }

    a0 = a1 = a2 = a3 = W4 * sextw(extwl(l, 0));

    t = extwl(l, 4);
    if (t) {
	t = sextw(t);
	a0 += W2 * t;
	a1 += W6 * t;
	a2 -= W6 * t;
	a3 -= W2 * t;
    }

    t = extwl(r, 0);
    if (t) {
	t = sextw(t);
	a0 += W4 * t;
	a1 -= W4 * t;
	a2 -= W4 * t;
	a3 += W4 * t;
    }

    t = extwl(r, 4);
    if (t) {
	t = sextw(t);
	a0 += W6 * t;
	a1 -= W2 * t;
	a2 += W2 * t;
	a3 -= W6 * t;
    }

    a0 += 1 << (ROW_SHIFT - 1);
    a1 += 1 << (ROW_SHIFT - 1);
    a2 += 1 << (ROW_SHIFT - 1);
    a3 += 1 << (ROW_SHIFT - 1);

    t = extwl(l, 2);
    if (!t) {
	b0 = 0;
	b1 = 0;
	b2 = 0;
	b3 = 0;
    } else {
	t = sextw(t);
	b0 = W1 * t;
	b1 = W3 * t;
	b2 = W5 * t;
	b3 = W7 * t;
    }

    t = extwl(l, 6);
    if (t) {
	t = sextw(t);
	b0 += W3 * t;
	b1 -= W7 * t;
	b2 -= W1 * t;
	b3 -= W5 * t;
    }

    t = extwl(r, 2);
    if (t) {
	t = sextw(t);
	b0 += W5 * t;
	b1 -= W1 * t;
	b2 += W7 * t;
	b3 += W3 * t;
    }

    t = extwl(r, 6);
    if (t) {
	t = sextw(t);
	b0 += W7 * t;
	b1 -= W5 * t;
	b2 += W3 * t;
	b3 -= W1 * t;
    }

    row[0] = (a0 + b0) >> ROW_SHIFT;
    row[1] = (a1 + b1) >> ROW_SHIFT;
    row[2] = (a2 + b2) >> ROW_SHIFT;
    row[3] = (a3 + b3) >> ROW_SHIFT;
    row[4] = (a3 - b3) >> ROW_SHIFT;
    row[5] = (a2 - b2) >> ROW_SHIFT;
    row[6] = (a1 - b1) >> ROW_SHIFT;
    row[7] = (a0 - b0) >> ROW_SHIFT;

    return 2;
}

static inline void idctSparseCol (int16_t * col)
{
	int a0, a1, a2, a3, b0, b1, b2, b3;
	uint64_t c0 = col[0];
	c0 *= W4;
	c0 += (1<<(COL_SHIFT-1));
	a0 = a1 = a2 = a3 = c0;

	if(col[8*2]){
		a0 +=  + W2*col[8*2];
		a1 +=  + W6*col[8*2];
		a2 +=  - W6*col[8*2];
		a3 +=  - W2*col[8*2];
	}

	if(col[8*4]){
		a0 += + W4*col[8*4];
		a1 += - W4*col[8*4];
		a2 += - W4*col[8*4];
		a3 += + W4*col[8*4];
	}

	if(col[8*6]){
		a0 += + W6*col[8*6];
		a1 += - W2*col[8*6];
		a2 += + W2*col[8*6];
		a3 += - W6*col[8*6];
	}

	if(!col[8*1]){
		b0 = 
		b1 = 
		b2 = 
		b3 = 0;
	}else{
		b0 = W1*col[8*1];
		b1 = W3*col[8*1];
		b2 = W5*col[8*1];
		b3 = W7*col[8*1];
	}

	if(col[8*3]){
		b0 += + W3*col[8*3];
		b1 += - W7*col[8*3];
		b2 += - W1*col[8*3];
		b3 += - W5*col[8*3];
	}

	if(col[8*5]){
		b0 += + W5*col[8*5];
		b1 += - W1*col[8*5];
		b2 += + W7*col[8*5];
		b3 += + W3*col[8*5];
	}

	if(col[8*7]){
		b0 += + W7*col[8*7];
		b1 += - W5*col[8*7];
		b2 += + W3*col[8*7];
		b3 += - W1*col[8*7];
	}

	col[8*0] = (a0 + b0) >> COL_SHIFT;
	col[8*7] = (a0 - b0) >> COL_SHIFT;
	col[8*1] = (a1 + b1) >> COL_SHIFT;
	col[8*6] = (a1 - b1) >> COL_SHIFT;
	col[8*2] = (a2 + b2) >> COL_SHIFT;
	col[8*5] = (a2 - b2) >> COL_SHIFT;
	col[8*3] = (a3 + b3) >> COL_SHIFT;
	col[8*4] = (a3 - b3) >> COL_SHIFT;
}

/* If all rows but the first one are zero after row transformation,
   all rows will be identical after column transformation.  */
static inline void idctCol2(int16_t *col)
{
	uint64_t l, r, s, t;
	uint64_t *lcol = (uint64_t *) col;
	uint64_t sgn = WORD_VEC(0x0100);
	uint64_t sxt = WORD_VEC(0xfd00);

	l = lcol[0]; /* first 4 */
	r = lcol[1]; /* last 4 */

	/* clear lowest bit to avoid overflow from next element */
	l &= WORD_VEC(0xfffe);
	r &= WORD_VEC(0xfffe);

	l += WORD_VEC(32);
	r += WORD_VEC(32);

	/* clear low 6 bits so the shift won't clobber next element */
	l &= WORD_VEC(0xffc0);
	r &= WORD_VEC(0xffc0);
	l >>= 6;
	r >>= 6;

	/* sign extend negative values */
	s = cmpbge(l, sgn);
	t = cmpbge(r, sgn);
	l |= zapnot(sxt, s);
	r |= zapnot(sxt, t);

	lcol[ 0] = l; lcol[ 1] = r;
	lcol[ 2] = l; lcol[ 3] = r;
	lcol[ 4] = l; lcol[ 5] = r;
	lcol[ 6] = l; lcol[ 7] = r;
	lcol[ 8] = l; lcol[ 9] = r;
	lcol[10] = l; lcol[11] = r;
	lcol[12] = l; lcol[13] = r;
	lcol[14] = l; lcol[15] = r;
}

static void idct_alpha (short *block)
{
	int i;
        int rowsZero = 1;       /* all rows except row 0 zero */
        int rowsConstant = 1;	/* all rows consist of a constant value */

	for (i = 0; i < 8; i++) {
		int sparseness = idctRowCondDC(block + 8 * i);

		if (i > 0 && sparseness > 0)
                        rowsZero = 0;
                if (sparseness == 2)
                        rowsConstant = 0;
	}

        if (rowsZero) {
                idctCol2(block);
        } else if (rowsConstant) {
		uint64_t *lblock = (uint64_t *) block;

		idctSparseCol(block);
		for (i = 0; i < 8; i += 2) {
			uint64_t v = (uint16_t) block[i * 8];
			uint64_t w = (uint16_t) block[i*8 + 8];
			v += v << 16;
			w += w << 16;
			v += v << 32;
			w += w << 32;
			lblock[0] = v;
			lblock[1] = v;
			lblock[2] = w;
			lblock[3] = w;
			lblock += 4;
		}
	} else {
		for (i = 0; i < 8; i++)
			idctSparseCol(block + i);
	}
}

static inline uint64_t clampneg(uint64_t x)
{
    uint64_t neg = WORD_VEC(0x7fff);
    uint64_t nb, r;
    nb = cmpbge(neg, x);
    r = nb >> 1;
    nb &= r;
    r = zapnot(x, nb);
    return r;
}

static inline uint64_t clampneg2(uint64_t x)
{
    uint64_t neg = WORD_VEC(0x3fff);
    uint64_t nb, r;
    nb = cmpbge(neg, x);
    r = nb >> 1;
    nb &= r;
    r = zapnot(x, nb);
    return r;
}

static inline uint64_t clamphigh(uint64_t x)
{
    uint64_t max = WORD_VEC(0x00ff);
    uint64_t mb, r;
    mb = cmpbge(max, x);
    mb >>= 1;
    r = zap(max, mb);
    r |= x;
    return r;
}

static inline uint64_t em_unpkbw(uint64_t x)
{
    return (x & 0xff) | ((x & (0xff << 8)) << 8) |
	((x & (0xffL << 16)) << 16) | ((x & (0xffL << 24)) << 24);
}

static inline uint64_t em_pkwb(uint64_t x)
{
    return (x & 0xff) | ((x & (0xff << 16)) >> 8) |
	((x & (0xffL << 32)) >> 16) | ((x & (0xffL << 48)) >> 24);
}

extern void put_pixels_clamped_alpha(const int16_t *block,
					    uint8_t *pixels, 
					    int line_size)
{
    int i;
    for(i = 0; i < 4; i++){
	uint64_t p1, p2, p3, p4;
	p1 = ldq(block);
	p2 = ldq(block+4);
	p3 = ldq(block+8);
	p4 = ldq(block+12);

	p1 = clampneg(p1);
	p2 = clampneg(p2);
	p3 = clampneg(p3);
	p4 = clampneg(p4);
	p1 = clamphigh(p1);
	p2 = clamphigh(p2);
	p3 = clamphigh(p3);
	p4 = clamphigh(p4);

	stl(em_pkwb(p1), pixels);
	stl(em_pkwb(p2), pixels+4);
	pixels += line_size;
	stl(em_pkwb(p3), pixels);
	stl(em_pkwb(p4), pixels+4);
	pixels += line_size;
	block += 16;
    }
}

extern void add_pixels_clamped_alpha(const int16_t *block,
					    uint8_t *pixels, 
					    int line_size)
{
    int i;
    for(i = 0; i < 4; i++){
	uint64_t p1, p2, p3, p4;
	uint64_t q1, q2, q3, q4;
	uint8_t *px2 = pixels + line_size;

	p1 = ldq(block);
	p2 = ldq(block+4);
	p3 = ldq(block+8);
	p4 = ldq(block+12);

	q1 = em_unpkbw(ldl(pixels));
	q2 = em_unpkbw(ldl(pixels+4));
	q3 = em_unpkbw(ldl(px2));
	q4 = em_unpkbw(ldl(px2+4));

	p1 &= ~WORD_VEC(0x8000); /* clear highest bit to avoid overflow */
	p2 &= ~WORD_VEC(0x8000);
	p3 &= ~WORD_VEC(0x8000);
	p4 &= ~WORD_VEC(0x8000);

	p1 += q1;
	p2 += q2;
	p3 += q3;
	p4 += q4;

	p1 &= ~WORD_VEC(0x8000); /* hibit would be set for e. g. -2 + 3 */
	p2 &= ~WORD_VEC(0x8000);
	p3 &= ~WORD_VEC(0x8000);
	p4 &= ~WORD_VEC(0x8000);

	p1 = clampneg2(p1);
	p2 = clampneg2(p2);
	p3 = clampneg2(p3);
	p4 = clampneg2(p4);

	p1 = clamphigh(p1); /* clamp to 255 */
	p2 = clamphigh(p2);
	p3 = clamphigh(p3);
	p4 = clamphigh(p4);

	stl(em_pkwb(p1), pixels);
	stl(em_pkwb(p2), pixels+4);
	stl(em_pkwb(p3), px2);
	stl(em_pkwb(p4), px2+4);

	pixels += line_size*2;
	block += 16;
    }
}

void idct_block_copy_alpha(int16_t * block, uint8_t * dest, int stride)
{
    idct_alpha(block);
    put_pixels_clamped_alpha_fn(block, dest, stride);
}

void idct_block_add_alpha(int16_t * block, uint8_t * dest, int stride)
{
    idct_alpha(block);
    add_pixels_clamped_alpha_fn(block, dest, stride);
}
