/*
   Decoding ahead
   Licence: GPL v2
   Author: Nick Kurshev
   Note: Threaded engine to decode frames ahead
*/

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#define DA_PREFIX "DEC_AHEAD:"
#include "dec_ahead.h"
#include "libao2/audio_out.h"
#include "libao2/audio_plugin.h"
#include "libvo/video_out.h"

#include "libmpcodecs/dec_video.h"
#include "sig_hand.h"

#ifdef ENABLE_DEC_AHEAD_DEBUG
#define MSG_T(args...) mp_msg(MSGT_GLOBAL, MSGL_DBG2, ## args )
#else
#define MSG_T(args...)
#endif
pthread_mutex_t vdecoding_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t vreading_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t vdec_active_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t vdec_locked_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t vdeca_mutex=PTHREAD_MUTEX_INITIALIZER;

pthread_mutex_t seek_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t seek_cond_mutex=PTHREAD_MUTEX_INITIALIZER;;
pthread_cond_t seek_cond=PTHREAD_COND_INITIALIZER;;

extern ao_functions_t *audio_out;
extern vo_functions_t *video_out;
extern volatile int xp_drop_frame;
extern volatile unsigned xp_n_frame_to_drop;
extern unsigned xp_num_frames;
extern volatile unsigned xp_drop_frame_cnt;
extern int output_quality;
extern int auto_quality;

volatile int dec_ahead_locked_frame=0;
volatile unsigned abs_dec_ahead_locked_frame=0;
volatile unsigned abs_dec_ahead_blitted_frame=0;
extern volatile unsigned abs_dec_ahead_active_frame;
extern volatile unsigned dec_ahead_active_frame;
extern volatile unsigned loc_dec_ahead_active_frame;
volatile int dec_ahead_in_lseek=NoSeek;
volatile int dec_ahead_in_resize=0;
volatile float dec_ahead_seek_num_frames=0;       /* frames played after seek */
volatile int dec_ahead_seek_num_frames_decoded=0; /* frames decoded after seek */
volatile int dec_ahead_num_frames_decoded=0;      /* frames decoded by thread */


static pthread_t pthread_id=0;
static pthread_attr_t our_attr;
extern int verbose;
extern float vo_fps;
static sh_video_t sh_video;
static sh_audio_t sh_audio;
extern demux_stream_t *d_video;

/* Support for '-loop' option */
extern int loop_times; /* it's const for xp mode */

extern float rel_seek_secs;	/* FIXME: in hope that user will not rewind */
extern int abs_seek_pos;	/* the movie at end of file :( */

extern void mpxp_seek( int xp_id, void *, int *, float );
extern int decore_audio( void );

pid_t dec_ahead_pid; /* Only for testing */

extern void update_osd( float v_pts );

shva_t *shva;
int xp_eof=0;
#define NORM_FRAME(a) ((a)%xp_num_frames)

static volatile int pthread_is_living=0;
static volatile int pthread_end_of_work=0;
void * dec_ahead_routine( void * arg )
{
    float frame_time=0;
    int blit_frame=0;
    int drop_param=0;
    int can_skip_dropping;
    volatile unsigned da_active_frame,lda_active_frame,ada_active_frame;
    int xp_id;
    pthread_is_living=1;
    xp_eof = 0;
    MSG_T("\nDEC_AHEAD: entering...\n");
    xp_id=init_signal_handling(sig_dec_ahead_video);
    pinfo[xp_id].current_module = "dec_ahead";
    dec_ahead_pid =
    pinfo[xp_id].pid = getpid(); /* Only for testing */
    pinfo[xp_id].thread_name = "video decoding ahead";
while(!xp_eof){
    if(pthread_end_of_work) break;
    /*--------------------  Decode a frame: -----------------------*/
    {   unsigned char* start=NULL;
	int in_size;
        if(dec_ahead_in_lseek==PreSeek) {
            dec_ahead_in_lseek=Seek;
            pthread_cond_signal(&seek_cond);
            LOCK_SEEK();
            UNLOCK_SEEK();
        }
	/* get it! */
	LOCK_VREADING();
	if(dec_ahead_in_lseek==Seek || dec_ahead_in_lseek==Loop)
	{ /* Get info from player after a seek */
            //*((char*)0x100) = 1; // Testing crash
	    sh_video.num_frames = dec_ahead_seek_num_frames;
	    sh_video.num_frames_decoded = dec_ahead_seek_num_frames_decoded;
	}
	/* prevent reent access to non-reent demuxer */
        //if(sh_video.num_frames>200)  *((char*)0x100) = 1; // Testing crash
        in_size=video_read_frame(&sh_video,&frame_time,&start,vo_fps);
	UNLOCK_VREADING();
	if(dec_ahead_in_lseek==Seek || dec_ahead_in_lseek==Loop)
	{
	    /* reset counters */
	    vo_get_active_frame(&dec_ahead_locked_frame);
	    LOCK_VDEC_ACTIVE();
	    LOCK_VDEC_LOCKED();
	    abs_dec_ahead_locked_frame = abs_dec_ahead_active_frame;
	    abs_dec_ahead_blitted_frame = loc_dec_ahead_active_frame;
	    dec_ahead_locked_frame = dec_ahead_active_frame;
	    UNLOCK_VDEC_LOCKED();
	    UNLOCK_VDEC_ACTIVE();
	    dec_ahead_in_lseek=NoSeek;
            pthread_cond_signal(&seek_cond);
	    MSG_T("\nDEC_AHEAD: reset counters to (%u %u) due lseek\n",dec_ahead_locked_frame,abs_dec_ahead_locked_frame);
	}
	if(in_size<0)
	{
	    /* Looping. */
	    if(loop_times>=0) {
		MSG_V("loop_times = %d\n", loop_times);
                while(abs_dec_ahead_locked_frame>abs_dec_ahead_active_frame &&
                      dec_ahead_in_lseek==NoSeek && !pthread_end_of_work)
                    usleep(1); /* Wait for video buffer to become empty */
                if(pthread_end_of_work) // User pressed quit
                    break;
                if(dec_ahead_in_lseek!=NoSeek) // User pressed seek
                    continue;
		dec_ahead_in_lseek=Loop;
		if(loop_times>1) loop_times--; 
		else
		if(loop_times==1) loop_times=-1;
		pinfo[xp_id].current_module="seek";
		abs_seek_pos=3; rel_seek_secs=0; // seek to start of movie (0%)
		LOCK_VREADING();
		mpxp_seek(xp_id,NULL,NULL,0.);
		decore_audio();
		UNLOCK_VREADING();
		rel_seek_secs=0;
		abs_seek_pos=0;
		pinfo[xp_id].current_module=NULL;
		continue;
	    }
	    LOCK_VDECA();
	    shva[dec_ahead_locked_frame].eof=1;
	    UNLOCK_VDECA();
	    xp_eof=1;
	    break; 
	}
	/* in_size==0: it's or broken stream or demuxer's bug */
	if(in_size==0 && !pthread_end_of_work) { dec_ahead_in_lseek=NoSeek; continue; }
	vo_set_frame_num(&dec_ahead_locked_frame);
#ifdef ENABLE_DEC_AHEAD_DEBUG
	if(xp_n_frame_to_drop && xp_drop_frame) 
	  MSG_T("\nDEC_AHEAD: frame drop request\n");
	if(verbose)
	    MSG_T("\nDEC_AHEAD: set frame %u\n",dec_ahead_locked_frame);
#endif
	/* ------- compute frame dropping --------------*/
	if(xp_n_frame_to_drop && xp_drop_frame)
	{
	    /* Adaptive frame dropping */
	    /* means we should evenly redistribute number of frame
	       to be dropped within of buffers of decoding ahead */
	    if(xp_n_frame_to_drop >= xp_num_frames) can_skip_dropping=0;
	    else can_skip_dropping=xp_num_frames%xp_n_frame_to_drop?0:1;
	    MSG_T("xp_num_frames=%u xp_n_f_to_dr=%u can_skip_dr=%i\n"
		    ,xp_num_frames,xp_n_frame_to_drop,can_skip_dropping);
	}
	else can_skip_dropping=1;
	/* ------------ sleep --------------- */
	LOCK_VDEC_ACTIVE();
	ada_active_frame= abs_dec_ahead_active_frame;
	lda_active_frame= loc_dec_ahead_active_frame;
	da_active_frame = dec_ahead_active_frame;
	UNLOCK_VDEC_ACTIVE();
	/* sleep if thread is too fast ;) */
	while(abs_dec_ahead_blitted_frame >= lda_active_frame+xp_num_frames-2)
	{
	    MSG_T("DEC_AHEAD: sleep (abs (blitted(%u)>=active+xp-2(%u)))\n"
		,abs_dec_ahead_blitted_frame,lda_active_frame+xp_num_frames-2);
	    if(pthread_end_of_work) goto pt_exit;
	    if(dec_ahead_in_lseek!=NoSeek) break;
	    usleep(1);
	    LOCK_VDEC_ACTIVE();
	    lda_active_frame= loc_dec_ahead_active_frame;
	    da_active_frame = dec_ahead_active_frame;
	    UNLOCK_VDEC_ACTIVE();
	}
	if(dec_ahead_in_lseek!=NoSeek) continue;
	LOCK_VDECODING();
#if 0
{ 
/* 
    We can't seriously examine question of too slow machines
    by motivation reasons
*/
int i,delay; /* sleeping 200 ms is far enough for 25 fps */
if(can_skip_dropping) delay=20;
else delay=xp_drop_frame?0:20;
for(i=0;i<delay;i++) usleep(0);
}
#endif
	drop_param = can_skip_dropping?0:xp_drop_frame;
	/* decode: */
	MSG_T("\nDEC_AHEAD: decode to %u (abs (blitted(%u)>=active+xp-2(%u)))\n"
		,abs_dec_ahead_blitted_frame,abs_dec_ahead_locked_frame,lda_active_frame+xp_num_frames-2);
	if(output_quality)
	{
	    unsigned total = xp_num_frames/2;
	    unsigned distance = abs_dec_ahead_blitted_frame-lda_active_frame;
	    int our_quality;
	    our_quality = output_quality*distance/total;
	    if(drop_param) set_video_quality(&sh_video,0);
	    else
	    if(auto_quality) set_video_quality(&sh_video,our_quality>0?our_quality:0);
	}
	blit_frame=decode_video(&sh_video,start,in_size,drop_param);
	if(output_quality)
	{
	    if(drop_param) set_video_quality(&sh_video,output_quality);
	}
	if(!can_skip_dropping)
	{
	    xp_drop_frame_cnt++;
	    xp_n_frame_to_drop--;
	}
	UNLOCK_VDECODING();
	if(dec_ahead_in_resize)
	{
	    dec_ahead_in_resize=0;
	}
#ifdef ENABLE_DEC_AHEAD_DEBUG
	if(verbose)
	{
	MSG_T("\nDEC_AHEAD: frame %u decoded (blit=%u blit_param=%u size=%i)\n"
	,abs_dec_ahead_locked_frame,blit_frame,drop_param,in_size);
	}
#endif
	LOCK_VDEC_LOCKED();
	abs_dec_ahead_locked_frame++;
	if(blit_frame)
	{
	    LOCK_VDECA();
	    shva[dec_ahead_locked_frame].v_pts = d_video->pts;
	    shva[dec_ahead_locked_frame].frametime=frame_time;
	    shva[dec_ahead_locked_frame].eof=0;
	    shva[dec_ahead_locked_frame].num_frames = sh_video.num_frames;
	    shva[dec_ahead_locked_frame].num_frames_decoded = sh_video.num_frames_decoded;
	    dec_ahead_num_frames_decoded = sh_video.num_frames_decoded;
	    UNLOCK_VDECA();
	    dec_ahead_locked_frame=(dec_ahead_locked_frame+1)%xp_num_frames;
	    abs_dec_ahead_blitted_frame++;
	}
	else 
	  if(in_size && !drop_param)
	  /* do not count broken frames */
		abs_dec_ahead_locked_frame--;
	UNLOCK_VDEC_LOCKED();
    }
    /*------------------------ frame decoded. --------------------*/
} /* while(!xp_eof)*/
  pt_exit:
  MSG_T("\nDEC_AHEAD: leaving...\n");
  pthread_is_living=0;
  pthread_end_of_work=0;
  return arg; /* terminate thread here !!! */
}

int uninit_dec_ahead( void )
{
  if(pthread_id && pthread_is_living) 
  {
    pthread_end_of_work=1;
    while(pthread_is_living) usleep(0);
    pthread_attr_destroy(&our_attr);
    free(shva);
  }
  return 0;
}

int init_dec_ahead(const sh_video_t *shv,const sh_audio_t *sha)
{
  pthread_attr_init(&our_attr);
  sh_video = *shv;
  if(!(shva = malloc(sizeof(shva_t)*xp_num_frames))) return ENOMEM;
  if(sha) sh_audio = *sha; /* currently is unused */
  dec_ahead_locked_frame=0;
  abs_dec_ahead_locked_frame=0;
  abs_dec_ahead_blitted_frame=0;
  dec_ahead_in_lseek=NoSeek;
  dec_ahead_in_resize=0;
  return 0;
}

int run_dec_ahead( void )
{
  int retval;
  retval = pthread_attr_setdetachstate(&our_attr,PTHREAD_CREATE_DETACHED);
  if(retval) 
  {
    if(verbose) printf("running thread: attr_setdetachstate fault!!!\n");
    return retval;
  }    
  pthread_attr_setscope(&our_attr,PTHREAD_SCOPE_SYSTEM);
#if 0
  /* requires root privelegies */
  pthread_attr_setschedpolicy(&our_attr,SCHED_FIFO);
#endif
  return pthread_create(&pthread_id,&our_attr,dec_ahead_routine,NULL);
}

extern void exit_sighandler(void);
void sig_dec_ahead_video( void )
{
    MSG_T("sig_dec_ahead_video\n");
    mp_msg_flush();

    xp_eof = 1;
    shva[dec_ahead_locked_frame].eof=1;
    /*
	Unlock all mutex 
	( man page says it may deadlock, but what is worse deadlock here or later? )
    */
    UNLOCK_VREADING();
    UNLOCK_VDEC_LOCKED();
    UNLOCK_VDEC_ACTIVE();
    UNLOCK_VDECA();
    UNLOCK_VDECODING();

    pthread_is_living=0;
    dec_ahead_in_lseek=NoSeek;
    pthread_kill(getpid(),SIGKILL);
    exit_sighandler();
}
