/*
   Decoding ahead
   Licence: GPL v2
   Author: Nick Kurshev
   Note: Threaded engine to decode frames ahead
*/

#ifndef __DEC_AHEAD_H
#define __DEC_AHEAD_H

#include <inttypes.h>
#include <pthread.h>
#include "libmpdemux/stream.h"
#include "libmpdemux/demuxer.h"
#include "libmpdemux/stheader.h"
#include "libvo/video_out.h"

//#define ENABLE_DEC_AHEAD_DEBUG 1

enum seek_states { NoSeek=0, PreSeek, Seek, Loop };

extern pthread_mutex_t vdec_active_mutex; /* it's related with video decoding (main process) */
extern pthread_mutex_t vdec_locked_mutex; /* it's related with video decoding (thread) */
extern pthread_mutex_t vreading_mutex; /* it's related with reading of video stream  */
extern pthread_mutex_t vdecoding_mutex; /* it's related with video decoding process */
extern pthread_mutex_t vdeca_mutex; /* it's related with video decoding attributes */

extern pthread_mutex_t seek_mutex; /* it's related with seek */
extern pthread_mutex_t seek_cond_mutex; /* it's related with seek */
extern pthread_cond_t seek_cond; /* it's related with seek */

#ifdef ENABLE_DEC_AHEAD_DEBUG
#include "mp_msg.h"
#ifndef DA_PREFIX
#define DA_PREFIX "dec_ahead:"
#endif
#define MSG_D(args...) { mp_msg(MSGT_GLOBAL, MSGL_V, ## args ); mp_msg_flush(); }
#else
#define MSG_D(args...)
#endif
#define LOCK_VDEC_ACTIVE() { MSG_D(DA_PREFIX"LOCK_VDEC_ACTIVE\n"); pthread_mutex_lock(&vdec_active_mutex); }
#define UNLOCK_VDEC_ACTIVE() { MSG_D(DA_PREFIX"UNLOCK_VDEC_ACTIVE\n"); pthread_mutex_unlock(&vdec_active_mutex); }
#define LOCK_VDEC_LOCKED() { MSG_D(DA_PREFIX"LOCK_VDEC_LOCKED\n"); pthread_mutex_lock(&vdec_locked_mutex); }
#define UNLOCK_VDEC_LOCKED() { MSG_D(DA_PREFIX"UNLOCK_VDEC_LOCKED\n"); pthread_mutex_unlock(&vdec_locked_mutex); }
#define LOCK_VDECODING() { MSG_D(DA_PREFIX"LOCK_VDECODING\n"); pthread_mutex_lock(&vdecoding_mutex); }
#define UNLOCK_VDECODING() { MSG_D(DA_PREFIX"UNLOCK_VDECODING\n"); pthread_mutex_unlock(&vdecoding_mutex); }
#define LOCK_VREADING() { MSG_D(DA_PREFIX"LOCK_VREADING\n"); pthread_mutex_lock(&vreading_mutex); }
#define UNLOCK_VREADING() { MSG_D(DA_PREFIX"UNLOCK_VREADING\n"); pthread_mutex_unlock(&vreading_mutex); }
#define LOCK_VDECA() { MSG_D(DA_PREFIX"LOCK_VDECA\n"); pthread_mutex_lock(&vdeca_mutex); }
#define UNLOCK_VDECA() { MSG_D(DA_PREFIX"UNLOCK_VDECA\n"); pthread_mutex_unlock(&vdeca_mutex); }
#define LOCK_SEEK() { MSG_D(DA_PREFIX"LOCK_SEEK\n"); pthread_mutex_lock(&seek_mutex); }
#define UNLOCK_SEEK() { MSG_D(DA_PREFIX"UNLOCK_SEEK\n"); pthread_mutex_unlock(&seek_mutex); }
#define LOCK_SEEK_COND() { MSG_D(DA_PREFIX"LOCK_SEEK_COND\n"); pthread_mutex_lock(&seek_cond_mutex); }
#define UNLOCK_SEEK_COND() { MSG_D(DA_PREFIX"UNLOCK_SEEK_COND\n"); pthread_mutex_unlock(&seek_cond_mutex); }

typedef struct sh_video_attr
{
  int eof;
  float frametime;
  float v_pts;
  float num_frames;             // number of frames played
  int num_frames_decoded;       // number of frames decoded
}shva_t;

extern shva_t* shva;

extern volatile int dec_ahead_locked_frame;
extern volatile unsigned abs_dec_ahead_locked_frame;
extern volatile unsigned abs_dec_ahead_blitted_frame;
extern volatile int dec_ahead_in_lseek;
extern volatile int dec_ahead_in_resize;
extern volatile float dec_ahead_seek_num_frames;       // frames played after seek
extern volatile int dec_ahead_seek_num_frames_decoded; // frames decoded after seek
extern volatile int dec_ahead_num_frames_decoded;      // frames decoded by thread

			/* 
			   stream - pointer to openned stream
			   astream - pointer to audio stream
			 */

extern int dec_ahead_pid; /* Only for testing */

extern int init_dec_ahead(const sh_video_t*stream,const sh_audio_t *astream);
extern int uninit_dec_ahead( void );
extern int run_dec_ahead( void );

extern void sig_dec_ahead_video( void );

#endif
