#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h> /* GLIBC specific. Exists under cygwin too! */
#define __AD_ACM
#include "ad_internal.h"
#include "codecs_ld.h"

#include "config.h"
#include "help_mp.h"

static ad_info_t info = 
{
	"DivX audio (WMA)",
	"acm",
	"Nick Kurshev",
	""
};

LIBAD_EXTERN(msacm)

/* ACM audio and VfW video codecs initialization */
/* based on the avifile library [http://divx.euro.ru] */
MMRESULT WINAPI (*acmStreamConvert_ptr)(
  HACMSTREAM has, PACMSTREAMHEADER pash, DWORD fdwConvert
);
MMRESULT WINAPI (*acmStreamOpen_ptr)(
  PHACMSTREAM phas, HACMDRIVER had, PWAVEFORMATEX pwfxSrc,
  PWAVEFORMATEX pwfxDst, PWAVEFILTER pwfltr, DWORD dwCallback,
  DWORD dwInstance, DWORD fdwOpen
);
MMRESULT WINAPI (*acmStreamClose_ptr)(
  HACMSTREAM has, DWORD fdwClose
);
MMRESULT WINAPI (*acmStreamPrepareHeader_ptr)(
  HACMSTREAM has, PACMSTREAMHEADER pash, DWORD fdwPrepare
);
MMRESULT WINAPI (*acmStreamSize_ptr)(
  HACMSTREAM has, DWORD cbInput, 
  LPDWORD pdwOutputBytes, DWORD fdwSize
);
MMRESULT WINAPI (*acmStreamUnprepareHeader_ptr)(
  HACMSTREAM has, PACMSTREAMHEADER pash, DWORD fdwUnprepare
);
static char **win32_codec_name;

static void *dll_handle;

static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  acmStreamConvert_ptr = dlsym(dll_handle,"acmStreamConvert");
  acmStreamOpen_ptr = dlsym(dll_handle,"acmStreamOpen");
  acmStreamClose_ptr = dlsym(dll_handle,"acmStreamClose");
  acmStreamPrepareHeader_ptr = dlsym(dll_handle,"acmStreamPrepareHeader");
  acmStreamSize_ptr = dlsym(dll_handle,"acmStreamSize");
  acmStreamUnprepareHeader_ptr = dlsym(dll_handle,"acmStreamUnprepareHeader");
  win32_codec_name = dlsym(dll_handle,"win32_codec_name");
  return acmStreamConvert_ptr && acmStreamOpen_ptr &&
	 acmStreamPrepareHeader_ptr && acmStreamSize_ptr &&
	 acmStreamUnprepareHeader_ptr && win32_codec_name && acmStreamClose_ptr;
}


int init_acm_audio_codec(sh_audio_t *sh_audio){
    HRESULT ret;
    WAVEFORMATEX *in_fmt=sh_audio->wf;
    unsigned int srcsize=0;

    MSG_V("======= Win32 (ACM) AUDIO Codec init =======\n");

    sh_audio->srcstream=0;

    sh_audio->o_wf.nChannels=in_fmt->nChannels;
    sh_audio->o_wf.nSamplesPerSec=in_fmt->nSamplesPerSec;
    sh_audio->o_wf.nAvgBytesPerSec=2*sh_audio->o_wf.nSamplesPerSec*sh_audio->o_wf.nChannels;
    sh_audio->o_wf.wFormatTag=WAVE_FORMAT_PCM;
    sh_audio->o_wf.nBlockAlign=2*in_fmt->nChannels;
    sh_audio->o_wf.wBitsPerSample=16;
//    sh_audio->o_wf.wBitsPerSample=in_fmt->wBitsPerSample;
    sh_audio->o_wf.cbSize=0;

    MSG_V("Input format:\n"
    "  wFormatTag %d\n"
    "  nChannels %d\n"
    "  nSamplesPerSec %d\n"
    "  nAvgBytesPerSec %d\n"
    "  nBlockAlign %d\n"
    "  wBitsPerSample %d\n"
    "  cbSize %d\n"
    "Output fmt:\n"
    "  wFormatTag %d\n"
    "  nChannels %d\n"
    "  nSamplesPerSec %d\n"
    "  nAvgBytesPerSec %d\n"
    "  nBlockAlign %d\n"
    "  wBitsPerSample %d\n"
    "  cbSize %d\n"
    , in_fmt->wFormatTag
    , in_fmt->nChannels
    , in_fmt->nSamplesPerSec
    , in_fmt->nAvgBytesPerSec
    , in_fmt->nBlockAlign
    , in_fmt->wBitsPerSample
    , in_fmt->cbSize
    , sh_audio->o_wf.wFormatTag
    , sh_audio->o_wf.nChannels
    , sh_audio->o_wf.nSamplesPerSec
    , sh_audio->o_wf.nAvgBytesPerSec
    , sh_audio->o_wf.nBlockAlign
    , sh_audio->o_wf.wBitsPerSample
    , sh_audio->o_wf.cbSize);

    *win32_codec_name = sh_audio->codec->dll_name;
    ret=(*acmStreamOpen_ptr)(&sh_audio->srcstream,(HACMDRIVER)NULL,
                    in_fmt,&sh_audio->o_wf,
		    NULL,0,0,0);
    if(ret){
        if(ret==ACMERR_NOTPOSSIBLE)
            MSG_ERR("ACM_Decoder: Unappropriate audio format\n");
        else
            MSG_ERR("ACM_Decoder: acmStreamOpen error: %d", (int)ret);
        sh_audio->srcstream=0;
        return 0;
    }
    MSG_V("Audio codec opened OK! ;-)\n");

    (*acmStreamSize_ptr)(sh_audio->srcstream, in_fmt->nBlockAlign, &srcsize, ACM_STREAMSIZEF_SOURCE);
    srcsize*=2;
    //if(srcsize<MAX_OUTBURST) srcsize=MAX_OUTBURST;
    if(!srcsize){
        MSG_WARN("Warning! ACM codec reports srcsize=0\n");
        srcsize=16384;
    }
    // limit srcsize to 4-16kb
    //while(srcsize && srcsize<4096) srcsize*=2;
    //while(srcsize>16384) srcsize/=2;
    sh_audio->audio_out_minsize=srcsize; // audio output min. size
    MSG_V("Audio ACM output buffer min. size: %ld\n",srcsize);

    (*acmStreamSize_ptr)(sh_audio->srcstream, srcsize, &srcsize, ACM_STREAMSIZEF_DESTINATION);
    sh_audio->audio_in_minsize=srcsize; // audio input min. size
    MSG_V("Audio ACM input buffer min. size: %ld\n",srcsize);
    
    if(srcsize<in_fmt->nBlockAlign) srcsize=in_fmt->nBlockAlign;

    sh_audio->a_in_buffer_size=2*sh_audio->audio_in_minsize;
    sh_audio->a_in_buffer=malloc(sh_audio->a_in_buffer_size);
    sh_audio->a_in_buffer_len=0;

    return 1;
}

int acm_decode_audio(sh_audio_t *sh_audio, void* a_buffer,int minlen,unsigned maxlen){
        ACMSTREAMHEADER ash;
        HRESULT hr;
        DWORD srcsize=0;
        DWORD len=minlen;
        (*acmStreamSize_ptr)(sh_audio->srcstream,len , &srcsize, ACM_STREAMSIZEF_DESTINATION);
        MSG_DBG3("acm says: srcsize=%ld  (buffsize=%d)  out_size=%d\n",srcsize,sh_audio->a_in_buffer_size,len);

        if(srcsize<sh_audio->wf->nBlockAlign){
           srcsize=sh_audio->wf->nBlockAlign;
           (*acmStreamSize_ptr)(sh_audio->srcstream, srcsize, &len, ACM_STREAMSIZEF_SOURCE);
           if(len>maxlen) len=maxlen;
        }

//        if(srcsize==0) srcsize=((WAVEFORMATEX *)&sh_audio->o_wf_ext)->nBlockAlign;
        if(srcsize>sh_audio->a_in_buffer_size) srcsize=sh_audio->a_in_buffer_size; // !!!!!!
        if(sh_audio->a_in_buffer_len<srcsize){
          sh_audio->a_in_buffer_len+=
            demux_read_data_r(sh_audio->ds,&sh_audio->a_in_buffer[sh_audio->a_in_buffer_len],
            srcsize-sh_audio->a_in_buffer_len);
        }
        MSG_DBG3("acm convert %d -> %d bytes\n",sh_audio->a_in_buffer_len,len);
        memset(&ash, 0, sizeof(ash));
        ash.cbStruct=sizeof(ash);
        ash.fdwStatus=0;
        ash.dwUser=0; 
        ash.pbSrc=sh_audio->a_in_buffer;
        ash.cbSrcLength=sh_audio->a_in_buffer_len;
        ash.pbDst=a_buffer;
        ash.cbDstLength=len;
        hr=(*acmStreamPrepareHeader_ptr)(sh_audio->srcstream,&ash,0);
        if(hr){
          MSG_V("ACM_Decoder: acmStreamPrepareHeader error %d\n",(int)hr);
					return -1;
        }
        hr=(*acmStreamConvert_ptr)(sh_audio->srcstream,&ash,0);
        if(hr){
          MSG_DBG2("ACM_Decoder: acmStreamConvert error %d\n",(int)hr);
          switch(hr)
	  {
	    case ACMERR_NOTPOSSIBLE:
	    case ACMERR_UNPREPARED:
		MSG_DBG2( "ACM_Decoder: acmStreamConvert error: probarly not initialized!\n");
	  }  
//					return -1;
        }
        if(verbose>1)
          MSG_DBG2("acm converted %d -> %d\n",ash.cbSrcLengthUsed,ash.cbDstLengthUsed);
        if(ash.cbSrcLengthUsed>=sh_audio->a_in_buffer_len){
          sh_audio->a_in_buffer_len=0;
        } else {
          sh_audio->a_in_buffer_len-=ash.cbSrcLengthUsed;
          memcpy(sh_audio->a_in_buffer,&sh_audio->a_in_buffer[ash.cbSrcLengthUsed],sh_audio->a_in_buffer_len);
        }
        len=ash.cbDstLengthUsed;
        hr=(*acmStreamUnprepareHeader_ptr)(sh_audio->srcstream,&ash,0);
        if(hr){
          MSG_V("ACM_Decoder: acmStreamUnprepareHeader error %d\n",(int)hr);
        }
        return len;
}

int close_acm_audio_codec(sh_audio_t *sh_audio)
{
    HRESULT ret;
    
    ret = (*acmStreamClose_ptr)(sh_audio->srcstream, 0);
    
    if (ret)
    switch(ret)
    {
	case ACMERR_BUSY:
	case ACMERR_CANCELED:
	    MSG_DBG2( "ACM_Decoder: stream busy, waiting..\n");
	    sleep(100);
	    return(close_acm_audio_codec(sh_audio));
	case ACMERR_UNPREPARED:
	case ACMERR_NOTPOSSIBLE:
	    return(0);
	default:
	    MSG_WARN( "ACM_Decoder: unknown error occured: %d\n", ret);
	    return(0);
    }    
/*    MSACM_UnregisterAllDrivers();*/
    return(1);
}

int init(sh_audio_t *sh_audio)
{
    int ret=acm_decode_audio(sh_audio,sh_audio->a_buffer,4096,sh_audio->a_buffer_size);
    if(ret<0){
        MSG_INFO("ACM decoding error: %d\n",ret);
        return 0;
    }
    sh_audio->a_buffer_len=ret;
  return 1;
}

int preinit(sh_audio_t *sh_audio)
{
  /* Win32 ACM audio codec: */
  if(!load_lib(wineld_name("libloader.so"))) return 0;
  if(init_acm_audio_codec(sh_audio)){
    sh_audio->i_bps=sh_audio->wf->nAvgBytesPerSec;
    sh_audio->channels=sh_audio->o_wf.nChannels;
    sh_audio->samplerate=sh_audio->o_wf.nSamplesPerSec;
  } else {
    MSG_ERR(MSGTR_ACMiniterror);
    return 0;
  }
  MSG_V("INFO: Win32/DShow (%s) audio codec init OK!\n",LIBDIR"/wine/libloader.so");
  return 1;
}

void uninit(sh_audio_t *sh)
{
  close_acm_audio_codec(sh);
  dlclose(dll_handle);
}

int control(sh_audio_t *sh_audio,int cmd,void* arg, ...)
{
  int skip;
    switch(cmd)
    {
      case ADCTRL_RESYNC_STREAM:
          sh_audio->a_in_buffer_len=0;/* reset ACM/DShow audio buffer */
	  return CONTROL_TRUE;
      case ADCTRL_SKIP_FRAME:
		    skip=sh_audio->wf->nBlockAlign;
		    if(skip<16){
		      skip=(sh_audio->wf->nAvgBytesPerSec/16)&(~7);
		      if(skip<16) skip=16;
		    }
		    demux_read_data_r(sh_audio->ds,NULL,skip);
	  return CONTROL_TRUE;
      default:
	  return CONTROL_UNKNOWN;
    }
  return CONTROL_UNKNOWN;
}

int decode_audio(sh_audio_t *sh_audio,unsigned char *buf,int minlen,int maxlen)
{
  return acm_decode_audio(sh_audio,buf,minlen,maxlen);
}
