#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include "config.h"
#include "ad_internal.h"
#include "../libao2/afmt.h"
#include "interface/adpcm.h"
#include "codecs_ld.h"

static ad_info_t info = 
{
	"Duck DK3 ADPCM (rogue format number) audio decoder",
	"dk3adpcm",
	"Nick Kurshev",
	"This format number was used by Duck Corp. but not officially registered with Microsoft"
};

LIBAD_EXTERN(dk3adpcm)
static int (*dk3_adpcm_decode_block_ptr)(unsigned short *output, unsigned char *input);
#define dk3_adpcm_decode_block(a,b) (*dk3_adpcm_decode_block_ptr)(a,b)

static void *dll_handle;
static int load_dll(const char *libname)
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  dk3_adpcm_decode_block_ptr = dlsym(dll_handle,"dk3_adpcm_decode_block");
  return dk3_adpcm_decode_block_ptr != NULL;
}

int init(sh_audio_t *sh_audio)
{
  sh_audio->channels=sh_audio->wf->nChannels;
  sh_audio->samplerate=sh_audio->wf->nSamplesPerSec;
  sh_audio->i_bps=DK3_ADPCM_BLOCK_SIZE*
    (sh_audio->channels*sh_audio->samplerate) / DK3_ADPCM_SAMPLES_PER_BLOCK;
  return 1;
}

int preinit(sh_audio_t *sh_audio)
{
  sh_audio->audio_out_minsize=DK3_ADPCM_SAMPLES_PER_BLOCK * 4;
  sh_audio->ds->ss_div=DK3_ADPCM_SAMPLES_PER_BLOCK;
  sh_audio->audio_in_minsize=
  sh_audio->ds->ss_mul=DK3_ADPCM_BLOCK_SIZE;
  return load_dll(codec_name("libadpcm.so"));
}

void uninit(sh_audio_t *sh)
{
  dlclose(dll_handle);
}

int control(sh_audio_t *sh,int cmd,void* arg, ...)
{
  return CONTROL_UNKNOWN;
}

int decode_audio(sh_audio_t *sh_audio,unsigned char *buf,int minlen,int maxlen)
{
  int len=-1;
   unsigned char ibuf[DK3_ADPCM_BLOCK_SIZE * 2]; /* bytes / stereo frame */
    if (demux_read_data_r(sh_audio->ds, ibuf,
          DK3_ADPCM_BLOCK_SIZE * sh_audio->wf->nChannels) != 
          DK3_ADPCM_BLOCK_SIZE * sh_audio->wf->nChannels) 
      return len; /* EOF */
    len = 2 * dk3_adpcm_decode_block(
          (unsigned short*)buf,ibuf);
    return len;
}
