#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include "config.h"
#include "ad_internal.h"
#include "../libao2/afmt.h"
#include "interface/adpcm.h"
#include "codecs_ld.h"

static ad_info_t info = 
{
	"Duck DK4 ADPCM (rogue format number) audio decoder",
	"dk4adpcm",
	"Nick Kurshev",
	"This format number was used by Duck Corp. but not officially registered with Microsoft"
};

LIBAD_EXTERN(dk4adpcm)
static int (*dk4_adpcm_decode_block_ptr)(unsigned short *output, unsigned char *input,
  int channels, int block_size);
#define dk4_adpcm_decode_block(a,b,c,d) (*dk4_adpcm_decode_block_ptr)(a,b,c,d)

static void *dll_handle;
static int load_dll(const char *libname)
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  dk4_adpcm_decode_block_ptr = dlsym(dll_handle,"dk4_adpcm_decode_block");
  return dk4_adpcm_decode_block_ptr != NULL;
}

int init(sh_audio_t *sh_audio)
{
  sh_audio->channels=sh_audio->wf->nChannels;
  sh_audio->samplerate=sh_audio->wf->nSamplesPerSec;
  sh_audio->i_bps = sh_audio->wf->nBlockAlign *
    (sh_audio->channels*sh_audio->samplerate) / DK4_ADPCM_SAMPLES_PER_BLOCK;
  return 1;
}

int preinit(sh_audio_t *sh_audio)
{
  sh_audio->audio_out_minsize=DK4_ADPCM_SAMPLES_PER_BLOCK * 4;
  sh_audio->ds->ss_div=DK4_ADPCM_SAMPLES_PER_BLOCK;
  sh_audio->audio_in_minsize=
  sh_audio->ds->ss_mul=sh_audio->wf->nBlockAlign;
  return load_dll(codec_name("libadpcm.so"));
}

void uninit(sh_audio_t *sh)
{
  dlclose(dll_handle);
}

int control(sh_audio_t *sh,int cmd,void* arg, ...)
{
  return CONTROL_UNKNOWN;
}

int decode_audio(sh_audio_t *sh_audio,unsigned char *buf,int minlen,int maxlen)
{
  int len=-1;
  static unsigned char *ibuf = NULL;
  if (!ibuf)
    ibuf = (unsigned char *)malloc(sh_audio->wf->nBlockAlign);
  if (demux_read_data_r(sh_audio->ds, ibuf, sh_audio->wf->nBlockAlign) != 
      sh_audio->wf->nBlockAlign)
          return len; /* EOF */
  len=2*dk4_adpcm_decode_block((unsigned short*)buf,ibuf,
          sh_audio->wf->nChannels, sh_audio->wf->nBlockAlign);
  return len;
}
