#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <dlfcn.h> /* GLIBC specific. Exists under cygwin too! */
#include "ad_internal.h"

#include "config.h"
#include "help_mp.h"
#include "cpudetect.h"

#include "interface/a52.h"
#include "codecs_ld.h"
#include "../libao2/afmt.h"

extern sample_t * a52_samples;
extern a52_state_t a52_state;
extern uint32_t a52_accel;
extern uint32_t a52_flags;
extern int a52_fillbuff(sh_audio_t *sh_audio);

static ad_info_t info = 
{
	"AC3 through SPDIF",
	"hwac3",
	"Nick Kurshev",
	""
};

LIBAD_EXTERN(hwac3)

static sample_t * (*a52_init_ptr) (uint32_t mm_accel);
#define a52_init(a) (*a52_init_ptr)(a)

static int (*ac3_iec958_build_burst_ptr)(int length, int data_type, int big_endian, unsigned char * data, unsigned char * out);
#define ac3_iec958_build_burst(a,b,c,d,e) (*ac3_iec958_build_burst_ptr)(a,b,c,d,e)

static void *dll_handle;
static int load_dll(const char *libname)
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  a52_init_ptr = dlsym(dll_handle,"a52_init");
  return a52_init_ptr?1:0;
}

static void *dll_handle2;
static int load2_dll(const char *libname)
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  ac3_iec958_build_burst_ptr = dlsym(dll_handle,"ac3_iec958_build_burst");
  return ac3_iec958_build_burst_ptr?1:0;
}

int preinit(sh_audio_t *sh)
{
  /* Dolby AC3 audio: */
  sh->audio_out_minsize=4*256*6;
  sh->audio_in_minsize=3840;
  sh->channels=2;
  sh->sample_format=AFMT_AC3;
  return load_dll(codec_name("liba52.so")) && load2_dll(codec_name("libac3-iec958.so"));
}

int init(sh_audio_t *sh_audio)
{
  /* Dolby AC3 passthrough:*/
  a52_samples=a52_init (a52_accel); /* doesn't require mmx optimzation */
  if (a52_samples == NULL) {
       MSG_ERR("A52 init failed\n");
       return 0;
  }
  if(a52_fillbuff(sh_audio)<0) {
       MSG_ERR("A52 sync failed\n");
       return 0;
  }
 /* 
  sh_audio->samplerate=ai.samplerate;   // SET by a52_fillbuff()
  sh_audio->samplesize=ai.framesize;
  sh_audio->i_bps=ai.bitrate*(1000/8);  // SET by a52_fillbuff()
  sh_audio->ac3_frame=malloc(6144);
  sh_audio->o_bps=sh_audio->i_bps;  // XXX FIXME!!! XXX

   o_bps is calculated from samplesize*channels*samplerate
   a single ac3 frame is always translated to 6144 byte packet. (zero padding)*/
  sh_audio->channels=2;
  sh_audio->samplesize=2;   /* 2*2*(6*256) = 6144 (very TRICKY!)*/
  return 1;
}

void uninit(sh_audio_t *sh)
{
  dlclose(dll_handle);
  dlclose(dll_handle2);
}

int control(sh_audio_t *sh,int cmd,void* arg, ...)
{
    switch(cmd)
    {
      case ADCTRL_RESYNC_STREAM:
          sh->a_in_buffer_len=0;   // reset ACM/DShow audio buffer
	  return CONTROL_TRUE;
      case ADCTRL_SKIP_FRAME:
	  a52_fillbuff(sh); break; // skip AC3 frame
	  return CONTROL_TRUE;
      default:
	  return CONTROL_UNKNOWN;
    }
  return CONTROL_UNKNOWN;
}

int decode_audio(sh_audio_t *sh_audio,unsigned char *buf,int minlen,int maxlen)
{
  int len=-1;
  if(!sh_audio->a_in_buffer_len)
    if((len=a52_fillbuff(sh_audio))<0) return len; /*EOF*/
  sh_audio->a_in_buffer_len=0;
  len = ac3_iec958_build_burst(len, 0x01, 1, sh_audio->a_in_buffer, buf);
  return len;
}
