#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include "config.h"
#include "ad_internal.h"
#include "../libao2/afmt.h"
#include "interface/adpcm.h"
#include "codecs_ld.h"

static ad_info_t info = 
{
	"IMA ADPCM audio decoder",
	"imaadpcm",
	"Nick Kurshev",
	"ima4 (MOV files)"
};

LIBAD_EXTERN(imaadpcm)

static int (*ima_adpcm_decode_block_ptr)(unsigned short *output, unsigned char *input,
  int channels);
#define ima_adpcm_decode_block(a,b,c) (*ima_adpcm_decode_block_ptr)(a,b,c)

static void *dll_handle;
static int load_dll(const char *libname)
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  ima_adpcm_decode_block_ptr = dlsym(dll_handle,"ima_adpcm_decode_block");
  return ima_adpcm_decode_block_ptr != NULL;
}

int init(sh_audio_t *sh_audio)
{
  /* IMA-ADPCM 4:1 audio codec:*/
  sh_audio->channels=sh_audio->wf->nChannels;
  sh_audio->samplerate=sh_audio->wf->nSamplesPerSec;
  /* decodes 34 byte -> 64 short*/
  sh_audio->i_bps=IMA_ADPCM_BLOCK_SIZE*(sh_audio->channels*sh_audio->samplerate)/
		  IMA_ADPCM_SAMPLES_PER_BLOCK;  /* 1:4 */
  return 1;
}

int preinit(sh_audio_t *sh_audio)
{
  sh_audio->audio_out_minsize=4096;
  sh_audio->ds->ss_div=IMA_ADPCM_SAMPLES_PER_BLOCK;
  sh_audio->audio_in_minsize=
  sh_audio->ds->ss_mul=IMA_ADPCM_BLOCK_SIZE * sh_audio->wf->nChannels;
  return load_dll(codec_name("libadpcm.so"));
}

void uninit(sh_audio_t *sh)
{
  dlclose(dll_handle);
}

int control(sh_audio_t *sh,int cmd,void* arg, ...)
{
  return CONTROL_UNKNOWN;
}

int decode_audio(sh_audio_t *sh_audio,unsigned char *buf,int minlen,int maxlen)
{
  int len=-1;
    unsigned char ibuf[IMA_ADPCM_BLOCK_SIZE * 2]; /* bytes / stereo frame */
    if (demux_read_data_r(sh_audio->ds, ibuf,
      IMA_ADPCM_BLOCK_SIZE * sh_audio->wf->nChannels) != 
      IMA_ADPCM_BLOCK_SIZE * sh_audio->wf->nChannels) 
	return len;
    len=2*ima_adpcm_decode_block((unsigned short*)buf,ibuf, sh_audio->wf->nChannels);
    return len;
}
