#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h> /* GLIBC specific. Exists under cygwin too! */
#include "ad_internal.h"
#include "interface/mp3.h"
#include "config.h"
#include "../cpudetect.h"
#include "../mm_accel.h"
#include "codecs_ld.h"

static ad_info_t info = 
{
	"MPEG layer-2, layer-3",
	"mp3lib",
	"Nick Kurshev",
	"Optimized to MMX/SSE/3Dnow!"
};

LIBAD_EXTERN(mp3)

extern unsigned mplayer_accel;
extern int mplayer_audio_read(char *buf,int size);
extern int fakemono;
static sh_audio_t* dec_audio_sh=NULL;

static void (*MP3_Init_ptr)(int fakemono,unsigned accel,int (*audio_read)(char *buf,int size));
#define MP3_Init(a,b,c) (*MP3_Init_ptr)(a,b,c)
static int (*MP3_DecodeFrame_ptr)(unsigned char *hova,short single);
#define MP3_DecodeFrame(a,b) (*MP3_DecodeFrame_ptr)(a,b)
static void (*MP3_PrintHeader_ptr)(void);
#define MP3_PrintHeader() (*MP3_PrintHeader_ptr)()

static int (*MP3_channels_ptr);
static int (*MP3_bitrate_ptr);
static int (*MP3_samplerate_ptr);


static void *dll_handle;
static int load_dll(const char *libname)
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  MP3_DecodeFrame_ptr = dlsym(dll_handle,"MP3_DecodeFrame");
  MP3_Init_ptr = dlsym(dll_handle,"MP3_Init");
  MP3_PrintHeader_ptr = dlsym(dll_handle,"MP3_PrintHeader");
  MP3_channels_ptr = dlsym(dll_handle,"MP3_channels");
  MP3_bitrate_ptr = dlsym(dll_handle,"MP3_bitrate");
  MP3_samplerate_ptr = dlsym(dll_handle,"MP3_samplerate");
  return MP3_DecodeFrame_ptr && MP3_Init_ptr && MP3_PrintHeader_ptr &&
	 MP3_channels_ptr && MP3_bitrate_ptr && MP3_samplerate_ptr;
}


int preinit(sh_audio_t *sh)
{
  sh->audio_out_minsize=4608;
  return load_dll(codec_name("libMP3.so"));
}

int init(sh_audio_t *sh)
{
  // MPEG Audio:
  dec_audio_sh=sh; // save sh_audio for the callback:
  MP3_Init(fakemono,mplayer_accel,&mplayer_audio_read);
  *MP3_samplerate_ptr=*MP3_channels_ptr=0;
  sh->a_buffer_len=MP3_DecodeFrame(sh->a_buffer,-1);
  sh->channels=2; // hack
  sh->samplerate=*MP3_samplerate_ptr;
  sh->i_bps=*MP3_bitrate_ptr*(1000/8);
  MP3_PrintHeader();
  return 1;
}

void uninit(sh_audio_t *sh)
{
  dlclose(dll_handle);
}

int control(sh_audio_t *sh,int cmd,void* arg, ...)
{
    switch(cmd)
    {
      case ADCTRL_RESYNC_STREAM:
          MP3_DecodeFrame(NULL,-2); // resync
          MP3_DecodeFrame(NULL,-2); // resync
          MP3_DecodeFrame(NULL,-2); // resync
	  return CONTROL_TRUE;
      case ADCTRL_SKIP_FRAME:
	  MP3_DecodeFrame(NULL,-2); // skip MPEG frame
	  return CONTROL_TRUE;
      default:
	  return CONTROL_UNKNOWN;
    }
  return CONTROL_UNKNOWN;
}

int decode_audio(sh_audio_t *sh_audio,unsigned char *buf,int minlen,int maxlen)
{
   return MP3_DecodeFrame(buf,-1);
}
