#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include "config.h"
#include "ad_internal.h"
#include "../libao2/afmt.h"
#include "interface/adpcm.h"
#include "codecs_ld.h"

static ad_info_t info = 
{
	"MS ADPCM audio decoder",
	"msadpcm",
	"Nick Kurshev",
	""
};

LIBAD_EXTERN(msadpcm)

static int (*ms_adpcm_decode_block_ptr)(unsigned short *output, unsigned char *input,
  int channels, int block_size);
#define ms_adpcm_decode_block(a,b,c,d) (*ms_adpcm_decode_block_ptr)(a,b,c,d)

static void *dll_handle;
static int load_dll(const char *libname)
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  ms_adpcm_decode_block_ptr = dlsym(dll_handle,"ms_adpcm_decode_block");
  return ms_adpcm_decode_block_ptr != NULL;
}

int init(sh_audio_t *sh_audio)
{
  sh_audio->channels=sh_audio->wf->nChannels;
  sh_audio->samplerate=sh_audio->wf->nSamplesPerSec;
  sh_audio->i_bps = sh_audio->wf->nBlockAlign *
    (sh_audio->channels*sh_audio->samplerate) / MS_ADPCM_SAMPLES_PER_BLOCK;
  return 1;
}

int preinit(sh_audio_t *sh_audio)
{
  sh_audio->audio_out_minsize=sh_audio->wf->nBlockAlign * 8;
  sh_audio->ds->ss_div = MS_ADPCM_SAMPLES_PER_BLOCK;
  sh_audio->audio_in_minsize =
  sh_audio->ds->ss_mul = sh_audio->wf->nBlockAlign;
  return load_dll(codec_name("libadpcm.so"));
}

void uninit(sh_audio_t *sh)
{
  dlclose(dll_handle);
}

int control(sh_audio_t *sh,int cmd,void* arg, ...)
{
  return CONTROL_UNKNOWN;
}

int decode_audio(sh_audio_t *sh_audio,unsigned char *buf,int minlen,int maxlen)
{
  int len=-1;
  static unsigned char *ibuf = NULL;
  if (!ibuf)
   ibuf = (unsigned char *)malloc
        (sh_audio->wf->nBlockAlign * sh_audio->wf->nChannels);
  if (demux_read_data_r(sh_audio->ds, ibuf,
      sh_audio->wf->nBlockAlign) != 
      sh_audio->wf->nBlockAlign) 
         return len; /* EOF */
  len= 2 * ms_adpcm_decode_block(
          (unsigned short*)buf,ibuf, sh_audio->wf->nChannels,
          sh_audio->wf->nBlockAlign);
  return len;
}
