#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include "config.h"
#include "ad_internal.h"
#include "../libao2/afmt.h"
#include "interface/roqav.h"
#include "interface/adpcm.h"
#include "codecs_ld.h"

static ad_info_t info = 
{
	"Id RoQ File Audio Decoder",
	"roqaudio",
	"Nick Kurshev",
	"RoQA is an internal MPlayer FOURCC"
};

LIBAD_EXTERN(roqaudio)

void *(*roq_decode_audio_init_ptr)(void);
#define roq_decode_audio_init() (*roq_decode_audio_init_ptr)()

int (*roq_decode_audio_ptr)(unsigned short *output, unsigned char *input,
  int encoded_size, int channels, void *context);
#define roq_decode_audio(a,b,c,d,e) (*roq_decode_audio_ptr)(a,b,c,d,e)

static void *dll_handle;
static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  roq_decode_audio_init_ptr = dlsym(dll_handle,"roq_decode_audio_init");
  roq_decode_audio_ptr = dlsym(dll_handle,"roq_decode_audio");
  return roq_decode_audio_init_ptr && roq_decode_audio_ptr;
}

int init(sh_audio_t *sh_audio)
{
  sh_audio->channels=sh_audio->wf->nChannels;
  sh_audio->samplerate=sh_audio->wf->nSamplesPerSec;
  sh_audio->i_bps = 44100;
  return 1;
}

int preinit(sh_audio_t *sh_audio)
{
  /* minsize was stored in wf->nBlockAlign by the RoQ demuxer */
  if(!load_lib(codec_name("libroqav.so"))) return 0;
  sh_audio->audio_out_minsize=sh_audio->wf->nBlockAlign;
  sh_audio->audio_in_minsize=sh_audio->audio_out_minsize / 2; /* FIXME? */
  sh_audio->ds->ss_div=DK3_ADPCM_SAMPLES_PER_BLOCK;
  sh_audio->ds->ss_mul=DK3_ADPCM_BLOCK_SIZE;
  sh_audio->context = roq_decode_audio_init();
  return 1;
}

void uninit(sh_audio_t *sh)
{
  dlclose(dll_handle);
}

int control(sh_audio_t *sh,int cmd,void* arg, ...)
{
  return CONTROL_UNKNOWN;
}

int decode_audio(sh_audio_t *sh_audio,unsigned char *buf,int minlen,int maxlen)
{
  int len=-1;
  static unsigned char *ibuf = NULL;
  unsigned char header_data[6];
  int read_len;

  if (!ibuf) ibuf = (unsigned char *)malloc(sh_audio->audio_out_minsize / 2);
  /* figure out how much data to read */
  if (demux_read_data_r(sh_audio->ds, header_data, 6) != 6) return len; /* EOF */
  read_len = (header_data[5] << 24) | (header_data[4] << 16) |
	     (header_data[3] << 8) | header_data[2];
  read_len += 2;  /* 16-bit arguments */
  if (demux_read_data_r(sh_audio->ds, ibuf, read_len) != read_len) return len;
  len = 2 * roq_decode_audio((unsigned short *)buf, ibuf,
			     read_len, sh_audio->channels, sh_audio->context);          
  return len;
}
