#include "config.h"

#include <stdio.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <stdlib.h>
#include <unistd.h>

#include "mp_msg.h"
#include "help_mp.h"

#include "osdep/timer.h"
#include "osdep/shmem.h"

extern int verbose; // defined in mplayer.c

#include "stream.h"
#include "demuxer.h"
#include "parse_es.h"

#include "codec-cfg.h"

#include "libvo/video_out.h"

#include "stheader.h"
#include "vd.h"

#include "dec_video.h"

// ===================================================================

extern int benchmark;
extern double video_time_usage;
extern double vout_time_usage;
extern double max_video_time_usage;
extern double cur_video_time_usage;
extern double min_video_time_usage;
extern double max_vout_time_usage;
extern double cur_vout_time_usage;
extern double min_vout_time_usage;
extern vo_vaa_t vo_vaa;

#include "postproc/postprocess.h"

#include "cpudetect.h"

extern int v_hw_equ_cap;
extern int v_bright;
extern int v_cont;
extern int v_hue;
extern int v_saturation;
static vo_vaa_t vo_vaa;
void init_video_vaa( void )
{
    if(vo_query_vaa(&vo_vaa)==VO_NOTIMPL)
	memset(&vo_vaa,0,sizeof(vo_vaa_t));
   /*
	get_hw_eq
   */
   if(vo_vaa.get_video_eq)
   {
	vidix_video_eq_t veq;
	if(vo_vaa.get_video_eq(&veq) == 0)
	{
	    v_hw_equ_cap = veq.cap;
	    if(veq.cap & VEQ_CAP_BRIGHTNESS) v_bright = veq.brightness/10;
	    if(veq.cap & VEQ_CAP_CONTRAST) v_cont = veq.contrast/10;
	    if(veq.cap & VEQ_CAP_HUE) v_hue = veq.hue/10;
	    if(veq.cap & VEQ_CAP_SATURATION) v_saturation=veq.saturation/10;
	    /*
	    v_red_intensity=veq.red_intensity/10;
	    v_green_intensity=veq.green_intensity/10;
	    v_blue_intensity=veq.blue_intensity/10;
	    */
	}
   }
}


int divx_quality=0;

vd_functions_t* mpvdec=NULL;

int get_video_quality_max(sh_video_t *sh_video){
  if(mpvdec){
    int ret=mpvdec->control(sh_video,VDCTRL_QUERY_MAX_PP_LEVEL,NULL);
    if(ret>=0) return ret;
  }
 return 0;
}

void set_video_quality(sh_video_t *sh_video,int quality){
  if(mpvdec)
    mpvdec->control(sh_video,VDCTRL_SET_PP_LEVEL, (void*)(&quality));
}

int set_video_colors(sh_video_t *sh_video,char *item,int value)
{
    if(vo_vaa.get_video_eq)
    {
	vidix_video_eq_t veq;
	if(vo_vaa.get_video_eq(&veq) == 0)
	{
	    int v_hw_equ_cap = veq.cap;
	    if(v_hw_equ_cap != 0)
	    {
		if(vo_vaa.set_video_eq)
		{
		    vidix_video_eq_t veq;
		    veq.flags = VEQ_FLG_ITU_R_BT_601; /* Fixme please !!! */
		    if(strcmp(item,"Brightness") == 0)
		    {
			if(!(v_hw_equ_cap & VEQ_CAP_BRIGHTNESS)) goto try_sw_control;
			veq.brightness = value*10;
			veq.cap = VEQ_CAP_BRIGHTNESS;
		    }
		    else
		    if(strcmp(item,"Contrast") == 0)
		    {
			if(!(v_hw_equ_cap & VEQ_CAP_CONTRAST)) goto try_sw_control;
			veq.contrast = value*10;
			veq.cap = VEQ_CAP_CONTRAST;
		    }
		    else
		    if(strcmp(item,"Saturation") == 0)
		    {
			if(!(v_hw_equ_cap & VEQ_CAP_SATURATION)) goto try_sw_control;
			veq.saturation = value*10;
			veq.cap = VEQ_CAP_SATURATION;
		    }
		    else
		    if(strcmp(item,"Hue") == 0)
		    {
			if(!(v_hw_equ_cap & VEQ_CAP_HUE)) goto try_sw_control;
			veq.hue = value*10;
			veq.cap = VEQ_CAP_HUE;
		    }
		    else goto try_sw_control;;
		    vo_vaa.set_video_eq(&veq);
		}
		return 1;
	    }
	}
    }
    try_sw_control:
    if(mpvdec) return mpvdec->control(sh_video,VDCTRL_SET_EQUALIZER,item,(int)value)==CONTROL_OK?1:0;
    return 0;
}

void uninit_video(sh_video_t *sh_video){
    if(!sh_video->inited) return;
    mp_msg(MSGT_DECVIDEO,MSGL_V,"uninit video: %d  \n",sh_video->codec->driver_name);
    mpvdec->uninit(sh_video);
    sh_video->inited=0;
}

int init_video(sh_video_t *sh_video,const char* codecname,const char * vfm,int status){
    sh_video->codec=NULL;
    while((sh_video->codec=find_codec(sh_video->format,
      sh_video->bih?((unsigned int*) &sh_video->bih->biCompression):NULL,
      sh_video->codec,0) )){
	// ok we found one codec
	int i;
	if(sh_video->codec->flags&CODECS_FLAG_SELECTED) continue; // already tried & failed
	if(codecname && strcmp(sh_video->codec->codec_name,codecname)) continue; // -vc
	if(vfm && strcmp(sh_video->codec->driver_name,vfm)!=0) continue; // vfm doesn't match
	if(sh_video->codec->status<status) continue; // too unstable
	sh_video->codec->flags|=CODECS_FLAG_SELECTED; // tagging it
	// ok, it matches all rules, let's find the driver!
	for (i=0; mpcodecs_vd_drivers[i] != NULL; i++)
	    if(strcmp(mpcodecs_vd_drivers[i]->info->driver_name,sh_video->codec->driver_name)==0) break;
	mpvdec=mpcodecs_vd_drivers[i];
	if(!mpvdec) continue;
	// it's available, let's try to init!
	mp_msg(MSGT_DECVIDEO,MSGL_INFO,"Opening video decoder: [%s] %s\n",mpvdec->info->codec_name,mpvdec->info->driver_name);
	if(!mpvdec->init(sh_video)){
	    mp_msg(MSGT_DECVIDEO,MSGL_INFO,"VDecoder init failed :(\n");
	    continue; // try next...
	}
	// Yeah! We got it!
	sh_video->inited=1;
	return 1;
    }
    return 0;
}


int decode_video(sh_video_t *sh_video,unsigned char *start,int in_size,int drop_frame){
mp_image_t *mpi=NULL;
unsigned int t;
unsigned int t2;
double tt;

t=GetTimer();
mpi=mpvdec->decode(sh_video, start, in_size, drop_frame);
/* ------------------------ frame decoded. -------------------- */

#ifdef ARCH_X86
// some codecs is broken, and doesn't restore MMX state :(
// it happens usually with broken/damaged files.
if(gCpuCaps.has3DNow){
	__asm __volatile ("femms\n\t":::"memory");
}
else if(gCpuCaps.hasMMX){
	__asm __volatile ("emms\n\t":::"memory");
}
#endif

if(!mpi) return 0; // error / skipped frame

t2=GetTimer();t=t2-t;
tt = t*0.000001f;
video_time_usage+=tt;
if(benchmark)
{
    if(tt > max_video_time_usage) max_video_time_usage=tt;
    if(tt < min_video_time_usage) min_video_time_usage=tt;
    cur_video_time_usage=tt;
}

if(drop_frame) return 0;

vo_flush_frame();

if(!(mpi->flags&(MP_IMGFLAG_DIRECT|MP_IMGFLAG_DRAW_CALLBACK))){
    // blit frame:
    if(mpi->flags&MP_IMGFLAG_PLANAR)
        vo_draw_slice(mpi->planes,mpi->stride,sh_video->disp_w,sh_video->disp_h,0,0);
    else
        vo_draw_frame(mpi->planes);
}
    t2=GetTimer()-t2;
    tt=t2*0.000001f;
    vout_time_usage+=tt;
    if(benchmark)
    {
	if(tt > max_vout_time_usage) max_vout_time_usage = tt;
	if(tt < min_vout_time_usage) min_vout_time_usage = tt;
	cur_vout_time_usage=tt;
    }

  return 1;
}


