#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include "config.h"

#include "vd_internal.h"
#include "codecs_ld.h"

static vd_info_t info = {
	"Cinepak Video decoder",
	"cinepak",
	"A'rpi",
	"Dr. Tim Ferguson, http://www.csse.monash.edu.au/~timf/videocodec.html",
	"native codec"
};

LIBVD_EXTERN(cinepak)

static void *(*decode_cinepak_init_ptr)(void);
#define decode_cinepak_init() (*decode_cinepak_init_ptr)()
static void (*decode_cinepak_ptr)(void *context, unsigned char *buf, int size, mp_image_t *mpi);
#define decode_cinepak(a,b,c,d) (*decode_cinepak_ptr)(a,b,c,d)


static void *dll_handle;
static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  decode_cinepak_init_ptr = dlsym(dll_handle,"decode_cinepak_init");
  decode_cinepak_ptr = dlsym(dll_handle,"decode_cinepak");
  return decode_cinepak_ptr && decode_cinepak_init_ptr;
}

// to set/get/query special features/parameters
static int control(sh_video_t *sh,int cmd,void* arg,...){
    return CONTROL_UNKNOWN;
}

static int init(sh_video_t *sh){
    if(!load_lib(codec_name("libcinepak.so"))) return 0;
    sh->context = decode_cinepak_init();
    return mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_YUY2);
}

// uninit driver
static void uninit(sh_video_t *sh){
  dlclose(dll_handle);
}

// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    mp_image_t* mpi;
    if(len<=0) return NULL; // skipped frame
    
    mpi=mpcodecs_get_image(sh, MP_IMGTYPE_STATIC, MP_IMGFLAG_PRESERVE | MP_IMGFLAG_ACCEPT_STRIDE, 
	(sh->disp_w+3)&(~3),
	(sh->disp_h+3)&(~3));
    
    if(!mpi){	// temporary!
	MSG_ERR("couldn't allocate image for cinepak codec\n");
	return NULL;
    }

#if 0    
    MSG_V("mpi: %p/%d %p/%d %p/%d (%d) (%d)  \n",
	mpi->planes[0], mpi->stride[0],
	mpi->planes[1], mpi->stride[1],
	mpi->planes[2], mpi->stride[2],
	mpi->planes[1]-mpi->planes[0],
	mpi->planes[2]-mpi->planes[1]);
#endif

//    decode_cinepak(sh->context, data, len, mpi->planes[0], sh->disp_w, sh->disp_h,
//	(mpi->flags&MP_IMGFLAG_YUV)?16:(mpi->imgfmt&255), mpi->stride[0]);

    decode_cinepak(sh->context, data, len, mpi);
    
    return mpi;
}

