#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include "config.h"

#include "vd_internal.h"
#include "codecs_ld.h"

static vd_info_t info = {
	"Creative YUV decoder",
	"cyuv",
	"A'rpi",
	"Dr. Tim Ferguson",
	"native codec"
};

LIBVD_EXTERN(cyuv)

static void (*decode_cyuv_ptr)(
  unsigned char *buf,
  int size,
  unsigned char *frame,
  int width,
  int height,
  int format);
#define decode_cyuv(a,b,c,d,e,f) (*decode_cyuv_ptr)(a,b,c,d,e,f)

static void *dll_handle;
static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  decode_cyuv_ptr = dlsym(dll_handle,"decode_cyuv");
  return decode_cyuv_ptr != NULL;
}

// to set/get/query special features/parameters
static int control(sh_video_t *sh,int cmd,void* arg,...){
    return CONTROL_UNKNOWN;
}

// init driver
static int init(sh_video_t *sh){
    if(!load_lib(codec_name("libcyuv.so"))) return 0;
    return mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_UYVY);
}

// uninit driver
static void uninit(sh_video_t *sh){
  dlclose(dll_handle);
}

// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    mp_image_t* mpi;
    if(len<=0) return NULL; // skipped frame
    
    mpi=mpcodecs_get_image(sh, MP_IMGTYPE_TEMP, 0, 
	sh->disp_w, sh->disp_h);
    if(!mpi) return NULL;

    decode_cyuv(data, len, mpi->planes[0], sh->disp_w, sh->disp_h, sh->codec->outfmt[sh->outfmtidx]);

    return mpi;
}
