#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include "config.h"

#include "vd_internal.h"
#include "codecs_ld.h"

static vd_info_t info = {
	"Autodesk FLI/FLC Animation decoder",
	"fli",
	"A'rpi",
	"Mike Melanson",
	"native codec"
};

LIBVD_EXTERN(fli)

static void *(*init_fli_decoder_ptr)(int width, int height);
#define init_fli_decoder(a,b) (*init_fli_decoder_ptr)(a,b)

static void (*decode_fli_frame_ptr)(
  unsigned char *encoded,
  int encoded_size,
  unsigned char *decoded,
  int width,
  int height,
  int bytes_per_pixel,
  void *context);
#define decode_fli_frame(a,b,c,d,e,f,g) (*decode_fli_frame_ptr)(a,b,c,d,e,f,g)

static void *dll_handle;
static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  init_fli_decoder_ptr = dlsym(dll_handle,"init_fli_decoder");
  decode_fli_frame_ptr = dlsym(dll_handle,"decode_fli_frame");
  return init_fli_decoder_ptr && decode_fli_frame_ptr;
}

// to set/get/query special features/parameters
static int control(sh_video_t *sh,int cmd,void* arg,...){
    return CONTROL_UNKNOWN;
}

// init driver
static int init(sh_video_t *sh){
    if(!load_lib(codec_name("libfli.so"))) return 0;
    if(!mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_BGR24)) return 0;
    sh->context = init_fli_decoder(sh->disp_w, sh->disp_h);
    return 1;
}

// uninit driver
static void uninit(sh_video_t *sh){
  dlclose(dll_handle);
}

// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    mp_image_t* mpi;
    if(len<=0) return NULL; // skipped frame
    
    mpi=mpcodecs_get_image(sh, MP_IMGTYPE_STATIC, MP_IMGFLAG_PRESERVE, 
	sh->disp_w, sh->disp_h);
    if(!mpi) return NULL;

    decode_fli_frame(
        data, len, mpi->planes[0],
        sh->disp_w, sh->disp_h,
        mpi->bpp/8,
        sh->context);
    
    return mpi;
}
