#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h> /* GLIBC specific. Exists under cygwin too! */

#include "config.h"

#include "vd_internal.h"
#include "../cpudetect.h"
#include "../mm_accel.h"
#include "../postproc/postprocess.h"
#include "codecs_ld.h"

static vd_info_t info = 
{
	"MPEG 1/2 Video decoder",
	"libmpeg2",
	"A'rpi",
	"Aaron & Walken",
	"native"
};

LIBVD_EXTERN(libmpeg2)

#include "libmpdemux/parse_es.h"

#include "libvo/video_out.h"
#include "interface/mpeg2.h"
#include "interface/mpeg2_internal.h"

extern unsigned mplayer_accel;
static picture_t **picture_ptr;	// exported from libmpeg2/decode.c

static void (*mpeg2_init_ptr) (unsigned);
#define mpeg2_init(a) (*mpeg2_init_ptr)(a)
static int (*mpeg2_decode_data_ptr) (vo_draw_slice_t,postprocess_t,uint8_t * data_start, uint8_t * data_end,int framedrop);
#define mpeg2_decode_data(a,b,c,d,e) (*mpeg2_decode_data_ptr)(a,b,c,d,e)
static void (*mpeg2_allocate_image_buffers_ptr)(picture_t * picture);
#define mpeg2_allocate_image_buffers(a) (*mpeg2_allocate_image_buffers_ptr)(a)
static void (*mpeg2_free_image_buffers_ptr) (picture_t * picture);
#define mpeg2_free_image_buffers(a) (*mpeg2_free_image_buffers_ptr)(a)

static void *dll_handle;
static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  mpeg2_init_ptr = dlsym(dll_handle,"mpeg2_init");
  picture_ptr = dlsym(dll_handle,"picture");
  mpeg2_decode_data_ptr = dlsym(dll_handle,"mpeg2_decode_data");
  mpeg2_free_image_buffers_ptr = dlsym(dll_handle,"mpeg2_free_image_buffers");
  mpeg2_allocate_image_buffers_ptr = dlsym(dll_handle,"mpeg2_allocate_image_buffers");
  return mpeg2_init_ptr && mpeg2_decode_data_ptr && mpeg2_allocate_image_buffers_ptr &&
	 mpeg2_free_image_buffers_ptr && picture_ptr;
}


// to set/get/query special features/parameters
static int control(sh_video_t *sh,int cmd,void* arg,...){
    switch(cmd){
    case VDCTRL_QUERY_MAX_PP_LEVEL:
	return GET_PP_QUALITY_MAX;
    case VDCTRL_SET_PP_LEVEL: {
	int quality=*((int*)arg);
	if(quality<0 || quality>GET_PP_QUALITY_MAX) quality=GET_PP_QUALITY_MAX;
	divx_quality = quality;
	(*picture_ptr)->pp_options=divx_quality;
	return CONTROL_OK;
    }
    }
    return CONTROL_UNKNOWN;
}

// init driver
static int init(sh_video_t *sh){
    if(!load_lib(codec_name("libmpeg2.so"))) return 0;
    mpeg2_init(mplayer_accel);
    (*picture_ptr)->pp_options=divx_quality;
    // send seq header to the decoder:  *** HACK ***
    mpeg2_decode_data(NULL,NULL,videobuffer,videobuffer+videobuf_len,0);
    mpeg2_allocate_image_buffers (*picture_ptr);
    return mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_YV12);
}

// uninit driver
static void uninit(sh_video_t *sh){
    mpeg2_free_image_buffers (*picture_ptr);
    dlclose(dll_handle);
}

// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    mp_image_t* mpi=mpcodecs_get_image(sh, MP_IMGTYPE_EXPORT, MP_IMGFLAG_DRAW_CALLBACK,
	sh->disp_w, sh->disp_h);
    mpeg2_decode_data(vo_draw_slice, postprocess, data, data+len,flags&3); // decode
    return mpi;
}

