#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include "config.h"

#include "vd_internal.h"
#include "codecs_ld.h"

static vd_info_t info = {
	"Microsoft Video 1 / CRAM decoder",
	"msvidc",
	"A'rpi",
	"Mike Melanson",
	"native codec"
};

LIBVD_EXTERN(msvidc)

static void (*AVI_Decode_Video1_16_ptr)(
  char *encoded,
  int encoded_size,
  char *decoded,
  int width,
  int height,
  int bytes_per_pixel);
#define AVI_Decode_Video1_16(a,b,c,d,e,f) (*AVI_Decode_Video1_16_ptr)(a,b,c,d,e,f)

static void (*AVI_Decode_Video1_8_ptr)(
  char *encoded,
  int encoded_size,
  char *decoded,
  int width,
  int height,
  unsigned char *palette_map,
  int bytes_per_pixel);
#define AVI_Decode_Video1_8(a,b,c,d,e,f,g) (*AVI_Decode_Video1_8_ptr)(a,b,c,d,e,f,g)

static void *dll_handle;
static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  AVI_Decode_Video1_8_ptr = dlsym(dll_handle,"AVI_Decode_Video1_8");
  AVI_Decode_Video1_16_ptr = dlsym(dll_handle,"AVI_Decode_Video1_16");
  return AVI_Decode_Video1_16_ptr && AVI_Decode_Video1_8_ptr;
}

// to set/get/query special features/parameters
static int control(sh_video_t *sh,int cmd,void* arg,...){
    return CONTROL_UNKNOWN;
}

// init driver
static int init(sh_video_t *sh){
    if(!load_lib(codec_name("libmsvidc.so"))) return 0;
    return mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_BGR24);
}

// uninit driver
static void uninit(sh_video_t *sh){
  dlclose(dll_handle);
}

// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    mp_image_t* mpi;
    if(len<=0) return NULL; // skipped frame
    
    mpi=mpcodecs_get_image(sh, MP_IMGTYPE_STATIC, MP_IMGFLAG_PRESERVE, 
	sh->disp_w, sh->disp_h);
    if(!mpi) return NULL;
    
    if (sh->bih->biBitCount == 16)
      AVI_Decode_Video1_16(
        data, len, mpi->planes[0],
        sh->disp_w, sh->disp_h,
        mpi->bpp/8);
    else
      AVI_Decode_Video1_8(
        data, len, mpi->planes[0],
        sh->disp_w, sh->disp_h,
	(unsigned char *)sh->bih+40,
        mpi->bpp/8);
    
    return mpi;
}
