#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include "config.h"

#include "vd_internal.h"
#include "codecs_ld.h"

static vd_info_t info = {
	"Quicktime Animation (RLE) decoder",
	"qtrle",
	"A'rpi",
	"Mike Melanson",
	"native codec"
};

LIBVD_EXTERN(qtrle)

static void (*qt_decode_rle_ptr)(
  unsigned char *encoded,
  int encoded_size,
  unsigned char *decoded,
  int width,
  int height,
  int encoded_bpp,
  int bytes_per_pixel);

#define qt_decode_rle(a,b,c,d,e,f,g) (*qt_decode_rle_ptr)(a,b,c,d,e,f,g)

static void *dll_handle;
static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  qt_decode_rle_ptr = dlsym(dll_handle,"qt_decode_rle");
  return qt_decode_rle_ptr != NULL;
}

// to set/get/query special features/parameters
static int control(sh_video_t *sh,int cmd,void* arg,...){
    return CONTROL_UNKNOWN;
}

// init driver
static int init(sh_video_t *sh){
  if(!load_lib(codec_name("libqtrle.so"))) return 0;
  if (sh->bih->biBitCount != 24){
    MSG_ERR(
    "    *** FYI: This Quicktime file is using %d-bit RLE Animation\n" \
    "    encoding, which is not yet supported by MPlayer. But if you upload\n" \
    "    this Quicktime file to the MPlayer FTP, the team could look at it.\n",
    sh->bih->biBitCount);
    return 0;
  }
    
    return mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_BGR24);
}

// uninit driver
static void uninit(sh_video_t *sh){
  dlclose(dll_handle);
}

// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    mp_image_t* mpi;
    if(len<=0) return NULL; // skipped frame
    
    mpi=mpcodecs_get_image(sh, MP_IMGTYPE_STATIC, MP_IMGFLAG_PRESERVE, 
	sh->disp_w, sh->disp_h);
    if(!mpi) return NULL;

    qt_decode_rle(
        data,len, mpi->planes[0],
        sh->disp_w, sh->disp_h,
        sh->bih->biBitCount,
        mpi->bpp/8);
    
    return mpi;
}
