#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include "config.h"

#include "vd_internal.h"
#include "codecs_ld.h"

static vd_info_t info = {
	"Apple Graphics (SMC) decoder",
	"qtsmc",
	"A'rpi",
	"Mike Melanson",
	"native codec"
};

LIBVD_EXTERN(qtsmc)

static int (*qt_init_decode_smc_ptr)(void);
#define qt_init_decode_smc() (*qt_init_decode_smc_ptr)()

static void (*qt_decode_smc_ptr)(
  unsigned char *encoded,
  int encoded_size,
  unsigned char *decoded,
  int width,
  int height,
  unsigned char *palette_map,
  int bytes_per_pixel);
#define qt_decode_smc(a,b,c,d,e,f,g) (*qt_decode_smc_ptr)(a,b,c,d,e,f,g)

static void *dll_handle;
static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  qt_init_decode_smc_ptr = dlsym(dll_handle,"qt_init_decode_smc");
  qt_decode_smc_ptr = dlsym(dll_handle,"qt_decode_smc");
  return qt_init_decode_smc_ptr && qt_decode_smc_ptr;
}

// to set/get/query special features/parameters
static int control(sh_video_t *sh,int cmd,void* arg,...){
    return CONTROL_UNKNOWN;
}

// init driver
static int init(sh_video_t *sh){
    if(!load_lib(codec_name("libqtsmc.so"))) return 0;
    if (qt_init_decode_smc() != 0){
	MSG_ERR( "SMC decoder could not allocate enough memory");
	return 0;
    }
    
    return mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_BGR24);
}

// uninit driver
static void uninit(sh_video_t *sh){
  dlclose(dll_handle);
}

// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    mp_image_t* mpi;
    if(len<=0) return NULL; // skipped frame
    
    mpi=mpcodecs_get_image(sh, MP_IMGTYPE_STATIC, MP_IMGFLAG_PRESERVE, 
	sh->disp_w, sh->disp_h);
    if(!mpi) return NULL;

    qt_decode_smc(
        data,len, mpi->planes[0],
        sh->disp_w, sh->disp_h,
        (unsigned char *)sh->bih+40,
        mpi->bpp/8);
    
    return mpi;
}
