#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include "config.h"

#include "vd_internal.h"
#include "codecs_ld.h"

static vd_info_t info = {
	"Id RoQ File Video decoder",
	"roqvideo",
	"A'rpi",
	"Mike Melanson",
	"native codec"
};

LIBVD_EXTERN(roqvideo)

#include "interface/roqav.h"

static void *(*roq_decode_video_init_ptr)(void);
#define roq_decode_video_init() (*roq_decode_video_init_ptr)()
static void (*roq_decode_video_ptr)(void *context, unsigned char *encoded, 
  int encoded_size, mp_image_t *mpi);
#define roq_decode_video(a,b,c,d) (*roq_decode_video_ptr)(a,b,c,d)

static void *dll_handle;
static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  roq_decode_video_init_ptr = dlsym(dll_handle,"roq_decode_video_init");
  roq_decode_video_ptr = dlsym(dll_handle,"roq_decode_video");
  return roq_decode_video_ptr && roq_decode_video_init_ptr;
}

// to set/get/query special features/parameters
static int control(sh_video_t *sh,int cmd,void* arg,...){
    return CONTROL_UNKNOWN;
}

// init driver
static int init(sh_video_t *sh){
    if(!load_lib(codec_name("libroqav.so"))) return 0;
    sh->context = roq_decode_video_init();
    return mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_YV12);
}

// uninit driver
static void uninit(sh_video_t *sh){
  dlclose(dll_handle);
}

// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    mp_image_t* mpi;
    if(len<=0) return NULL; // skipped frame
    
    mpi=mpcodecs_get_image(sh, MP_IMGTYPE_IP, MP_IMGFLAG_PRESERVE | MP_IMGFLAG_READABLE, 
	sh->disp_w, sh->disp_h);
    if(!mpi) return NULL;

    roq_decode_video(sh->context, data, len, mpi);

    return mpi;
}
