#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <dlfcn.h> /* GLIBC specific. Exists under cygwin too! */

#include "help_mp.h"

#include "vd_internal.h"
#include "../libmpdemux/wine/vfw.h"
#include "../libmpdemux/wine/driver.h"

#include "codecs_ld.h"

static vd_info_t info_vfw = {
	"Win32/VfW video codecs",
	"vfw",
	"A'rpi",
	"based on http://avifile.sf.net",
	"win32 codecs"
};

static vd_info_t info_vfwex = {
	"Win32/VfWex video codecs",
	"vfwex",
	"A'rpi",
	"based on http://avifile.sf.net",
	"win32 codecs"
};

#define info info_vfw
LIBVD_EXTERN(vfw)
#undef info

#define info info_vfwex
LIBVD_EXTERN(vfwex)
#undef info

static HIC	VFWAPI	(*ICOpen_ptr)(long fccType, long fccHandler, UINT wMode);
static LRESULT	VFWAPI (*ICSendMessage_ptr)(HIC hic, unsigned int msg, long dw1, long dw2);
static long VFWAPIV (*ICDecompress_ptr)(HIC hic,long dwFlags,LPBITMAPINFOHEADER lpbiFormat,void* lpData,LPBITMAPINFOHEADER lpbi,void* lpBits);
static long VFWAPIV (*ICDecompressEx_ptr)(HIC hic,long dwFlags,LPBITMAPINFOHEADER lpbiFormat,void* lpData,LPBITMAPINFOHEADER lpbi,void* lpBits);
static long VFWAPIV (*ICUniversalEx_ptr)(HIC hic,int command,LPBITMAPINFOHEADER lpbiFormat,LPBITMAPINFOHEADER lpbi);
static LRESULT VFWAPI (*ICClose_ptr)(HIC hic);
static int vfw_ex=0;

#define ICSendMessage(a,b,c,d) (*ICSendMessage_ptr)(a,b,c,d)
#define ICUniversalEx(a,b,c,d) (*ICUniversalEx_ptr)(a,b,c,d)
#define ICClose(a) (*ICClose_ptr)(a)

static char **win32_codec_name;

static void *dll_handle;

static int load_lib( const char *libname )
{
  if(!(dll_handle=ld_codec(libname,NULL))) return 0;
  ICOpen_ptr = dlsym(dll_handle,"ICOpen");
  ICClose_ptr = dlsym(dll_handle,"ICClose");
  ICSendMessage_ptr = dlsym(dll_handle,"ICSendMessage");
  ICDecompress_ptr = dlsym(dll_handle,"ICDecompress");
  ICDecompressEx_ptr = dlsym(dll_handle,"ICDecompressEx");
  ICUniversalEx_ptr = dlsym(dll_handle,"ICUniversalEx");
  win32_codec_name = dlsym(dll_handle,"win32_codec_name");
  return ICOpen_ptr && ICSendMessage_ptr &&
	 ICDecompress_ptr && ICDecompressEx_ptr &&
	 ICUniversalEx_ptr && win32_codec_name;
}

int init_vfw_video_codec(sh_video_t *sh_video,int ex){
  HRESULT ret;
  int yuv=0;
  unsigned int outfmt=sh_video->codec->outfmt[sh_video->outfmtidx];
  char *temp;
  int temp_len;
  int i;

  MSG_V("======= Win32 (VFW) VIDEO Codec init =======\n");

  memset(&sh_video->o_bih, 0, sizeof(BITMAPINFOHEADER));
  sh_video->o_bih.biSize = sizeof(BITMAPINFOHEADER);

  *win32_codec_name = sh_video->codec->dll_name;
  sh_video->hic = (*ICOpen_ptr)( 0x63646976, sh_video->bih->biCompression, ICMODE_DECOMPRESS);
  if(!sh_video->hic){
    MSG_ERR("ICOpen failed! unknown codec / wrong parameters?\n");
    return 0;
  }

//  sh_video->bih->biBitCount=32;

  temp_len = ICDecompressGetFormatSize(sh_video->hic, sh_video->bih);

  if (temp_len < sh_video->o_bih.biSize)
    temp_len = sh_video->o_bih.biSize;

  temp = malloc(temp_len);
  MSG_V("ICDecompressGetFormatSize ret: %d\n", temp_len);

  // Note: DivX.DLL overwrites 4 bytes _AFTER_ the o_bih header, so it corrupts
  // the sh_video struct content. We call it with an 1024-byte temp space and
  // then copy out the data we need:
  memset(temp,0x77,temp_len);
//  memcpy(temp,sh_video->bih,sizeof(BITMAPINFOHEADER));
//  sh_video->o_bih.biSize = temp_len;

  ret = ICDecompressGetFormat(sh_video->hic, sh_video->bih, temp);
  if(ret < 0){
    MSG_ERR("ICDecompressGetFormat failed: Error %d\n", (int)ret);
    for (i=0; i < temp_len; i++) MSG_DBG2( "%02x ", temp[i]);
    return 0;
  }
  MSG_V("ICDecompressGetFormat OK\n");
  
  memcpy(&sh_video->o_bih,temp,sizeof(sh_video->o_bih));

  if (temp_len > sizeof(sh_video->o_bih))
  {
    MSG_V( "Extra info in o_bih (%d bytes)!\n",
	temp_len-sizeof(sh_video->o_bih));
    for(i=sizeof(sh_video->o_bih);i<temp_len;i++) MSG_DBG2( "%02X ",temp[i]);
  }
  free(temp);
#if 0
  // workaround for pegasus MJPEG:
  if(!sh_video->o_bih.biWidth) sh_video->o_bih.biWidth=sh_video->bih->biWidth;
  if(!sh_video->o_bih.biHeight) sh_video->o_bih.biHeight=sh_video->bih->biHeight;
  if(!sh_video->o_bih.biPlanes) sh_video->o_bih.biPlanes=sh_video->bih->biPlanes;
#endif

  switch (outfmt) {

/* planar format */
  case IMGFMT_YV12:
  case IMGFMT_I420:
  case IMGFMT_IYUV:
      sh_video->o_bih.biBitCount=12;
      yuv=1;
      break;

/* packed format */
  case IMGFMT_YUY2:
  case IMGFMT_UYVY:
  case IMGFMT_YVYU:
      sh_video->o_bih.biBitCount=16;
      yuv=1;
      break;

/* rgb/bgr format */
  case IMGFMT_RGB8:
  case IMGFMT_BGR8:
      sh_video->o_bih.biBitCount=8;
      break;

  case IMGFMT_RGB15:
  case IMGFMT_RGB16:
  case IMGFMT_BGR15:
  case IMGFMT_BGR16:
      sh_video->o_bih.biBitCount=16;
      break;

  case IMGFMT_RGB24:
  case IMGFMT_BGR24:
      sh_video->o_bih.biBitCount=24;
      break;

  case IMGFMT_RGB32:
  case IMGFMT_BGR32:
      sh_video->o_bih.biBitCount=32;
      break;

  default:
      MSG_ERR("unsupported image format: 0x%x\n", outfmt);
      return 0;
  }

  sh_video->o_bih.biSizeImage = sh_video->o_bih.biWidth * sh_video->o_bih.biHeight * (sh_video->o_bih.biBitCount/8);
  
  if(!(sh_video->codec->outflags[sh_video->outfmtidx]&CODECS_FLAG_FLIP)) {
      sh_video->o_bih.biHeight=-sh_video->bih->biHeight; // flip image!
  }

  if(yuv && !(sh_video->codec->outflags[sh_video->outfmtidx] & CODECS_FLAG_YUVHACK))
	 sh_video->o_bih.biCompression = outfmt;
  else
         sh_video->o_bih.biCompression = 0;

    MSG_V("Starting decompression, format:\n"
	"  biSize %d\n"
	"  biWidth %d\n"
	"  biHeight %d\n"
	"  biPlanes %d\n"
	"  biBitCount %d\n"
	"  biCompression 0x%x ('%.4s')\n"
	"  biSizeImage %d\n"
	"Dest fmt:\n"
	"  biSize %d\n"
	"  biWidth %d\n"
	"  biHeight %d\n"
	"  biPlanes %d\n"
	"  biBitCount %d\n"
	"  biCompression 0x%x ('%.4s')\n"
	"  biSizeImage %d\n"
	, sh_video->bih->biSize
	, sh_video->bih->biWidth
	, sh_video->bih->biHeight
	, sh_video->bih->biPlanes
	, sh_video->bih->biBitCount
	, sh_video->bih->biCompression, (char *)&sh_video->bih->biCompression
	, sh_video->bih->biSizeImage
	, sh_video->o_bih.biSize
	, sh_video->o_bih.biWidth
	, sh_video->o_bih.biHeight
	, sh_video->o_bih.biPlanes
	, sh_video->o_bih.biBitCount
	, sh_video->o_bih.biCompression, (char *)&sh_video->o_bih.biCompression
	, sh_video->o_bih.biSizeImage);

  ret = ex ?
      ICDecompressQueryEx(sh_video->hic, sh_video->bih, &sh_video->o_bih) :
      ICDecompressQuery(sh_video->hic, sh_video->bih, &sh_video->o_bih);
  if(ret){
    MSG_ERR("ICDecompressQuery failed: Error %d\n", (int)ret);
//    return 0;
  } else
  MSG_V("ICDecompressQuery OK\n");

  ret = ex ?
      ICDecompressBeginEx(sh_video->hic, sh_video->bih, &sh_video->o_bih) :
      ICDecompressBegin(sh_video->hic, sh_video->bih, &sh_video->o_bih);
  if(ret){
    MSG_ERR("ICDecompressBegin failed: Error %d\n", (int)ret);
//    return 0;
  }

  sh_video->our_out_buffer = (char*)memalign(64,sh_video->o_bih.biSizeImage);
  if(!sh_video->our_out_buffer){
    MSG_ERR(MSGTR_NoMemForDecodedImage, sh_video->o_bih.biSizeImage);
    return 0;
  }

  if(yuv && sh_video->codec->outflags[sh_video->outfmtidx] & CODECS_FLAG_YUVHACK)
    sh_video->o_bih.biCompression = outfmt;

//  avi_header.our_in_buffer=malloc(avi_header.video.dwSuggestedBufferSize); // FIXME!!!!

  (*ICSendMessage_ptr)(sh_video->hic, ICM_USER+80, (long)(&divx_quality) ,NULL);

  MSG_V("VIDEO CODEC Init OK!!! ;-)\n");
  return 1;
}


int vfw_set_postproc(sh_video_t* sh_video,int quality){
  // Works only with opendivx/divx4 based DLL
  return (*ICSendMessage_ptr)(sh_video->hic, ICM_USER+80, (long)(&quality) ,NULL);
}

int vfw_decode_video(sh_video_t* sh_video,void* start,int in_size,int drop_frame,int ex){
    HRESULT ret;

    sh_video->bih->biSizeImage = in_size;

    if(ex)
      ret = (*ICDecompressEx_ptr)(sh_video->hic, 
	  ( (sh_video->ds->flags&1) ? 0 : ICDECOMPRESS_NOTKEYFRAME ) |
	  ( (drop_frame==2 && !(sh_video->ds->flags&1))?(ICDECOMPRESS_HURRYUP|ICDECOMPRESS_PREROL):0 ) , 
                         sh_video->bih,   start,
                        &sh_video->o_bih,
                        drop_frame ? 0 : sh_video->our_out_buffer);
    else
      ret = (*ICDecompress_ptr)(sh_video->hic, 
	  ( (sh_video->ds->flags&1) ? 0 : ICDECOMPRESS_NOTKEYFRAME ) |
	  ( (drop_frame==2 && !(sh_video->ds->flags&1))?(ICDECOMPRESS_HURRYUP|ICDECOMPRESS_PREROL):0 ) , 
                         sh_video->bih,   start,
                        &sh_video->o_bih,
                        drop_frame ? 0 : sh_video->our_out_buffer);

    return (int)ret;
}

int vfw_close_video_codec(sh_video_t *sh_video, int ex)
{
    HRESULT ret;
    
    ret = ICDecompressEnd(sh_video->hic);
    if (ret)
    {
	MSG_WARN( "ICDecompressEnd failed: %d\n", ret);
	return(0);
    }

    ret = ICClose(sh_video->hic);
    if (ret)
    {
	MSG_WARN( "ICClose failed: %d\n", ret);
	return(0);
    }

    return(1);
}

// to set/get/query special features/parameters
static int control(sh_video_t *sh,int cmd,void* arg,...){
    switch(cmd){
    case VDCTRL_QUERY_MAX_PP_LEVEL:
	return 9;
    case VDCTRL_SET_PP_LEVEL:
	vfw_set_postproc(sh,10*(*((int*)arg)));
	return CONTROL_OK;
    }
    return CONTROL_UNKNOWN;
}

// init driver
static int init(sh_video_t *sh){
    if(!load_lib(wineld_name("libloader.so"))) return 0;
    if(strcmp(sh->codec->driver_name,"vfwex") == 0) vfw_ex=1;
    if(!init_vfw_video_codec(sh,vfw_ex)) return 0;
    MSG_V("INFO: Win32 video codec (%s) init OK!\n",LIBDIR"/wine/libloader.so");
    return mpcodecs_config_vo(sh,sh->disp_w,sh->disp_h,IMGFMT_YUY2);
}

// uninit driver
static void uninit(sh_video_t *sh)
{
  vfw_close_video_codec(sh,vfw_ex);
  dlclose(dll_handle);
}

// decode a frame
static mp_image_t* decode(sh_video_t *sh,void* data,int len,int flags){
    mp_image_t* mpi;
    int ret;
    if(len<=0) return NULL; // skipped frame

    mpi=mpcodecs_get_image(sh, MP_IMGTYPE_IP, MP_IMGFLAG_ACCEPT_WIDTH, 
	sh->disp_w, sh->disp_h);
    if(!mpi){	// temporary!
	MSG_ERR("couldn't allocate image for vfw codec\n");
	return NULL;
    }

    // set buffer:
    sh->our_out_buffer=mpi->planes[0];

    // set stride:  (trick discovered by Andreas Ackermann - thanx!)
    sh->bih->biWidth=mpi->width; //mpi->stride[0]/(mpi->bpp/8);
    sh->o_bih.biWidth=mpi->width; //mpi->stride[0]/(mpi->bpp/8);

    if((ret=vfw_decode_video(sh,data,len,flags&3,vfw_ex))){
      MSG_WARN("Error decompressing frame, err=%d\n",ret);
      return NULL;
    }
    
    return mpi;
}
