#include <stdio.h>
#include "../config.h"
#include "../mp_msg.h"
#define MSG_V(args...) mp_msg(MSGT_GLOBAL,MSGL_V, ##args )

#ifdef USE_FASTMEMCPY
#include "fastmemcpy.h"
/*
  aclib - advanced C library ;)
  This file contains functions which improve and expand standard C-library
  see aclib_template.c ... this file only contains runtime cpu detection and config options stuff
  runtime cpu detection by michael niedermayer (michaelni@gmx.at) is under GPL
*/
#include <stddef.h>
#include "../cpudetect.h"

#define BLOCK_SIZE 4096
#define CONFUSION_FACTOR 0
//Feel free to fine-tune the above 2, it might be possible to get some speedup with them :)

//#define STATISTICS

#ifdef ARCH_X86
#define CAN_COMPILE_X86_ASM
#endif

//Note: we have MMX, MMX2, 3DNOW version there is no 3DNOW+MMX2 one
//Plain C versions
//#if !defined (HAVE_MMX) || defined (RUNTIME_CPUDETECT)
//#define COMPILE_C
//#endif

#ifdef CAN_COMPILE_X86_ASM

#undef HAVE_MMX
#undef HAVE_MMX2
#undef HAVE_3DNOW
#undef HAVE_SSE

//MMX versions
#undef RENAME
#define HAVE_MMX
#undef HAVE_MMX2
#undef HAVE_3DNOW
#define RENAME(a) a ## _MMX
#include "aclib_template.c"

//MMX2 versions
#undef RENAME
#define HAVE_MMX
#define HAVE_MMX2
#undef HAVE_3DNOW
#define RENAME(a) a ## _MMX2
#include "aclib_template.c"

//3DNOW versions
#undef RENAME
#define HAVE_MMX
#undef HAVE_MMX2
#define HAVE_3DNOW
#define RENAME(a) a ## _3DNow
#include "aclib_template.c"

#endif // CAN_COMPILE_X86_ASM

extern int verbose;
static void * init_fast_memcpy(void * to, const void * from, size_t len)
{
#ifdef CAN_COMPILE_X86_ASM
	// ordered per speed fasterst first
	if(gCpuCaps.hasMMX2)
	{
		MSG_V("Using MMX2 optimized memcpy\n");
		fast_memcpy_ptr = fast_memcpy_MMX2;
	}
	else if(gCpuCaps.has3DNow)
	{
		MSG_V("Using 3DNow optimized memcpy\n");
		fast_memcpy_ptr = fast_memcpy_3DNow;
	}
	else if(gCpuCaps.hasMMX)
	{
		MSG_V("Using MMX optimized memcpy\n");
		fast_memcpy_ptr = fast_memcpy_MMX;
	}
	else
#endif //CAN_COMPILE_X86_ASM
	{
		MSG_V("Using generic memcpy\n");
		fast_memcpy_ptr = memcpy; /* prior to mmx we use the standart memcpy */
	}
	return (*fast_memcpy_ptr)(to,from,len);	
}

static void * init_mem2agpcpy(void * to, const void * from, size_t len)
{
#ifdef CAN_COMPILE_X86_ASM
	// ordered per speed fasterst first
	if(gCpuCaps.hasMMX2)
	{
		MSG_V("Using MMX2 optimized agpcpy\n");
		mem2agpcpy_ptr = mem2agpcpy_MMX2;
	}
	else if(gCpuCaps.has3DNow)
	{
		MSG_V("Using 3DNow optimized agpcpy\n");
		mem2agpcpy_ptr = mem2agpcpy_3DNow;
	}
	else if(gCpuCaps.hasMMX)
	{
		MSG_V("Using MMX optimized agpcpy\n");
		mem2agpcpy_ptr = mem2agpcpy_MMX;
	}
	else
#endif //CAN_COMPILE_X86_ASM
	{
		MSG_V("Using generic optimized agpcpy\n");
		mem2agpcpy_ptr = memcpy; /* prior to mmx we use the standart memcpy */
	}
	return (*mem2agpcpy_ptr)(to,from,len);	
}

void *(*fast_memcpy_ptr)(void * to, const void * from, size_t len) = init_fast_memcpy;
void *(*mem2agpcpy_ptr)(void * to, const void * from, size_t len) = init_mem2agpcpy;

#endif /* use fastmemcpy */

