/* 
 * video_out_pgm.c, pgm interface
 *
 *
 * Copyright (C) 1996, MPEG Software Simulation Group. All Rights Reserved. 
 *
 * Hacked into mpeg2dec by
 * 
 * Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * 15 & 16 bpp support added by Franck Sicard <Franck.Sicard@solsoft.fr>
 *
 * Xv image suuport by Gerd Knorr <kraxel@goldbach.in-berlin.de>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "config.h"
#include "video_out.h"
#include "video_out_internal.h"

LIBVO_EXTERN (pgm)

static vo_info_t vo_info = 
{
	"PGM file",
	"pgm",
	"walken",
	""
};

static int image_width;
static int image_height;
static char header[1024];
static int framenum = 0;

static uint8_t *image=NULL;

char vo_pgm_filename[24];

static uint32_t
config(uint32_t width, uint32_t height, uint32_t d_width, uint32_t d_height, uint32_t fullscreen, char *title, uint32_t format,const vo_tune_info_t *info)
{
	image_height = height;
	image_width = width;
	image=malloc(width*height*3/2);
	snprintf (header, 1024, "P5\n\n%d %d\n255\n", width, height*3/2);
	UNUSED(d_width);
	UNUSED(d_height);
	UNUSED(fullscreen);
	UNUSED(title);
	UNUSED(format);
	UNUSED(info);
	return 0;
}

static const vo_info_t*
get_info(void)
{
    return &vo_info;
}

static void flip_page (unsigned idx)
{
    FILE * f;
    UNUSED(idx);
    snprintf (vo_pgm_filename, 24, "%08d.pgm", framenum++);

    f = fopen (vo_pgm_filename, "wb");  if (f == NULL) return;
    fwrite (header, strlen (header), 1, f);
    fwrite (image, image_width, image_height*3/2, f);
    fclose (f);

    return;
}

static uint32_t draw_slice(vo_slice_t *slice)
{
    unsigned i;
    // copy Y:
    uint8_t *dst=image+image_width*slice->y+slice->x;
    uint8_t *src=slice->src[0];
    for(i=0;i<slice->h;i++){
        memcpy(dst,src,slice->w);
        src+=slice->stride[0];
        dst+=image_width;
    }
{
    // copy U+V:
    uint8_t *src1=slice->src[1];
    uint8_t *src2=slice->src[2];
    uint8_t *dst=image+image_width*image_height+image_width*(slice->y/2)+(slice->x/2);
    for(i=0;i<slice->h/2;i++){
        memcpy(dst,src1,slice->w/2);
        memcpy(dst+image_width/2,src2,slice->w/2);
        src1+=slice->stride[1];
        src2+=slice->stride[2];
        dst+=image_width;
    }

}
    
    return 0;
}

static uint32_t
query_format(vo_query_fourcc_t *format)
{
    if(format->fourcc==IMGFMT_YV12) return 1;
//    switch(format){
//    case IMGFMT_YV12:
//    case IMGFMT_RGB|24:
//    case IMGFMT_BGR|24:
//        return 1;
//    }
    return 0;
}

static void
uninit(void)
{
    if(image){ free(image);image=NULL;}
}

static uint32_t preinit(const char *arg)
{
    if(arg) 
    {
	MSG_ERR("vo_pgm: Unknown subdevice: %s\n",arg);
	return ENOSYS;
    }
    return 0;
}

static uint32_t control(uint32_t request, void *data, ...)
{
  switch (request) {
  case VOCTRL_QUERY_FORMAT:
    return query_format((vo_query_fourcc_t*)data);
  case VOCTRL_DRAW_SLICE:
    return draw_slice((vo_slice_t*)data);
  }
  return VO_NOTIMPL;
}
