/*
 * vo_xv.c, X11 Xv interface
 *
 * Copyright (C) 1996, MPEG Software Simulation Group. All Rights Reserved.
 *
 * Hacked into mpeg2dec by
 *
 * Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * Xv image suuport by Gerd Knorr <kraxel@goldbach.in-berlin.de>
 * fullscreen support by Pontscho
 * double buffering support by A'rpi
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "../dec_ahead.h"
#include "video_out.h"
#include "video_out_internal.h"

LIBVO_EXTERN(xv)

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#include <errno.h>

#include "x11_common.h"

#include "fastmemcpy.h"
#include "sub.h"
#include "aspect.h"
#include "../mp_image.h"
#include "dri_vo.h"

static vo_info_t vo_info =
{
        "X11/Xv",
        "xv",
        "Gerd Knorr <kraxel@goldbach.in-berlin.de>",
        ""
};

extern int verbose;
extern int enable_xp;
/* since it doesn't seem to be defined on some platforms */
int XShmGetEventBase(Display*);

/* X11 related variables */
static unsigned depth;
static XWindowAttributes attribs;

#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>
// FIXME: dynamically allocate this stuff
static void allocate_xvimage(int);
static unsigned int ver,rel,req,ev,err;
static unsigned int formats, adaptors,i,xv_port,xv_format,xv_bpp;
static XvAdaptorInfo        *ai;
static XvImageFormatValues  *fo;

static unsigned expose_idx=0,num_buffers=1; // default
extern unsigned vo_da_buffs;
static XvImage* xvimage[MAX_DRI_BUFFERS];
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

static XShmSegmentInfo Shminfo[MAX_DRI_BUFFERS];

static uint32_t image_width;
static uint32_t image_height;
static uint32_t image_format;
static int flip_flag;

static Window                 mRoot;
static uint32_t               drwX,drwY,drwWidth,drwHeight,drwBorderWidth,drwDepth;
static uint32_t               drwcX,drwcY,dwidth,dheight;

static int __xv_set_video_eq( const vidix_video_eq_t *info,int use_reset)
{
 XvAttribute *attributes;
 int howmany, xv_min,xv_max,xv_atomka;
 static int was_reset = 0;
/* get available attributes */
     attributes = XvQueryPortAttributes(mDisplay, xv_port, &howmany);
     /* first pass try reset */
     if(use_reset)
     {
	for (i = 0; (int)i < howmany && attributes; i++)
        {
            if (attributes[i].flags & XvSettable && !strcmp(attributes[i].name,"XV_SET_DEFAULTS"
))
            {
		was_reset = 1;
		MSG_DBG2("vo_xv: reset gamma correction\n");
                xv_atomka = XInternAtom(mDisplay, attributes[i].name, True);
                XvSetPortAttribute(mDisplay, xv_port, xv_atomka, attributes[i].max_value);
	    }
        }
	/* for safety purposes */
	if(!was_reset) return ENOSYS;
     }
     for (i = 0; (int)i < howmany && attributes; i++)
     {
            if (attributes[i].flags & XvSettable)
            {
                xv_min = attributes[i].min_value;
                xv_max = attributes[i].max_value;
                xv_atomka = XInternAtom(mDisplay, attributes[i].name, True);
/* since we have SET_DEFAULTS first in our list, we can check if it's available
   then trigger it if it's ok so that the other values are at default upon query */
                if (xv_atomka != None)
                {
		    int port_value,port_min,port_max,port_mid;
		    if(strcmp(attributes[i].name,"XV_BRIGHTNESS") == 0
			      && (info->cap & VEQ_CAP_BRIGHTNESS))
				port_value = info->brightness;
		    else
		    if(strcmp(attributes[i].name,"XV_SATURATION") == 0
			      && (info->cap & VEQ_CAP_SATURATION))
				port_value = info->saturation;
		    else
		    if(strcmp(attributes[i].name,"XV_CONTRAST") == 0
			      && (info->cap & VEQ_CAP_CONTRAST))
				port_value = info->contrast;
		    else
		    if(strcmp(attributes[i].name,"XV_HUE") == 0
			      && (info->cap & VEQ_CAP_HUE))
				port_value = info->hue;
		    else
                    /* Note: since 22.01.2002 GATOS supports these attrs for radeons (NK) */
		    if(strcmp(attributes[i].name,"XV_RED_INTENSITY") == 0
			      && (info->cap & VEQ_CAP_RGB_INTENSITY))
				port_value = info->red_intensity;
		    else
		    if(strcmp(attributes[i].name,"XV_GREEN_INTENSITY") == 0
			      && (info->cap & VEQ_CAP_RGB_INTENSITY))
				port_value = info->green_intensity;
		    else
		    if(strcmp(attributes[i].name,"XV_BLUE_INTENSITY") == 0
			      && (info->cap & VEQ_CAP_RGB_INTENSITY))
				port_value = info->blue_intensity;
		    else continue;
		    /* means that user has untouched this parameter since
		       NVidia driver has default == min for XV_HUE but not mid */
		    if(!port_value && use_reset) continue;
		    port_min = xv_min;
		    port_max = xv_max;
		    port_mid = (port_min + port_max) / 2;
		    port_value = port_mid + (port_value * (port_max - port_min)) / 2000;
		    MSG_DBG2("vo_xv: set gamma %s to %i (min %i max %i mid %i)\n",attributes[i].name,port_value,port_min,port_max,port_mid);
                    XvSetPortAttribute(mDisplay, xv_port, xv_atomka, port_value);
                }
        }
    }
    return 0;
}


static int xv_set_video_eq( const vidix_video_eq_t *info)
{
    return __xv_set_video_eq(info,0);
}

static int xv_get_video_eq( vidix_video_eq_t *info)
{
 XvAttribute *attributes;
 int howmany, xv_min,xv_max,xv_atomka;
/* get available attributes */
     memset(info,0,sizeof(vidix_video_eq_t));
     attributes = XvQueryPortAttributes(mDisplay, xv_port, &howmany);
     for (i = 0; (int)i < howmany && attributes; i++)
     {
            if (attributes[i].flags & XvGettable)
            {
                xv_min = attributes[i].min_value;
                xv_max = attributes[i].max_value;
                xv_atomka = XInternAtom(mDisplay, attributes[i].name, True);
/* since we have SET_DEFAULTS first in our list, we can check if it's available
   then trigger it if it's ok so that the other values are at default upon query */
                if (xv_atomka != None)
                {
		    int port_value,port_min,port_max,port_mid;
                    XvGetPortAttribute(mDisplay, xv_port, xv_atomka, &port_value);
		    MSG_DBG2("vo_xv: get: %s = %i\n",attributes[i].name,port_value);

		    port_min = xv_min;
		    port_max = xv_max;
		    port_mid = (port_min + port_max) / 2;		    
		    port_value = ((port_value - port_mid)*2000)/(port_max-port_min);
		    
		    MSG_DBG2("vo_xv: assume: %s = %i\n",attributes[i].name,port_value);
		    
		    if(strcmp(attributes[i].name,"XV_BRIGHTNESS") == 0)
		    {
			info->cap |= VEQ_CAP_BRIGHTNESS;
			info->brightness = port_value;
		    }
		    else
		    if(strcmp(attributes[i].name,"XV_SATURATION") == 0)
		    {
			info->cap |= VEQ_CAP_SATURATION;
			info->saturation = port_value;
		    }
		    else
		    if(strcmp(attributes[i].name,"XV_CONTRAST") == 0)
		    {
			info->cap |= VEQ_CAP_CONTRAST;
			info->contrast = port_value;
		    }
		    else
		    if(strcmp(attributes[i].name,"XV_HUE") == 0)
		    {
			info->cap |= VEQ_CAP_HUE;
			info->hue = port_value;
		    }
		    else
                    /* Note: since 22.01.2002 GATOS supports these attrs for radeons (NK) */
		    if(strcmp(attributes[i].name,"XV_RED_INTENSITY") == 0)
		    {
			info->cap |= VEQ_CAP_RGB_INTENSITY;
			info->red_intensity = port_value;
		    }
		    else
		    if(strcmp(attributes[i].name,"XV_GREEN_INTENSITY") == 0)
		    {
			info->cap |= VEQ_CAP_RGB_INTENSITY;
			info->green_intensity = port_value;
		    }
		    else
		    if(strcmp(attributes[i].name,"XV_BLUE_INTENSITY") == 0)
		    {
			info->cap |= VEQ_CAP_RGB_INTENSITY;
			info->blue_intensity = port_value;
		    }
		    else continue;
                }
        }
    }
    return 0;
}

extern int vo_gamma_brightness;
extern int vo_gamma_saturation;
extern int vo_gamma_contrast;
extern int vo_gamma_hue;
extern int vo_gamma_red_intensity;
extern int vo_gamma_green_intensity;
extern int vo_gamma_blue_intensity;

static void set_gamma_correction( void )
{
  vidix_video_eq_t info;
  /* try all */
  info.cap = VEQ_CAP_BRIGHTNESS | VEQ_CAP_CONTRAST | VEQ_CAP_SATURATION |
	     VEQ_CAP_HUE | VEQ_CAP_RGB_INTENSITY;
  info.flags = 0; /* doesn't matter for xv */
  info.brightness = vo_gamma_brightness;
  info.contrast = vo_gamma_contrast;
  info.saturation = vo_gamma_saturation;
  info.hue = vo_gamma_hue;
  info.red_intensity = vo_gamma_red_intensity;
  info.green_intensity = vo_gamma_green_intensity;
  info.blue_intensity = vo_gamma_blue_intensity;
  /* reset with XV_SET_DEFAULTS only once */
  __xv_set_video_eq(&info,1);
}

/* unofficial gatos extensions */
#define FOURCC_RGBA32	0x41424752
#define FOURCC_RGB24	0x00000000
#define FOURCC_RGBT16	0x54424752
#define FOURCC_RGB16	0x32424752

/*
 * connect to server, create and map window,
 * allocate colors and (shared) memory
 */
static uint32_t config(uint32_t width, uint32_t height, uint32_t d_width, uint32_t d_height, uint32_t flags, char *title, uint32_t format,const vo_tune_info_t *info)
{
// int screen;
 char *hello = (title == NULL) ? "Xv render" : title;
// char *name = ":0.0";
 XSizeHints hint;
 XVisualInfo vinfo;
 XvPortID xv_p;

 XGCValues xgcv;
 XSetWindowAttributes xswa;
 unsigned long xswamask,i;
#ifdef HAVE_XF86VM
 int vm=0;
 unsigned int modeline_width, modeline_height;
 static uint32_t vm_width;
 static uint32_t vm_height;
#endif

 UNUSED(info);
 aspect_save_orig(width,height);
 aspect_save_prescale(d_width,d_height);

 image_height = height;
 image_width = width;
 image_format=format;

 vo_fs=flags&1;
 if ( vo_fs )
  { vo_old_width=d_width; vo_old_height=d_height; }
     
#ifdef HAVE_XF86VM
 if( flags&0x02 ) vm = 1;
#endif
 flip_flag=flags&8;
 num_buffers=vo_doublebuffering?vo_da_buffs:1;
 

 aspect_save_screenres(vo_screenwidth,vo_screenheight);

   hint.x = 0;
   hint.y = 0;
   hint.width = d_width;
   hint.height = d_height;
   aspect(&d_width,&d_height,A_NOZOOM);
#ifdef HAVE_XF86VM
    if ( vm )
      {
	if ((d_width==0) && (d_height==0))
	  { vm_width=image_width; vm_height=image_height; }
	else
	  { vm_width=d_width; vm_height=d_height; }
	vo_vm_switch(vm_width, vm_height,&modeline_width, &modeline_height);
	hint.x=(vo_screenwidth-modeline_width)/2;
	hint.y=(vo_screenheight-modeline_height)/2;
	hint.width=modeline_width;
	hint.height=modeline_height;
      }
    else
#endif
   if ( vo_fs )
    {
     hint.width=vo_screenwidth;
     hint.height=vo_screenheight;
#ifdef X11_FULLSCREEN
     /* this code replaces X11_FULLSCREEN hack in mplayer.c
      * aspect() is available through aspect.h for all vos.
      * besides zooming should only be done with -zoom,
      * but I leave the old -fs behaviour so users don't get
      * irritated for now (and send lots o' mails ;) ::atmos
      */

     aspect(&d_width,&d_height,A_ZOOM);
#endif

    }
   dwidth=d_width; dheight=d_height; //XXX: what are the copy vars used for?
   hint.flags = PPosition | PSize;
   XGetWindowAttributes(mDisplay, DefaultRootWindow(mDisplay), &attribs);
   depth=attribs.depth;
   if (depth != 15 && depth != 16 && depth != 24 && depth != 32) depth = 24;
   XMatchVisualInfo(mDisplay, mScreen, depth, TrueColor, &vinfo);

   xswa.background_pixel = 0;
   xswa.border_pixel     = 0;
   xswamask = CWBackPixel | CWBorderPixel;

    if ( WinID>=0 ){
      vo_window = WinID ? ((Window)WinID) : RootWindow(mDisplay,mScreen);
      XUnmapWindow( mDisplay,vo_window );
      XChangeWindowAttributes( mDisplay,vo_window,xswamask,&xswa );
    } else 

   vo_window = XCreateWindow(mDisplay, RootWindow(mDisplay,mScreen),
       hint.x, hint.y, hint.width, hint.height,
       0, depth,CopyFromParent,vinfo.visual,xswamask,&xswa);

   vo_x11_classhint( mDisplay,vo_window,"xv" );
   vo_hidecursor(mDisplay,vo_window);

   XSelectInput(mDisplay, vo_window, StructureNotifyMask | KeyPressMask | 
	((WinID==0) ? 0 : (PointerMotionMask
		| ButtonPressMask | ButtonReleaseMask
   )));
   XSetStandardProperties(mDisplay, vo_window, hello, hello, None, NULL, 0, &hint);
   if ( vo_fs ) vo_x11_decoration( mDisplay,vo_window,0 );
   XMapWindow(mDisplay, vo_window);
#ifdef HAVE_XINERAMA
   vo_x11_xinerama_move(mDisplay,vo_window);
#endif
   vo_gc = XCreateGC(mDisplay, vo_window, 0L, &xgcv);
   XFlush(mDisplay);
   XSync(mDisplay, False);
#ifdef HAVE_XF86VM
    if ( vm )
     {
      /* Grab the mouse pointer in our window */
      XGrabPointer(mDisplay, vo_window, True, 0,
                   GrabModeAsync, GrabModeAsync,
                   vo_window, None, CurrentTime);
      XSetInputFocus(mDisplay, vo_window, RevertToNone, CurrentTime);
     }
#endif

 xv_port = 0;
 if (Success == XvQueryExtension(mDisplay,&ver,&rel,&req,&ev,&err))
  {
   /* check for Xvideo support */
   if (Success != XvQueryAdaptors(mDisplay,DefaultRootWindow(mDisplay), &adaptors,&ai))
    {
     MSG_ERR("Xv: XvQueryAdaptors failed");
     return -1;
    }
   /* check adaptors */
   for (i = 0; i < adaptors && xv_port == 0; i++)
    {
     if ((ai[i].type & XvInputMask) && (ai[i].type & XvImageMask))
	 for (xv_p = ai[i].base_id; xv_p < ai[i].base_id+ai[i].num_ports; ++xv_p)
	 {
	     if (!XvGrabPort(mDisplay, xv_p, CurrentTime)) {
		 xv_port = xv_p;
		 break;
	     } else {
		 MSG_ERR("Xv: could not grab port %i\n", (int)xv_p);
	     }
	 }
    }
   /* check image formats */
   if (xv_port != 0)
    {
     fo = XvListImageFormats(mDisplay, xv_port, (int*)&formats);
     xv_format=0;
     if(format==IMGFMT_BGR32) format=FOURCC_RGBA32;
     if(format==IMGFMT_BGR16) format=FOURCC_RGB16;
     for(i = 0; i < formats; i++){
       MSG_INFO("Xvideo image format: 0x%x (%4.4s) %s\n", fo[i].id,(char*)&fo[i].id, (fo[i].format == XvPacked) ? "packed" : "planar");
       if (fo[i].id == (int)format) xv_format = fo[i].id;
     }
     if (!xv_format) xv_port = 0;
    }

   if (xv_port != 0)
    {
     MSG_INFO( "using Xvideo port %d for hw scaling\n",xv_port );
       
       switch (xv_format){
	case IMGFMT_IF09:
	case IMGFMT_YVU9:
			  xv_bpp=9;
			  break;
	case IMGFMT_YV12:
	case IMGFMT_I420:
        case IMGFMT_IYUV:
			  xv_bpp=12;
			  break;
	case IMGFMT_YUY2:
	case IMGFMT_YVYU:
			  xv_bpp=16;
			  break;	
	case IMGFMT_UYVY:
			  xv_bpp=16;
			  break;
	case FOURCC_RGBA32:
			  xv_bpp=32;
			  break;
	case FOURCC_RGB16:
			  xv_bpp=16;
			  break;
	default:
			  xv_bpp = 16;
       }

      for(i=0;i<num_buffers;++i) allocate_xvimage(i);

     set_gamma_correction();

     XGetGeometry( mDisplay,vo_window,&mRoot,&drwX,&drwY,&drwWidth,&drwHeight,&drwBorderWidth,&drwDepth );
     drwX=0; drwY=0;
     XTranslateCoordinates( mDisplay,vo_window,mRoot,0,0,&drwcX,&drwcY,&mRoot );
     MSG_INFO( "[xv] dcx: %d dcy: %d dx: %d dy: %d dw: %d dh: %d\n",drwcX,drwcY,drwX,drwY,drwWidth,drwHeight );

     aspect(&dwidth,&dheight,A_NOZOOM);
     if ( vo_fs )
      {
       aspect(&dwidth,&dheight,A_ZOOM);
       drwX=( vo_screenwidth - (dwidth > vo_screenwidth?vo_screenwidth:dwidth) ) / 2;
       drwcX+=drwX;
       drwY=( vo_screenheight - (dheight > vo_screenheight?vo_screenheight:dheight) ) / 2;
       drwcY+=drwY;
       drwWidth=(dwidth > vo_screenwidth?vo_screenwidth:dwidth);
       drwHeight=(dheight > vo_screenheight?vo_screenheight:dheight);
       MSG_INFO( "[xv-fs] dcx: %d dcy: %d dx: %d dy: %d dw: %d dh: %d\n",drwcX,drwcY,drwX,drwY,drwWidth,drwHeight );
      }
     saver_off(mDisplay);  // turning off screen saver
     return 0;
    }
  }

 MSG_FATAL("Sorry, Xv not supported by this X11 version/driver\n");
 MSG_FATAL("******** Try with  -vo x11  or  -vo sdl  *********\n");
 return 1;
}

static const vo_info_t * get_info(void)
{ return &vo_info; }

static void allocate_xvimage(int foo)
{
 /*
  * allocate XvImages.  FIXME: no error checking, without
  * mit-shm this will bomb...
  */
 xvimage[foo] = XvShmCreateImage(mDisplay, xv_port, xv_format, 0, image_width, image_height, &Shminfo[foo]);

 Shminfo[foo].shmid    = shmget(IPC_PRIVATE, xvimage[foo]->data_size, IPC_CREAT | 0777);
 Shminfo[foo].shmaddr  = (char *) shmat(Shminfo[foo].shmid, 0, 0);
 Shminfo[foo].readOnly = False;

 xvimage[foo]->data = Shminfo[foo].shmaddr;
 XShmAttach(mDisplay, &Shminfo[foo]);
 XSync(mDisplay, False);
 shmctl(Shminfo[foo].shmid, IPC_RMID, 0);
 memset(xvimage[foo]->data,128,xvimage[foo]->data_size);
 return;
}

static void deallocate_xvimage(int foo)
{
 XShmDetach( mDisplay,&Shminfo[foo] );
 shmdt( Shminfo[foo].shmaddr );
 XFlush( mDisplay );
 XSync(mDisplay, False);
 return;
}

static uint32_t check_events(void)
{
 uint32_t e=vo_x11_check_events(mDisplay);
 if(e&VO_EVENT_RESIZE)
  {
   XGetGeometry( mDisplay,vo_window,&mRoot,&drwX,&drwY,&drwWidth,&drwHeight,&drwBorderWidth,&drwDepth );
   drwX=0; drwY=0;
   XTranslateCoordinates( mDisplay,vo_window,mRoot,0,0,&drwcX,&drwcY,&mRoot );
   MSG_V( "[xv-resize] dcx: %d dcy: %d dx: %d dy: %d dw: %d dh: %d\n",drwcX,drwcY,drwX,drwY,drwWidth,drwHeight );

   aspect(&dwidth,&dheight,A_NOZOOM);
   if ( vo_fs )
    {
     aspect(&dwidth,&dheight,A_ZOOM);
     drwX=( vo_screenwidth - (dwidth > vo_screenwidth?vo_screenwidth:dwidth) ) / 2;
     drwcX+=drwX;
     drwY=( vo_screenheight - (dheight > vo_screenheight?vo_screenheight:dheight) ) / 2;
     drwcY+=drwY;
     drwWidth=(dwidth > vo_screenwidth?vo_screenwidth:dwidth);
     drwHeight=(dheight > vo_screenheight?vo_screenheight:dheight);
     MSG_V( "[xv-fs] dcx: %d dcy: %d dx: %d dy: %d dw: %d dh: %d\n",drwcX,drwcY,drwX,drwY,drwWidth,drwHeight );
    }
  }
 if ( e & VO_EVENT_EXPOSE )
  {
   XvShmPutImage(mDisplay, xv_port, vo_window, vo_gc, xvimage[expose_idx], 0, 0,  image_width, image_height, drwX, drwY, 1, 1, False);
   XvShmPutImage(mDisplay, xv_port, vo_window, vo_gc, xvimage[expose_idx], 0, 0,  image_width, image_height, drwX,drwY,drwWidth,(vo_fs?drwHeight - 1:drwHeight), False);
  }
  return e|VO_EVENT_FORCE_UPDATE;
}

static void flip_page(unsigned idx)
{
 expose_idx=idx;
 XvShmPutImage(mDisplay, xv_port, vo_window, vo_gc, xvimage[idx],
         0, 0,  image_width, image_height,
         drwX,drwY,drwWidth,(vo_fs?drwHeight - 1:drwHeight),
         False);
 if (num_buffers>1) XFlush(mDisplay);
 else XSync(mDisplay, False);
 return;
}

static uint32_t query_format(vo_query_fourcc_t* format)
{
 XvPortID xv_p;
 if (!vo_x11_init()) return 0;
 xv_port = 0;
 if (Success == XvQueryExtension(mDisplay,&ver,&rel,&req,&ev,&err))
  {
   /* check for Xvideo support */
   if (Success != XvQueryAdaptors(mDisplay,DefaultRootWindow(mDisplay), &adaptors,&ai))
    {
     MSG_ERR("Xv: XvQueryAdaptors failed");
     return -1;
    }
   /* check adaptors */
   for (i = 0; i < adaptors && xv_port == 0; i++)
    {
     if ((ai[i].type & XvInputMask) && (ai[i].type & XvImageMask))
	 for (xv_p = ai[i].base_id; xv_p < ai[i].base_id+ai[i].num_ports; ++xv_p)
	 {
	     if (!XvGrabPort(mDisplay, xv_p, CurrentTime)) {
		 xv_port = xv_p;
		 break;
	     } else {
		 MSG_ERR("Xv: could not grab port %i\n", (int)xv_p);
	     }
	 }
    }
   /* check image formats */
   if (xv_port != 0)
    {
     fo = XvListImageFormats(mDisplay, xv_port, (int*)&formats);
     xv_format=0;
     for(i = 0; i < formats; i++){
       if(fo[i].id == (int)format->fourcc) return 1;
       if(fo[i].id == FOURCC_RGBA32 && format->fourcc == IMGFMT_BGR32) return 1;
       if(fo[i].id == FOURCC_RGB16 && format->fourcc == IMGFMT_BGR16) return 1;
     }
     if (!xv_format) xv_port = 0;
    }
  }
return 0;
}

static void uninit(void) 
{
 unsigned i;
 saver_on(mDisplay); // screen saver back on
 for( i=0;i<num_buffers;i++ ) deallocate_xvimage( i );
#ifdef HAVE_XF86VM
 vo_vm_close(mDisplay);
#endif
 vo_x11_uninit(mDisplay, vo_window);
}

static uint32_t preinit(const char *arg)
{
    if(arg) 
    {
	MSG_ERR("vo_xv: Unknown subdevice: %s\n",arg);
	return ENOSYS;
    }
    return 0;
}

static void query_vaa(vo_vaa_t *vaa)
{
  memset(vaa,0,sizeof(vo_vaa_t));
  vaa->get_video_eq = xv_get_video_eq;
  vaa->set_video_eq = xv_set_video_eq;
}

#ifndef max
#define max(a,b) ((a)>(b)?(a):(b))
#endif
#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#endif

static void xv_dri_get_surface_caps(dri_surface_cap_t *caps)
{
    unsigned i,n;
    caps->caps = DRI_CAP_TEMP_VIDEO | DRI_CAP_UPSCALER | DRI_CAP_DOWNSCALER |
		DRI_CAP_HORZSCALER | DRI_CAP_VERTSCALER;
    caps->fourcc = image_format;
    caps->width=image_width;
    caps->height=image_height;
    caps->x=0;
    caps->y=0;
    caps->w=image_width;
    caps->h=image_height;
    n=min(4,xvimage[0]->num_planes);
    for(i=0;i<n;i++)
	caps->strides[i] = xvimage[0]->pitches[i];
    for(;i<4;i++)
	caps->strides[i] = 0;
    if(image_format == IMGFMT_YV12)
    {
	unsigned ts;
	ts = caps->strides[2];
	caps->strides[2] = caps->strides[1];
	caps->strides[1] = ts;
    }
}

static void xv_dri_get_surface(dri_surface_t *surf)
{
    unsigned i,n;
    n=min(4,xvimage[0]->num_planes);
    for(i=0;i<n;i++)
	surf->planes[i] = xvimage[surf->idx]->data + xvimage[surf->idx]->offsets[i];
    for(;i<4;i++)
	surf->planes[i] = 0;
    if(image_format == IMGFMT_YV12)
    {
	void * tp;
	tp = surf->planes[2];
	surf->planes[2] = surf->planes[1];
	surf->planes[1] = tp;
    }
}

static uint32_t control(uint32_t request, void *data, ...)
{
  switch (request) {
  case VOCTRL_QUERY_VAA:
    query_vaa((vo_vaa_t*)data);
    return VO_TRUE;
  case VOCTRL_QUERY_FORMAT:
    return query_format((vo_query_fourcc_t*)data);
  case VOCTRL_FULLSCREEN:
    vo_x11_fullscreen();
    return VO_TRUE;
  case VOCTRL_CHECK_EVENTS:
    *(uint32_t *)data = check_events();
    if(enable_xp && (*(uint32_t *)data & VO_EVENT_RESIZE)==VO_EVENT_RESIZE)
		    LOCK_VDECODING(); /* just for compatibility with other vo */
    return VO_TRUE;
  case VOCTRL_GET_NUM_FRAMES:
	*(uint32_t *)data = num_buffers;
	return VO_TRUE;
  case DRI_GET_SURFACE_CAPS:
	xv_dri_get_surface_caps(data);
	return VO_TRUE;
  case DRI_GET_SURFACE: 
	xv_dri_get_surface(data);
	return VO_TRUE;
  }
  return VO_NOTIMPL;
}
