/* MplayerXP (C) 2000-2002. by A'rpi/ESP-team (C) 2002. by Nick Kurshev */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>

#include "version.h"
#include "config.h"
#include "sig_hand.h"
#include "mp_msg.h"

#define HELP_MP_DEFINE_STATIC
#include "help_mp.h"

#include "cfgparser.h"
#include "cfg-mplayer-def.h"

#ifdef USE_SUB
#include "subreader.h"
#endif

#include "libvo/video_out.h"
extern void* mDisplay; // Display* mDisplay;

#include "libvo/sub.h"

#include "libao2/audio_out.h"
#include "libao2/audio_plugin.h"

#include "codec-cfg.h"

#include "dvdauth.h"
#ifdef USE_DVDNAV
#include <dvdnav.h>
#endif
#include "spudec.h"
#include "vobsub.h"

#include "osdep/getch2.h"
#include "osdep/keycodes.h"
#include "osdep/timer.h"
#include "osdep/shmem.h"

#include "cpudetect.h"
#include "mm_accel.h"

#include "input/input.h"

int slave_mode=0;
int verbose=0;

#define ABS(x) (((x)>=0)?(x):(-(x)))

#ifdef HAVE_RTC
#include <linux/rtc.h>
#endif

#ifdef USE_TV
#include "libmpdemux/tv.h"

extern int tv_param_on;
extern tvi_handle_t *tv_handler;
#endif


/**************************************************************************
             Playtree
**************************************************************************/
#include "playtree.h"

play_tree_t* playtree;

#define PT_NEXT_ENTRY 1
#define PT_PREV_ENTRY -1
#define PT_NEXT_SRC 2
#define PT_PREV_SRC -2
#define PT_UP_NEXT 3
#define PT_UP_PREV -3

/**************************************************************************
             Config
**************************************************************************/

m_config_t* mconfig;

/**************************************************************************
             Decoding ahead
**************************************************************************/
#include "dec_ahead.h"

int enable_xp=1;
volatile int dec_ahead_active_frame=0;
volatile unsigned abs_dec_ahead_active_frame=0;
volatile unsigned loc_dec_ahead_active_frame=0;
int xp_drop_frame=0;
unsigned xp_n_frame_to_drop=0;
volatile unsigned xp_drop_frame_cnt=0;
unsigned xp_num_frames=0;
extern volatile int xp_eof;
demux_stream_t *d_video=NULL;
int osd_show_av_delay = 0;
int osd_show_sub_delay = 0;
int osd_function=OSD_PLAY;
int output_quality=0;

int xp_id=0;
pid_t mplayer_pid;

/*
   Acceleration for codecs
*/
unsigned mplayer_accel=0;

/**************************************************************************
             Config file
**************************************************************************/

static int cfg_inc_verbose(struct config *conf){ ++verbose; return 0;}

static int cfg_include(struct config *conf, char *filename){
	return m_config_parse_config_file(mconfig, filename);
}

#include "get_path.c"

/**************************************************************************
             Input media streaming & demultiplexer:
**************************************************************************/

static int max_framesize=0;

#include "stream.h"
#include "demuxer.h"
#include "stheader.h"
#include "parse_es.h"

#include "libmpcodecs/dec_video.h"
#include "libmpcodecs/dec_audio.h"

/* Common FIFO functions, and keyboard/event FIFO code */
#include "fifo.c"
int use_stdin=0;
/**************************************************************************/

int vo_inited=0;
ao_functions_t *audio_out=NULL;

/* benchmark: */
double video_time_usage=0;
double vout_time_usage=0;
static double audio_time_usage=0;
static int total_time_usage_start=0;
int benchmark=0;
static unsigned bench_dropped_frames=0;
double max_video_time_usage=0;
double cur_video_time_usage=0;
double min_video_time_usage=0;
double max_vout_time_usage=0;
double cur_vout_time_usage=0;
double min_vout_time_usage=0;
static double max_audio_time_usage=0;
static double cur_audio_time_usage=0;
static double min_audio_time_usage=0;

/* quality's options: */
int auto_quality=0;

int osd_level=2;

// seek:
char *seek_to_sec=NULL;
off_t seek_to_byte=0;
off_t step_sec=0;
int loop_times=-1;
float rel_seek_secs=0;
int abs_seek_pos=0;

/* codecs: */
int has_audio=1;
int has_video=1;
char *audio_codec=NULL;  /* override audio codec */
char *video_codec=NULL;  /* override video codec */
char *audio_family=NULL; /* override audio codec family */
char *video_family=NULL; /* override video codec family */

// streaming:
int audio_id=-1;
int video_id=-1;
int dvdsub_id=-1;
int vobsub_id=-1;
char* audio_lang=NULL;
char* dvdsub_lang=NULL;
static char* spudec_ifo=NULL;
static int vcd_track=0;

// cache2:
static int stream_cache_size=0;
#ifdef USE_STREAM_CACHE
extern int cache_fill_status;
#else
#define cache_fill_status 0
#endif

// dump:
static char *stream_dump_name="stream.dump";
static int stream_dump_type=0;

// A-V sync:
static float default_max_pts_correction=-1;//0.01f;
static float max_pts_correction=0;//default_max_pts_correction;
static float c_total=0;
static float audio_delay=0;

static int dapsync=0;
static int softsleep=0;

static float force_fps=0;
static int force_srate=0;
static int frame_dropping=0; // option  0=no drop  1= drop vo  2= drop decode
static int play_n_frames=-1;
static uint32_t our_n_frames=0;

// screen info:
char* video_driver=NULL; //"mga"; // default
char* audio_driver=NULL;

// libvo opts: (defiend at libmpcodecs/vd.c)
extern int opt_screen_size_x;
extern int opt_screen_size_y;
extern int screen_size_xy;
extern float movie_aspect;
extern int fullscreen;
extern int vidmode;
extern int softzoom;
extern int flip;
extern int vo_flags;

// sub:
char *font_name=NULL;
float font_factor=0.75;
char *sub_name=NULL;
float sub_delay=0;
float sub_fps=0;
int   sub_auto = 1;
char *vobsub_name=NULL;
/*DSP!!char *dsp=NULL;*/

extern char *vo_subdevice;
extern char *ao_subdevice;

static stream_t* stream=NULL;

//char* current_module=NULL; // for debugging

#ifdef HAVE_RTC
int nortc;
#endif

static unsigned int inited_flags=0;
#define INITED_VO	0x00000001
#define INITED_AO	0x00000002
#define INITED_GETCH2	0x00000004
#define INITED_LIRC	0x00000008
#define INITED_SPUDEC	0x00000010
#define INITED_STREAM	0x00000020
#define INITED_INPUT	0x00000040
#define INITED_ALL	0xFFFFFFFF

#undef MSG_ERR
#undef MSG_FATAL
#undef MSG_INFO
#undef MSG_WARN
#undef MSG_OK
#undef MSG_HINT
#undef MSG_V
#undef MSG_STATUS
#undef MSG_DBG2
#define MSG_ERR(args...) mp_msg(MSGT_CPLAYER,MSGL_ERR, ##args )
#define MSG_FATAL(args...) mp_msg(MSGT_CPLAYER,MSGL_FATAL, ##args )
#define MSG_INFO(args...) mp_msg(MSGT_CPLAYER,MSGL_INFO, ##args )
#define MSG_WARN(args...) mp_msg(MSGT_CPLAYER,MSGL_WARN, ##args )
#define MSG_OK(args...) mp_msg(MSGT_CPLAYER,MSGL_OK, ##args )
#define MSG_HINT(args...) mp_msg(MSGT_CPLAYER,MSGL_HINT, ##args )
#define MSG_V(args...) mp_msg(MSGT_CPLAYER,MSGL_V, ##args )
#define MSG_STATUS(args...) mp_msg(MSGT_CPLAYER,MSGL_STATUS, ##args )
#define MSG_DBG2(args...) mp_msg(MSGT_CPLAYER,MSGL_DBG2, ##args )

void uninit_player(unsigned int mask){
  fflush(stdout);
  fflush(stderr);
  mask=inited_flags&mask;
  if(enable_xp) 
  {
    pinfo[xp_id].current_module="uninit_xp";
    uninit_dec_ahead();
  }
  if (mask&INITED_SPUDEC){
    inited_flags&=~INITED_SPUDEC;
    pinfo[xp_id].current_module="uninit_spudec";
    spudec_free(vo_spudec);
    vo_spudec=NULL;
  }
  if(mask&INITED_VO){
    inited_flags&=~INITED_VO;
    pinfo[xp_id].current_module="uninit_vo";
    vo_uninit();
  }

  if(mask&INITED_AO){
    inited_flags&=~INITED_AO;
    pinfo[xp_id].current_module="uninit_ao";
    audio_out->uninit();
  }

  if(mask&INITED_GETCH2){
    inited_flags&=~INITED_GETCH2;
    pinfo[xp_id].current_module="uninit_getch2";
  // restore terminal:
    getch2_disable();
  }

  if(mask&INITED_STREAM){
    inited_flags&=~INITED_STREAM;
    pinfo[xp_id].current_module="uninit_stream";
    if(stream) free_stream(stream);
    stream=NULL;
  }

  if(mask&INITED_INPUT){
    inited_flags&=~INITED_INPUT;
    pinfo[xp_id].current_module="uninit_input";
    mp_input_uninit();
  }

  pinfo[xp_id].current_module=NULL;

}

void exit_player(char* how){

  fflush(stdout);
  fflush(stderr);
  uninit_player(INITED_ALL);

  pinfo[xp_id].current_module="exit_player";

  if(how) MSG_HINT(MSGTR_Exiting,how);
  MSG_DBG2("max framesize was %d bytes\n",max_framesize);
  mp_msg_uninit();
  if(how) exit(0);
  return; /* Still try coredump!!!*/
}

void exit_sighandler(void)
{
  static int sig_count=0;
  ++sig_count;
  if(sig_count==2) exit(1);
  if(sig_count>2){
    // can't stop :(
    kill(getpid(),SIGKILL);
  }
  if(!enable_xp || getpid()==mplayer_pid) /* Is there a reason for not exit only thread here? */
      exit_player(NULL);
}

//extern void write_avi_header_1(FILE *f,int fcc,float fps,int width,int height);

extern void mp_register_options(m_config_t* cfg);

#include "mixer.h"
#include "cfg-mplayer.h"

void parse_cfgfiles( m_config_t* conf )
{
char *conffile;
int conffile_fd;
if (m_config_parse_config_file(conf, "/etc/mplayer.conf") < 0)
  exit(1);
if ((conffile = get_path("")) == NULL) {
  MSG_WARN(MSGTR_NoHomeDir);
} else {
  mkdir(conffile, 0777);
  free(conffile);
  if ((conffile = get_path("config")) == NULL) {
    MSG_ERR(MSGTR_GetpathProblem);
  } else {
    if ((conffile_fd = open(conffile, O_CREAT | O_EXCL | O_WRONLY, 0666)) != -1) {
      MSG_INFO(MSGTR_CreatingCfgFile, conffile);
      write(conffile_fd, default_config, strlen(default_config));
      close(conffile_fd);
    }
    if (m_config_parse_config_file(conf, conffile) < 0)
      exit(1);
    free(conffile);
  }
}
}

// When libmpdemux perform a blocking operation (network connection or cache filling)
// if the operation fail we use this function to check if it was interrupted by the user.
// The function return a new value for eof.
static int libmpdemux_was_interrupted(int eof)
{
  mp_cmd_t* cmd;
  if((cmd = mp_input_get_cmd(0,0)) != NULL) {
       switch(cmd->id) {
       case MP_CMD_QUIT:
	 exit_player(MSGTR_Exit_quit);
       case MP_CMD_PLAY_TREE_STEP: {
	 eof = (cmd->args[0].v.i > 0) ? PT_NEXT_ENTRY : PT_PREV_ENTRY;
       } break;
       case MP_CMD_PLAY_TREE_UP_STEP: {
	 eof = (cmd->args[0].v.i > 0) ? PT_UP_NEXT : PT_UP_PREV;
       } break;	  
       case MP_CMD_PLAY_ALT_SRC_STEP: {
	 eof = (cmd->args[0].v.i > 0) ?  PT_NEXT_SRC : PT_PREV_SRC;
       } break;
       }
       mp_cmd_free(cmd);
  }
  return eof;
}

#ifdef ARCH_X86
static void get_mmx_optimizations( void )
{
  GetCpuCaps(&gCpuCaps);
  MSG_INFO("CPUflags: Type: %d MMX: %d MMX2: %d 3DNow: %d 3DNow2: %d SSE: %d SSE2: %d\n",
      gCpuCaps.cpuType,gCpuCaps.hasMMX,gCpuCaps.hasMMX2,
      gCpuCaps.has3DNow, gCpuCaps.has3DNowExt,
      gCpuCaps.hasSSE, gCpuCaps.hasSSE2);
#ifdef RUNTIME_CPUDETECT
  MSG_INFO("Compiled with RUNTIME CPU Detection - warning, it's untested!\n");
#else
  MSG_INFO("Compiled for x86 CPU with features:");
#ifdef HAVE_MMX
  MSG_INFO(" MMX");
#endif
#ifdef HAVE_MMX2
  MSG_INFO(" MMX2");
#endif
#ifdef HAVE_3DNOW
  MSG_INFO(" 3DNow");
#endif
#ifdef HAVE_3DNOWEX
  MSG_INFO(" 3DNowEx");
#endif
#ifdef HAVE_SSE
  MSG_INFO(" SSE");
#endif
  MSG_INFO("\n");
#endif
  if(gCpuCaps.hasMMX) 		mplayer_accel |= MM_ACCEL_X86_MMX;
  if(gCpuCaps.hasMMX2) 		mplayer_accel |= MM_ACCEL_X86_MMXEXT;
  if(gCpuCaps.hasSSE) 		mplayer_accel |= MM_ACCEL_X86_SSE;
  if(gCpuCaps.has3DNow) 	mplayer_accel |= MM_ACCEL_X86_3DNOW;
  if(gCpuCaps.has3DNowExt) 	mplayer_accel |= MM_ACCEL_X86_3DNOWEXT;
}
#endif

static void init_player( void )
{
  unsigned i;
    if(video_driver && strcmp(video_driver,"help")==0)
    {
	vo_print_help();
	exit(0);
    }

    if(audio_driver && strcmp(audio_driver,"help")==0){
      MSG_INFO("Available audio output drivers:\n");
      i=0;
      while (audio_out_drivers[i]) {
        const ao_info_t *info = audio_out_drivers[i++]->info;
	MSG_INFO("\t%s\t%s\n", info->short_name, info->name);
      }
      MSG_INFO("\n");
      exit(0);
    }

    /* check codec.conf*/
    if(!parse_codec_cfg(get_path("codecs.conf"))){
      if(!parse_codec_cfg(CONFDIR"/codecs.conf")){
        MSG_HINT(MSGTR_CopyCodecsConf);
        exit(0);
      }
    }

    if(audio_codec && strcmp(audio_codec,"help")==0){
      MSG_INFO("Available audio codecs:\n");
      list_codecs(1);
      MSG_INFO("\n");
      exit(0);
    }
    if(video_codec && strcmp(video_codec,"help")==0){
      MSG_INFO("Available video codecs:\n");
      list_codecs(0);
      MSG_INFO("\n");
      exit(0);
    }
}

static demuxer_t *demuxer=NULL;

static demux_stream_t *d_audio=NULL;
static demux_stream_t *d_dvdsub=NULL;

static sh_audio_t *sh_audio=NULL;
static sh_video_t *sh_video=NULL;

int decore_audio( void )
{
  int eof = 0;
/*========================== PLAY AUDIO ============================*/
while(sh_audio){
  unsigned int t;
  double tt;
  int playsize;
  
  ao_data.pts=sh_audio->timer*90000.0;
  playsize=audio_out->get_space();
  
  if(!playsize) {
    if(sh_video)
      break; // buffer is full, do not block here!!!
    usec_sleep(10000); // Wait a tick before retry
    continue;
  }
  
  if(playsize>MAX_OUTBURST) playsize=MAX_OUTBURST; // we shouldn't exceed it!
  //if(playsize>outburst) playsize=outburst;

  // Update buffer if needed
  pinfo[xp_id].current_module="decode_audio";   // Enter AUDIO decoder module
  t=GetTimer();
  while(sh_audio->a_buffer_len<playsize && !d_audio->eof){
    int ret=decode_audio(sh_audio,&sh_audio->a_buffer[sh_audio->a_buffer_len],
        playsize-sh_audio->a_buffer_len,sh_audio->a_buffer_size-sh_audio->a_buffer_len);
    if(ret>0) sh_audio->a_buffer_len+=ret; 
    else {
      if(!sh_video)
	eof = PT_NEXT_ENTRY;
      break;
    }
  }
  pinfo[xp_id].current_module="play_audio";   // Leave AUDIO decoder module
  t=GetTimer()-t;
  tt = t*0.000001f;
  audio_time_usage+=tt;
  if(benchmark)
  {
    if(tt > max_audio_time_usage) max_audio_time_usage = tt;
    if(tt < min_audio_time_usage) min_audio_time_usage = tt;
    cur_audio_time_usage=tt;
  }
  if(playsize>sh_audio->a_buffer_len) playsize=sh_audio->a_buffer_len;
  
  playsize=audio_out->play(sh_audio->a_buffer,playsize,0);

  if(playsize>0){
      sh_audio->a_buffer_len-=playsize;
      memcpy(sh_audio->a_buffer,&sh_audio->a_buffer[playsize],sh_audio->a_buffer_len);
      sh_audio->timer+=playsize/(float)(sh_audio->o_bps);
  }

  break;
 } // if(sh_audio)
 return eof;
}

typedef struct
{
    int drop_frame_cnt;
    int too_slow_frame_cnt;
    int too_fast_frame_cnt;
}video_stat_t;

#ifdef USE_OSD
void update_osd( float v_pts )
{
static char osd_text_buffer[64];
static int osd_last_pts=-303;
//================= Update OSD ====================
  if(osd_level>=2){
      int pts=v_pts;
      char osd_text_tmp[64];
      if(pts==osd_last_pts-1) ++pts; else osd_last_pts=pts;
      vo_osd_text=osd_text_buffer;
#ifdef USE_DVDNAV
      if (osd_show_dvd_nav_delay) {
          sprintf(osd_text_tmp, "DVDNAV: %s", dvd_nav_text);
          osd_show_dvd_nav_delay--;
      } else
#endif
      if (osd_show_sub_delay) {
	  sprintf(osd_text_tmp, "Sub delay: %d ms",(int)(sub_delay*1000));
	  osd_show_sub_delay--;
      } else
      if (osd_show_av_delay) {
	  sprintf(osd_text_tmp, "A-V delay: %d ms",(int)(audio_delay*1000));
	  osd_show_av_delay--;
      } else
#ifdef ENABLE_DEC_AHEAD_DEBUG
          if(verbose) sprintf(osd_text_tmp,"%c %02d:%02d:%02d abs frame: %lu",osd_function,pts/3600,(pts/60)%60,pts%60,abs_dec_ahead_active_frame);
	  else sprintf(osd_text_tmp,"%c %02d:%02d:%02d",osd_function,pts/3600,(pts/60)%60,pts%60);
#else
          sprintf(osd_text_tmp,"%c %02d:%02d:%02d",osd_function,pts/3600,(pts/60)%60,pts%60);
#endif
      if(strcmp(vo_osd_text, osd_text_tmp)) {
	      strcpy(vo_osd_text, osd_text_tmp);
	      vo_osd_changed(OSDTYPE_OSD);
      }
  } else {
      if(vo_osd_text) {
      vo_osd_text=NULL;
	  vo_osd_changed(OSDTYPE_OSD);
      }
  }
}
#endif

int mp09_decore_video( int rtc_fd, video_stat_t *vstat, float *aq_sleep_time, float *v_pts )
{
/*========================== PLAY VIDEO ============================*/

  static float next_frame_time=0;
  static int frame_time_remaining=0; // flag
  static int dropped_frames=0; // how many frames dropped since last non-dropped frame
  static int drop_frame=0;     // current dropping status
  static int total_frame_cnt=0;
  static unsigned int lastframeout_ts=0;
  float frame_time=next_frame_time;
  float time_frame=0;
  float AV_delay=0; // average of A-V timestamp differences
  float time_frame_corr_avg=0;
  int blit_frame=0;
  int delay_corrected=1;
  vo_pts=sh_video->timer*90000.0;
  vo_fps=sh_video->fps;

  if(!frame_time_remaining){
    //--------------------  Decode a frame: -----------------------
    while(1)
    {   unsigned char* start=NULL;
	int in_size;
	// get it!
	pinfo[xp_id].current_module="video_read_frame";
        in_size=video_read_frame(sh_video,&next_frame_time,&start,force_fps);
	if(in_size<0) return 1;
	if(in_size>max_framesize) max_framesize=in_size; // stats
	sh_video->timer+=frame_time;
	time_frame+=frame_time;  // for nosound
	// check for frame-drop:
	pinfo[xp_id].current_module="check_framedrop";
	if(sh_audio && !d_audio->eof){
	    float delay=audio_out->get_delay();
	    float d=(sh_video->timer)-(sh_audio->timer-delay);
	    // we should avoid dropping to many frames in sequence unless we
	    // are too late. and we allow 100ms A-V delay here:
	    if(d<-dropped_frames*frame_time-0.100){
		drop_frame=frame_dropping;
		++vstat->drop_frame_cnt;
		++dropped_frames;
	    } else {
		drop_frame=dropped_frames=0;
	    }
	    ++total_frame_cnt;
	}
	// decode:
	pinfo[xp_id].current_module="decode_video";
	blit_frame=decode_video(sh_video,start,in_size,drop_frame);
	break;
    }
    //------------------------ frame decoded. --------------------

    MSG_DBG2("*** ftime=%5.3f ***\n",frame_time);

#if 0
    if(sh_video->vf_inited<0){
	MSG_FATAL("\nFATAL: Couldn't initialize video filters (-vop) or video output (-vo) !\n");
	eof=1; goto goto_next_file;
    }
#endif
  }

// ==========================================================================
    
    pinfo[xp_id].current_module="draw_osd";
    vo_draw_osd();

    pinfo[xp_id].current_module="calc_sleep_time";

#if 0
{	// debug frame dropping code
	  float delay=audio_out->get_delay();
	  MSG_V("\r[V] %5.3f [A] %5.3f => {%5.3f}  (%5.3f) [%d]   \n",
	      sh_video->timer,sh_audio->timer-delay,
	      sh_video->timer-(sh_audio->timer-delay),
	      delay,drop_frame);
}
#endif

    if(drop_frame && !frame_time_remaining){

      time_frame=0;	// don't sleep!
      blit_frame=0;	// don't display!
      
    } else {

      // It's time to sleep...
      
      frame_time_remaining=0;
      time_frame-=GetRelativeTime(); // reset timer

      if(sh_audio && !d_audio->eof){
	  float delay=audio_out->get_delay();
	  MSG_DBG2("delay=%f\n",delay);

	if(!dapsync){

	      /* Arpi's AV-sync */

          time_frame=sh_video->timer;
          time_frame-=sh_audio->timer-delay;

	} else {  // if(!dapsync)

	      /* DaP's AV-sync */

    	      float SH_AV_delay;
	      /* SH_AV_delay = sh_video->timer - (sh_audio->timer - (float)((float)delay + sh_audio->a_buffer_len) / (float)sh_audio->o_bps); */
	      SH_AV_delay = sh_video->timer - (sh_audio->timer - (float)((float)delay) / (float)sh_audio->o_bps);
	      if(SH_AV_delay<-2*frame_time){
		  static int drop_message=0;
	          drop_frame=frame_dropping; // tricky!
	          ++vstat->drop_frame_cnt;
		  if(vstat->drop_frame_cnt>50 && AV_delay>0.5 && !drop_message){
	    	      drop_message=1;
	    	      MSG_WARN(MSGTR_SystemTooSlow);
	         }
		MSG_INFO("A-V SYNC: FRAMEDROP (SH_AV_delay=%.3f)!\n", SH_AV_delay);
	        MSG_DBG2("\nframe drop %d, %.2f\n", drop_frame, time_frame);
	        /* go into unlimited-TF cycle */
    		time_frame = SH_AV_delay;
	      } else {
#define	SL_CORR_AVG_LEN	125
	        /* don't adjust under framedropping */
	        time_frame_corr_avg = (time_frame_corr_avg * (SL_CORR_AVG_LEN - 1) +
	    				(SH_AV_delay - time_frame)) / SL_CORR_AVG_LEN;
#define	UNEXP_CORR_MAX	0.1	/* limit of unexpected correction between two frames (percentage) */
#define	UNEXP_CORR_WARN	1.0	/* warn limit of A-V lag (percentage) */
	        time_frame += time_frame_corr_avg;
	        if (SH_AV_delay - time_frame < (frame_time + time_frame_corr_avg) * UNEXP_CORR_MAX &&
		    SH_AV_delay - time_frame > (frame_time + time_frame_corr_avg) * -UNEXP_CORR_MAX)
		    time_frame = SH_AV_delay;
	        else {
		    if (SH_AV_delay - time_frame > (frame_time + time_frame_corr_avg) * UNEXP_CORR_WARN ||
		        SH_AV_delay - time_frame < (frame_time + time_frame_corr_avg) * -UNEXP_CORR_WARN)
		        MSG_WARN( "WARNING: A-V SYNC LAG TOO LARGE: %.3f {%.3f - %.3f} (too little UNEXP_CORR_MAX?)\n",
		  	    SH_AV_delay - time_frame, SH_AV_delay, time_frame);
		        time_frame += (frame_time + time_frame_corr_avg) * ((SH_AV_delay > time_frame) ?
		    		      UNEXP_CORR_MAX : -UNEXP_CORR_MAX);
	        }
	      }	/* /start dropframe */

	} // if(dapsync)

	if(delay>0.25) delay=0.25; else
	if(delay<0.10) delay=0.10;
	if(time_frame>delay*0.6){
	    // sleep time too big - may cause audio drops (buffer underrun)
	    frame_time_remaining=1;
	    time_frame=delay*0.5;
	}

      } else {

          // NOSOUND:
          if( (time_frame<-3*frame_time || time_frame>3*frame_time) || benchmark)
	      time_frame=0;
	  
      }

      (*aq_sleep_time)+=time_frame;

    }	// !drop_frame
    
//============================== SLEEP: ===================================

// flag 256 means: libvo driver does its timing (dvb card)
if(time_frame>0.001 && !(vo_flags&256)){

#ifdef HAVE_RTC
    if(rtc_fd>=0){
	// -------- RTC -----------
	pinfo[xp_id].current_module="sleep_rtc";
        while (time_frame > 0.000) {
	    unsigned long long rtc_ts;
	    if (read (rtc_fd, &rtc_ts, sizeof(rtc_ts)) <= 0)
		    MSG_ERR( "Linux RTC read error: %s\n", strerror(errno));
    	    time_frame-=GetRelativeTime();
	}
    } else
#endif
    {
	// -------- USLEEP + SOFTSLEEP -----------
	float min=softsleep?0.021:0.005;
	pinfo[xp_id].current_module="sleep_usleep";
        while(time_frame>min){
          if(time_frame<=0.020)
             usec_sleep(0); // sleeps 1 clock tick (10ms)!
          else
             usec_sleep(1000000*(time_frame-0.020));
          time_frame-=GetRelativeTime();
        }
	if(softsleep){
	    pinfo[xp_id].current_module="sleep_soft";
	    if(time_frame<0) MSG_WARN( "Warning! Softsleep underflow!\n");
	    while(time_frame>0) time_frame-=GetRelativeTime(); // burn the CPU
	}
    }

}

//if(!frame_time_remaining){	// should we display the frame now?

//====================== FLIP PAGE (VIDEO BLT): =========================

        pinfo[xp_id].current_module="flip_page";

	vo_check_events(); /* check events AST */
        if(blit_frame && !frame_time_remaining){
	   unsigned int t2=GetTimer();
	   double tt;
	   float j;
#define	FRAME_LAG_WARN	0.2
	   j = ((float)t2 - lastframeout_ts) / 1000000;
	   lastframeout_ts = GetTimer();
	   if (j < frame_time + frame_time * -FRAME_LAG_WARN)
		vstat->too_fast_frame_cnt++;
		/* printf ("PANIC: too fast frame (%.3f)!\n", j); */
	   else if (j > frame_time + frame_time * FRAME_LAG_WARN)
		vstat->too_slow_frame_cnt++;
		/* printf ("PANIC: too slow frame (%.3f)!\n", j); */

	   vo_flip_page();
//        usec_sleep(50000); // test only!
	   t2=GetTimer()-t2;
	   tt = t2*0.000001f;
	   vout_time_usage+=tt;
	   if(benchmark)
	   {
		/* we need compute draw_slice+flip_page here */
		cur_vout_time_usage+=tt;
		if(cur_vout_time_usage > max_vout_time_usage) max_vout_time_usage = cur_vout_time_usage;
		if(cur_vout_time_usage < min_vout_time_usage) min_vout_time_usage = cur_vout_time_usage;
		if((cur_video_time_usage + cur_vout_time_usage + cur_audio_time_usage)*vo_fps > 1)
							bench_dropped_frames ++;
	   }
	}

//====================== A-V TIMESTAMP CORRECTION: =========================

  pinfo[xp_id].current_module="av_sync";

  if(sh_audio){
    float a_pts=0;

    // unplayed bytes in our and soundcard/dma buffer:
    float delay=audio_out->get_delay()+(float)sh_audio->a_buffer_len/(float)sh_audio->o_bps;

    if(pts_from_bps){
	// PTS = sample_no / samplerate
        unsigned int samples=(sh_audio->audio.dwSampleSize)?
          ((ds_tell(d_audio)-sh_audio->a_in_buffer_len)/sh_audio->audio.dwSampleSize) :
          (d_audio->pack_no); // <- used for VBR audio
	samples+=sh_audio->audio.dwStart; // offset
        a_pts=samples*(float)sh_audio->audio.dwScale/(float)sh_audio->audio.dwRate;
	delay_corrected=1;
    } else {
      // PTS = (last timestamp) + (bytes after last timestamp)/(bytes per sec)
      a_pts=d_audio->pts;
      if(!delay_corrected) if(a_pts) delay_corrected=1;
      a_pts+=(ds_tell_pts(d_audio)-sh_audio->a_in_buffer_len)/(float)sh_audio->i_bps;
    }
    *v_pts=d_video->pts;

      MSG_DBG2("### A:%8.3f (%8.3f)  V:%8.3f  A-V:%7.4f  \n",a_pts,a_pts-audio_delay-delay,*v_pts,(a_pts-delay-audio_delay)-*v_pts);

      if(delay_corrected){
        float x;
	AV_delay=(a_pts-delay-audio_delay)-*v_pts;
        x=AV_delay*0.1f;
        if(x<-max_pts_correction) x=-max_pts_correction; else
        if(x> max_pts_correction) x= max_pts_correction;
        if(default_max_pts_correction>=0)
          max_pts_correction=default_max_pts_correction;
        else
          max_pts_correction=sh_video->frametime*0.10; // +-10% of time
	if(!frame_time_remaining){ sh_audio->timer+=x; c_total+=x;} // correction
        if(benchmark && verbose) MSG_STATUS("A:%6.1f V:%6.1f A-V:%7.3f ct:%7.3f  %3d/%3d  %2d%% %2d%% %4.1f%% %d %d %d%%\r",
	  a_pts-audio_delay-delay,*v_pts,AV_delay,c_total,
          (int)sh_video->num_frames,(int)sh_video->num_frames_decoded,
          (sh_video->timer>0.5)?(int)(100.0*video_time_usage/(double)sh_video->timer):0,
          (sh_video->timer>0.5)?(int)(100.0*vout_time_usage/(double)sh_video->timer):0,
          (sh_video->timer>0.5)?(100.0*audio_time_usage/(double)sh_video->timer):0
          ,vstat->drop_frame_cnt
	  ,output_quality
	  ,cache_fill_status
        );
        fflush(stdout);
      }
    }
    /* let it paints audio timer instead of video */
    if(sh_audio) *v_pts = sh_audio->timer-audio_out->get_delay();
    return 0;
}

int xp_decore_video( int rtc_fd, video_stat_t *vstat, float *aq_sleep_time, float *v_pts )
{
    float time_frame=0;
    float frame_time=0;
    float AV_delay=0; /* average of A-V timestamp differences */
    unsigned int lastframeout_ts=0;
    volatile int da_locked_frame;
    volatile unsigned ada_locked_frame;
    volatile unsigned ada_blitted_frame;
    int blit_frame=0;
    int delay_corrected=1;
    int final_frame=0;
    int num_frames_decoded = 0;
    LOCK_VDEC_ACTIVE();
    vo_get_active_frame(&dec_ahead_active_frame);
    UNLOCK_VDEC_ACTIVE();
    LOCK_VDECA();
    frame_time = shva[dec_ahead_active_frame].frametime;
    final_frame = shva[dec_ahead_active_frame].eof;
    sh_video->num_frames = shva[dec_ahead_active_frame].num_frames;
    sh_video->num_frames_decoded = shva[dec_ahead_active_frame].num_frames_decoded;
    num_frames_decoded = dec_ahead_num_frames_decoded;
    UNLOCK_VDECA();
    if(xp_eof && final_frame) return 1;
    /*------------------------ frame decoded. --------------------*/
    sh_video->timer+=frame_time;
    /* It's time to sleep ;)...*/
    pinfo[xp_id].current_module="sleep";
    GetRelativeTime(); /* reset timer */
    if(sh_audio && !d_audio->eof)
	time_frame=sh_video->timer-(sh_audio->timer-audio_out->get_delay());
    else /* NOSOUND: WARNING: We should to use something like
	    get_timer()-1/fps
	    here but for now it's far enough to use 1/fps only */
	time_frame=frame_time;
    (*aq_sleep_time)=0; /* we have other way to control that */

    if(!(vo_flags&256)){ /* flag 256 means: libvo driver does its timing (dvb card) */
#define XP_MIN_TIMESLICE 0.010 /* under Linux on x86 min time_slice = 10 ms */
    while(time_frame>XP_MIN_TIMESLICE)
    {
	/* free cpu for threads */
	usleep(1);
	time_frame-=GetRelativeTime();
    }
#ifdef HAVE_RTC
	if(rtc_fd>=0){
	    // -------- RTC -----------
	    while (time_frame > 0.000) {
		unsigned long long rtc_ts;
		if (read (rtc_fd, &rtc_ts, sizeof(rtc_ts)) <= 0)
		    perror ("read (rtc_fd)");
    		time_frame-=GetRelativeTime();
	    }
	} else
#endif
	{
	    // -------- USLEEP + SOFTSLEEP -----------
	    float min=softsleep?0.021:0.005;
	    while(time_frame>min){
		if(time_frame<=0.020)
		    usec_sleep(0); // sleeps 1 clock tick (10ms)!
		else
		    usec_sleep(1000000*(time_frame-0.020));
		time_frame-=GetRelativeTime();
	    }
	    if(softsleep){
		if(time_frame<0) MSG_WARN("Warning! softsleep underflow!\n");
		while(time_frame>0) time_frame-=GetRelativeTime(); // burn the CPU
	    }
	}
    }

    pinfo[xp_id].current_module="flip_page";
    vo_check_events();
/* blit frame */
    LOCK_VDEC_LOCKED();
    da_locked_frame = dec_ahead_locked_frame;
    ada_locked_frame = abs_dec_ahead_locked_frame;
    ada_blitted_frame = abs_dec_ahead_blitted_frame;
    UNLOCK_VDEC_LOCKED();
    /* don't flip if there is nothing new to display */
    blit_frame=!(
    dec_ahead_active_frame == da_locked_frame || /* like: we are in lseek */
     /* don't flip if frame which is being decoded is equal to frame
	which is being displayed or will be displayed.
	Many card can perform page flipping only during VSYNC
	so we can't flip to locked frame too */
    (dec_ahead_active_frame+1)%xp_num_frames == da_locked_frame ||
    (dec_ahead_active_frame-1+xp_num_frames)%xp_num_frames == da_locked_frame);
    if(xp_eof) blit_frame=1; /* force blitting until end of stream will be reached */
    if(!blit_frame)
    {
        static int drop_message=0;
        if(!drop_message &&
	    abs_dec_ahead_active_frame > 50 &&
	    ada_locked_frame <= abs_dec_ahead_active_frame - 50)
	    {
		drop_message=1;
		MSG_WARN(MSGTR_SystemTooSlow);
		MSG_D("\ndec_ahead_main: TOO SLOW (locked=%u active=%u)\n"
		,ada_locked_frame,abs_dec_ahead_active_frame);
	    }
	    MSG_D("\ndec_ahead_main: stalling %u (locked: %u blitted: %u local: %u)\n",dec_ahead_active_frame,da_locked_frame,ada_blitted_frame,loc_dec_ahead_active_frame);
    }
    else
    {
	unsigned int t2=GetTimer();
	double tt;
	float j;
#define	FRAME_LAG_WARN	0.2
	j = ((float)t2 - lastframeout_ts) / 1000000;
	lastframeout_ts = GetTimer();
	if (j < frame_time + frame_time * -FRAME_LAG_WARN)
	    vstat->too_fast_frame_cnt++;
		/* printf ("PANIC: too fast frame (%.3f)!\n", j); */
	else if (j > frame_time + frame_time * FRAME_LAG_WARN)
	    vstat->too_slow_frame_cnt++;
		/* printf ("PANIC: too slow frame (%.3f)!\n", j); */
	vo_flip_page();
	MSG_D("\ndec_ahead_main: schedule %u on screen (abs: %u local: %u blitted %u)\n",dec_ahead_active_frame,abs_dec_ahead_active_frame,loc_dec_ahead_active_frame,ada_blitted_frame);
	LOCK_VDEC_ACTIVE();
	dec_ahead_active_frame=(dec_ahead_active_frame+1)%xp_num_frames;
	loc_dec_ahead_active_frame++;
	UNLOCK_VDEC_ACTIVE();
#ifdef USE_OSD
	/*--------- add OSD to the next frame contents ---------*/
	{
	    float video_pts;
	    MSG_D("dec_ahead_main: draw_osd to %u\n",dec_ahead_active_frame);
	    pinfo[xp_id].current_module="draw_osd";
            video_pts = d_video->pts;
            if (sh_video) {
                /* paint video timer */
                LOCK_VDECA();
                video_pts = shva[dec_ahead_active_frame].v_pts;
                UNLOCK_VDECA();
            } else {
                /* let it paints audio timer instead of video */
                if(sh_audio) {
                    float a_pts=0;
                    if(pts_from_bps){
                        unsigned int samples=(sh_audio->audio.dwSampleSize)?
                            ((ds_tell(d_audio)-sh_audio->a_in_buffer_len)/sh_audio->audio.dwSampleSize) :
                            (d_audio->pack_no); // <- used for VBR audio
                        samples+=sh_audio->audio.dwStart; // offset
                        a_pts=samples*(float)sh_audio->audio.dwScale/(float)sh_audio->audio.dwRate;
                    } else {
                        // PTS = (last timestamp) + (bytes after last timestamp)/(bytes per sec)
                        a_pts=d_audio->pts;
                        a_pts+=(ds_tell_pts(d_audio)-sh_audio->a_in_buffer_len)/(float)sh_audio->i_bps;
                    }
                    video_pts = a_pts-audio_out->get_delay();
                }
            }
	    update_osd(video_pts);
	    vo_draw_osd();
	}
#endif
	t2=GetTimer()-t2;
	tt = t2*0.000001f;
	vout_time_usage+=tt;
	if(benchmark)
	{
		/* we need compute draw_slice+flip_page here */
		cur_vout_time_usage+=tt;
		if((cur_video_time_usage + cur_vout_time_usage + cur_audio_time_usage)*vo_fps > 1)
							bench_dropped_frames ++;
	}
    }
    LOCK_VDEC_ACTIVE();
    abs_dec_ahead_active_frame++;
    UNLOCK_VDEC_ACTIVE();
    /* ----- compute frame dropping ----------*/
    {
	unsigned n_frames_to_drop=0;
	static unsigned prev_n_fr=0;
	volatile int da_locked_frame;
	volatile unsigned ada_locked_frame;
	LOCK_VDEC_LOCKED();
	da_locked_frame = dec_ahead_locked_frame;
	ada_locked_frame = abs_dec_ahead_locked_frame;
	UNLOCK_VDEC_LOCKED();
	n_frames_to_drop=0;
	/* note: Arpi allows 100 ms of delay before frame dropping.
		 we will allows 6 frames of delay (that's the for 25 fps) */
	if(frame_dropping &&
	   abs_dec_ahead_active_frame+xp_num_frames-6 > ada_locked_frame &&
	   !xp_eof)
	 n_frames_to_drop = abs_dec_ahead_active_frame+xp_num_frames-ada_locked_frame;
#ifdef ENABLE_DEC_AHEAD_DEBUG
	if(n_frames_to_drop)
	{
		MSG_D("\n**** FRAME DROPPING ****** %u(%u) > %u\n"
		,abs_dec_ahead_active_frame
		,abs_dec_ahead_active_frame+xp_num_frames
		,ada_locked_frame);
	}
#endif
	if(n_frames_to_drop != prev_n_fr)
	{
	    LOCK_VDECODING();
	    xp_n_frame_to_drop=n_frames_to_drop;
	    UNLOCK_VDECODING();
	    prev_n_fr = n_frames_to_drop;
	}
    }
    pinfo[xp_id].current_module=NULL;
    
/*================ A-V TIMESTAMP CORRECTION: =========================*/
  if(sh_audio){
    float a_pts=0;

    // unplayed bytes in our and soundcard/dma buffer:
    float delay=audio_out->get_delay()+(float)sh_audio->a_buffer_len/(float)sh_audio->o_bps;

    if(pts_from_bps){
        unsigned int samples=(sh_audio->audio.dwSampleSize)?
          ((ds_tell(d_audio)-sh_audio->a_in_buffer_len)/sh_audio->audio.dwSampleSize) :
          (d_audio->pack_no); // <- used for VBR audio
	samples+=sh_audio->audio.dwStart; // offset
        a_pts=samples*(float)sh_audio->audio.dwScale/(float)sh_audio->audio.dwRate;
      delay_corrected=1;
    } else {
      // PTS = (last timestamp) + (bytes after last timestamp)/(bytes per sec)
      a_pts=d_audio->pts;
      if(!delay_corrected) if(a_pts) delay_corrected=1;
      a_pts+=(ds_tell_pts(d_audio)-sh_audio->a_in_buffer_len)/(float)sh_audio->i_bps;
    }
    LOCK_VDECA();
    *v_pts = shva[dec_ahead_active_frame].v_pts;
    UNLOCK_VDECA();

      MSG_DBG2("### A:%8.3f (%8.3f)  V:%8.3f  A-V:%7.4f  \n",a_pts,a_pts-audio_delay-delay,*v_pts,(a_pts-delay-audio_delay)-*v_pts);

      if(delay_corrected){
        float x;
	AV_delay=(a_pts-delay-audio_delay)-*v_pts;
        x=AV_delay*0.1f;
        if(x<-max_pts_correction) x=-max_pts_correction; else
        if(x> max_pts_correction) x= max_pts_correction;
        if(default_max_pts_correction>=0)
          max_pts_correction=default_max_pts_correction;
        else
	{
	    LOCK_VDECA();
	    max_pts_correction=shva[dec_ahead_active_frame].frametime*0.10; // +-10% of time
	    UNLOCK_VDECA();
	}
        sh_audio->timer+=x; c_total+=x;
        if(benchmark && verbose)
	    MSG_STATUS("A:%6.1f V:%6.1f A-V:%7.3f ct:%7.3f  %3d/%3d  %2d%% %2d%% %4.1f%% %d %d %d%%\r",
	    a_pts-audio_delay-delay,*v_pts,AV_delay,c_total,
            (int)sh_video->num_frames,num_frames_decoded,
            (sh_video->timer>0.5)?(int)(100.0*video_time_usage/(double)sh_video->timer):0,
            (sh_video->timer>0.5)?(int)(100.0*vout_time_usage/(double)sh_video->timer):0,
            (sh_video->timer>0.5)?(100.0*audio_time_usage/(double)sh_video->timer):0
            ,vstat->drop_frame_cnt
	    ,output_quality
	    ,cache_fill_status
    	    );
        fflush(stdout);
      }
    
  } else {
    // No audio:
    
    if(benchmark && verbose)
	MSG_STATUS("V:%6.1f  %3d  %2d%% %2d%% %4.1f%% %d %d %d%%\r",*v_pts,
        (int)sh_video->num_frames,
        (sh_video->timer>0.5)?(int)(100.0*video_time_usage/(double)sh_video->timer):0,
        (sh_video->timer>0.5)?(int)(100.0*vout_time_usage/(double)sh_video->timer):0,
        (sh_video->timer>0.5)?(100.0*audio_time_usage/(double)sh_video->timer):0
          ,vstat->drop_frame_cnt
	  ,output_quality
	  ,cache_fill_status
        );

      fflush(stdout);

  }
  return 0;
}

void mpxp_seek( int xp_id, video_stat_t *vstat, int *osd_visible, float v_pts )
{
  if(demux_seek(demuxer,rel_seek_secs,abs_seek_pos)){
      if(enable_xp)
      { // Send back frame info to decoding thread
          dec_ahead_seek_num_frames = sh_video->num_frames;
          dec_ahead_seek_num_frames_decoded = sh_video->num_frames_decoded;
      }

      // success:
      /* FIXME there should be real seeking for vobsub */
      if (vo_vobsub)
	vobsub_reset(vo_vobsub);

      if(sh_audio){
	if(verbose){
	    float a_pts=d_audio->pts;
            a_pts+=(ds_tell_pts(d_audio)-sh_audio->a_in_buffer_len)/(float)sh_audio->i_bps;
	    MSG_V("SEEK: A: %5.3f  V: %5.3f  A-V: %5.3f   \n",a_pts,v_pts,a_pts-v_pts);
	}
        MSG_STATUS("A:%6.1f  V:%6.1f  A-V:%7.3f  ct: ?   \r",d_audio->pts,v_pts,0.0f);
      } else {
        MSG_STATUS("A: ---   V:%6.1f   \r",v_pts);
      }
      fflush(stdout);

      if(sh_video){
	 pinfo[xp_id].current_module="seek_video_reset";
         vo_reset();
      }
      
      if(sh_audio){
        pinfo[xp_id].current_module="seek_audio_reset";
        audio_out->reset(); // stop audio, throwing away buffered data
      }
#ifdef USE_OSD
        // Set OSD:
      if(osd_level){
        int len=((demuxer->movi_end-demuxer->movi_start)>>8);
        if (len>0){
	   if(osd_visible) *osd_visible=sh_video->fps; // 1 sec
	   vo_osd_progbar_type=0;
	   vo_osd_progbar_value=(demuxer->filepos-demuxer->movi_start)/len;
	   vo_osd_changed(OSDTYPE_PROGBAR);
	}
      }
#endif
      if(sh_video) {
	c_total=0;
	max_pts_correction=0.1;
	if(osd_visible) *osd_visible=sh_video->fps; // to rewert to PLAY pointer after 1 sec
	audio_time_usage=0; video_time_usage=0; vout_time_usage=0;
	if(vstat)
	{
	    vstat->drop_frame_cnt=0;
	    vstat->too_slow_frame_cnt=0;
	    vstat->too_fast_frame_cnt=0;
	}
        if(vo_spudec) spudec_reset(vo_spudec);
      }
  }
}

int v_hw_equ_cap=0;
int v_bright=50;
int v_cont=50;
int v_hue=50;
int v_saturation=50;
/*
For future:
int v_red_intensity=0;
int v_green_intensity=0;
int v_blue_intensity=0;
*/


int main(int argc,char* argv[], char *envp[]){

#ifdef USE_SUB
static subtitle* subtitles=NULL;
float sub_last_pts = -303;
#endif

// for multifile support:
play_tree_iter_t* playtree_iter = NULL;

char* filename=NULL; //"MI2-Trailer.avi";
int file_format=DEMUXER_TYPE_UNKNOWN;
int (*decore_video_ptr)( int rtc_fd, video_stat_t *vstat, float *aq_sleep_time, float *v_pts )
			 = mp09_decore_video;

// movie info:
int eof=0;

int osd_visible=100;
video_stat_t vstat;
int rtc_fd=-1;
int i;

    mplayer_pid=
    pinfo[xp_id].pid=getpid();
    pinfo[xp_id].thread_name = "main";
    memset(&vstat,0,sizeof(video_stat_t));
//vo_tune_info_t vtune;

  mp_msg_init(MSGL_STATUS);

  MSG_INFO("%s",banner_text);
//  memset(&vtune,0,sizeof(vo_tune_info_t));
  /* Test for cpu capabilities (and corresponding OS support) for optimizing */
#ifdef ARCH_X86
  get_mmx_optimizations();
#endif

    playtree = play_tree_new();

    mconfig = m_config_new(playtree);
    m_config_register_options(mconfig,mplayer_opts);
    // TODO : add something to let modules register their options
    mp_register_options(mconfig);
    parse_cfgfiles(mconfig);

    if(m_config_parse_command_line(mconfig, argc, argv, envp) < 0) exit(1); // error parsing cmdline

    playtree = play_tree_cleanup(playtree);
    if(playtree) {
      playtree_iter = play_tree_iter_new(playtree,mconfig);
      if(playtree_iter) {  
	if(play_tree_iter_step(playtree_iter,0,0) != PLAY_TREE_ITER_ENTRY) {
	  play_tree_iter_free(playtree_iter);
	  playtree_iter = NULL;
	}
	filename = play_tree_iter_get_file(playtree_iter,1);
      }
    }

    if(enable_xp) vo_doublebuffering=1;
    else	  vo_use_bm = 0;
	
    init_player();

    if(!filename && !vcd_track && !dvd_title && !dvd_nav && !tv_param_on){
	// no file/vcd/dvd -> show HELP:
	MSG_INFO("%s",help_text);
	exit(0);
    }

    // Many users forget to include command line in bugreports...
    if(verbose){
      MSG_INFO("CommandLine:");
      for(i=1;i<argc;i++) MSG_INFO(" '%s'",argv[i]);
      MSG_INFO("\n");
    }

    mp_msg_init(verbose+MSGL_STATUS);

//------ load global data first ------


// check font
#ifdef USE_OSD
  if(font_name){
       vo_font=read_font_desc(font_name,font_factor,verbose>1);
       if(!vo_font) MSG_ERR(MSGTR_CantLoadFont,font_name);
  } else {
      // try default:
       vo_font=read_font_desc(get_path("font/font.desc"),font_factor,verbose>1);
       if(!vo_font)
       vo_font=read_font_desc(DATADIR"/font/font.desc",font_factor,verbose>1);
  }
#endif

  pinfo[xp_id].current_module="init_osd";
  vo_init_osd();
#ifdef HAVE_RTC
  if(!nortc)
  {
    if ((rtc_fd = open("/dev/rtc", O_RDONLY)) < 0)
	perror ("Linux RTC init: open");
    else {
	unsigned long irqp;

	/* if (ioctl(rtc_fd, RTC_IRQP_SET, _) < 0) { */
	if (ioctl(rtc_fd, RTC_IRQP_READ, &irqp) < 0) {
    	    perror ("Linux RTC init: ioctl (rtc_irqp_read)");
    	    close (rtc_fd);
    	    rtc_fd = -1;
	} else if (ioctl(rtc_fd, RTC_PIE_ON, 0) < 0) {
	    /* variable only by the root */
    	    perror ("Linux RTC init: ioctl (rtc_pie_on)");
    	    close (rtc_fd);
	    rtc_fd = -1;
	} else
	    MSG_V("Using Linux's hardware RTC timing (%ldHz)\n", irqp);
    }
  }
  if(rtc_fd<0)
#endif
	MSG_V("Using %s timing\n",softsleep?"software":"usleep()");

#ifdef USE_TERMCAP
  load_termcap(NULL); // load key-codes
#endif

// ========== Init keyboard FIFO (connection to libvo) ============
make_pipe(&keyb_fifo_get,&keyb_fifo_put);

/* Init input system */
pinfo[xp_id].current_module = "init_input";
mp_input_init();
if(keyb_fifo_get > 0)
  mp_input_add_key_fd(keyb_fifo_get,1,NULL,NULL);
if(slave_mode)
   mp_input_add_cmd_fd(0,1,NULL,NULL);
else if(!use_stdin)
  mp_input_add_key_fd(0,1,NULL,NULL);
inited_flags|=INITED_INPUT;
pinfo[xp_id].current_module = NULL;

xp_id = init_signal_handling(exit_sighandler);

// ******************* Now, let's see the per-file stuff ********************

play_next_file:

// We must enable getch2 here to be able to interrupt network connection
// or cache filling
if(!use_stdin && !slave_mode){
  getch2_enable();  // prepare stdin for hotkeys...
  inited_flags|=INITED_GETCH2;
}

    if(filename) MSG_OK(MSGTR_Playing, filename);

    pinfo[xp_id].current_module="vobsub";
    if (vobsub_name){
      vo_vobsub=vobsub_open(vobsub_name,1);
      if(vo_vobsub==NULL)
        MSG_ERR(MSGTR_CantLoadSub,vobsub_name);
    }else if(sub_auto && filename && (strlen(filename)>=5)){
      /* try to autodetect vobsub from movie filename ::atmos */
      char *buf = malloc((strlen(filename)-3) * sizeof(char));
      memset(buf,0,strlen(filename)-3); // make sure string is terminated
      strncpy(buf, filename, strlen(filename)-4); 
      vo_vobsub=vobsub_open(buf,0);
      free(buf);
    }
    if(vo_vobsub)
      sub_auto=0; // don't do autosub for textsubs if vobsub found

    pinfo[xp_id].current_module="mplayer";
    stream=NULL;
    demuxer=NULL;
    d_audio=NULL;
    d_video=NULL;
    sh_audio=NULL;
    sh_video=NULL;
    
// check video_out driver name:
    if (video_driver)
	if ((i = strcspn(video_driver, ":")) > 0)
	{
	    size_t i2 = strlen(video_driver);

	    if (video_driver[i] == ':')
	    {
		vo_subdevice = malloc(i2-i);
		if (vo_subdevice != NULL)
		    strncpy(vo_subdevice, (char *)(video_driver+i+1), i2-i);
		video_driver[i] = '\0';
	    }
	}
    pinfo[xp_id].current_module="vo_register";
    vo_inited = vo_register(video_driver);

  if(!vo_inited){
    MSG_FATAL(MSGTR_InvalidVOdriver,video_driver?video_driver:"?");
    exit_player(MSGTR_Exit_error);
  }
  pinfo[xp_id].current_module="vo_init";
  if((i=vo_init(vo_subdevice))!=0)
  {
    MSG_FATAL("Error opening/initializing the selected video_out (-vo) device!\n");
    exit_player(MSGTR_Exit_error);
  }
// check audio_out driver name:
    pinfo[xp_id].current_module="ao_init";
    if (audio_driver)
	if ((i = strcspn(audio_driver, ":")) > 0)
	{
	    size_t i2 = strlen(audio_driver);

	    if (audio_driver[i] == ':')
	    {
		ao_subdevice = malloc(i2-i);
		if (ao_subdevice != NULL)
		    strncpy(ao_subdevice, (char *)(audio_driver+i+1), i2-i);
		audio_driver[i] = '\0';
	    }
	}
  if(!audio_driver)
    audio_out=audio_out_drivers[0];
  else
  for (i=0; audio_out_drivers[i] != NULL; i++){
    const ao_info_t *info = audio_out_drivers[i]->info;
    if(strcmp(info->short_name,audio_driver) == 0){
      audio_out = audio_out_drivers[i];break;
    }
  }
  if (!audio_out){
    MSG_FATAL(MSGTR_InvalidAOdriver,audio_driver);
    exit_player(MSGTR_Exit_error);
  }
  /* Initailize audio plugin interface if used */
  if(ao_plugin_cfg.plugin_list){
    pinfo[xp_id].current_module="init_audio_plugin";
    for (i=0; audio_out_drivers[i] != NULL; i++){
      const ao_info_t *info = audio_out_drivers[i]->info;
      if(strcmp(info->short_name,"plugin") == 0){
	audio_out_drivers[i]->control(AOCONTROL_SET_PLUGIN_DRIVER,(long)audio_out);
	audio_out = audio_out_drivers[i];
	break;
      }
    }
  }

//============ Open & Sync STREAM --- fork cache2 ====================

  pinfo[xp_id].current_module="open_stream";
  stream=open_stream(filename,vcd_track,&file_format);
  if(!stream) { // error...
    eof = libmpdemux_was_interrupted(PT_NEXT_ENTRY);
    goto goto_next_file;
  }
  inited_flags|=INITED_STREAM;

  if(stream->type == STREAMTYPE_PLAYLIST) {
    play_tree_t* entry;
    // Handle playlist
    pinfo[xp_id].current_module="handle_playlist";
    MSG_V("Parsing playlist %s...\n",filename);
    entry = parse_playtree(stream);
    if(!entry) {      
      entry = playtree_iter->tree;
      if(play_tree_iter_step(playtree_iter,1,0) != PLAY_TREE_ITER_ENTRY) {
	eof = PT_NEXT_ENTRY;
	goto goto_next_file;
      }
      if(playtree_iter->tree == entry ) { // Loop with a single file
	if(play_tree_iter_up_step(playtree_iter,1,0) != PLAY_TREE_ITER_ENTRY) {
	  eof = PT_NEXT_ENTRY;
	  goto goto_next_file;
	}
      }
      play_tree_remove(entry,1,1);
      eof = PT_NEXT_SRC;
      goto goto_next_file;
    }
    play_tree_insert_entry(playtree_iter->tree,entry);
    entry = playtree_iter->tree;
    if(play_tree_iter_step(playtree_iter,1,0) != PLAY_TREE_ITER_ENTRY) {
      eof = PT_NEXT_ENTRY;
      goto goto_next_file;
    }      
    play_tree_remove(entry,1,1);
    eof = PT_NEXT_SRC;
    goto goto_next_file;
  }
  stream->start_pos+=seek_to_byte;

#ifdef HAVE_LIBCSS
  pinfo[xp_id].current_module="libcss";
  if (dvdimportkey) {
    if (dvd_import_key(dvdimportkey)) {
	MSG_FATAL(MSGTR_ErrorDVDkey);
	exit_player(MSGTR_Exit_error);
    }
    MSG_INFO(MSGTR_CmdlineDVDkey);
  }
  if (dvd_auth_device) {
    if (dvd_auth(dvd_auth_device,filename)) {
	MSG_FATAL("Error in DVD auth...\n");
	exit_player(MSGTR_Exit_error);
      } 
    MSG_INFO(MSGTR_DVDauthOk);
  }
#endif

if(stream_dump_type==5){
  unsigned char buf[4096];
  int len;
  FILE *f;
  pinfo[xp_id].current_module="dumpstream";
  stream_reset(stream);
  stream_seek(stream,stream->start_pos);
  f=fopen(stream_dump_name,"wb");
  if(!f){
    MSG_FATAL(MSGTR_CantOpenDumpfile);
    exit_player(MSGTR_Exit_error);
  }
  while(!stream->eof){
      len=stream_read(stream,buf,4096);
      if(len>0) fwrite(buf,len,1,f);
  }
  fclose(f);
  MSG_INFO(MSGTR_CoreDumped);
  exit_player(MSGTR_Exit_eof);
}

#ifdef USE_DVDREAD
if(stream->type==STREAMTYPE_DVD){
  pinfo[xp_id].current_module="dvd lang->id";
  if(audio_lang && audio_id==-1) audio_id=dvd_aid_from_lang(stream,audio_lang);
  if(dvdsub_lang && dvdsub_id==-1) dvdsub_id=dvd_sid_from_lang(stream,dvdsub_lang);
  pinfo[xp_id].current_module=NULL;
}
#endif

#ifdef USE_DVDNAV
  if (dvd_nav) stream_cache_size=0;	// must disable caching...
#endif

// CACHE2: initial prefill: 20%  later: 5%  (should be set by -cacheopts)
if(stream_cache_size){
  pinfo[xp_id].current_module="enable_cache";
  if(!stream_enable_cache(stream,stream_cache_size*1024,stream_cache_size*1024/5,stream_cache_size*1024/20))
    if((eof = libmpdemux_was_interrupted(PT_NEXT_ENTRY))) goto goto_next_file;
}

//============ Open DEMUXERS --- DETECT file type =======================

if(!has_audio) audio_id=-2; // do NOT read audio packets...

pinfo[xp_id].current_module="demux_open";

demuxer=demux_open(stream,file_format,audio_id,video_id,dvdsub_id);
if(!demuxer) goto goto_next_file; // exit_player(MSGTR_Exit_error); // ERROR

//file_format=demuxer->file_format;

d_audio=demuxer->audio;
d_video=demuxer->video;
d_dvdsub=demuxer->sub;

// DUMP STREAMS:
if((stream_dump_type)&&(stream_dump_type!=4)){
  FILE *f;
  demux_stream_t *ds=NULL;
  pinfo[xp_id].current_module="dump";
  // select stream to dump
  switch(stream_dump_type){
  case 1: ds=d_audio;break;
  case 2: ds=d_video;break;
  case 3: ds=d_dvdsub;break;
  }
  if(!ds){        
      MSG_FATAL(MSGTR_DumpSelectedSteramMissing);
      exit_player(MSGTR_Exit_error);
  }
  // disable other streams:
  if(d_audio && d_audio!=ds) {ds_free_packs(d_audio); d_audio->id=-2; }
  if(d_video && d_video!=ds) {ds_free_packs(d_video); d_video->id=-2; }
  if(d_dvdsub && d_dvdsub!=ds) {ds_free_packs(d_dvdsub); d_dvdsub->id=-2; }
  // let's dump it!
  f=fopen(stream_dump_name,"wb");
  if(!f){
    MSG_FATAL(MSGTR_CantOpenDumpfile);
    exit_player(MSGTR_Exit_error);
  }
  while(!ds->eof){
    unsigned char* start;
    int in_size=ds_get_packet(ds,&start);
    if( (demuxer->file_format==DEMUXER_TYPE_AVI || demuxer->file_format==DEMUXER_TYPE_ASF || demuxer->file_format==DEMUXER_TYPE_MOV)
	&& stream_dump_type==2) fwrite(&in_size,1,4,f);
    if(in_size>0) fwrite(start,in_size,1,f);
  }
  fclose(f);
  MSG_INFO(MSGTR_CoreDumped);
  exit_player(MSGTR_Exit_eof);
}

sh_audio=d_audio->sh;
sh_video=d_video->sh;

if(sh_video){

  pinfo[xp_id].current_module="video_read_properties";
  if(!video_read_properties(sh_video)) {
    MSG_ERR("Video: can't read properties\n");
    sh_video=d_video->sh=NULL;
  } else {
    MSG_INFO("[V] filefmt:%d  fourcc:0x%X  size:%dx%d  fps:%5.2f  ftime:=%6.4f\n",
	   demuxer->file_format,sh_video->format, sh_video->disp_w,sh_video->disp_h,
	   sh_video->fps,sh_video->frametime
	   );

    vo_fps = sh_video->fps;
    /* need to set fps here for output encoders to pick it up in their init */
    if(force_fps){
      sh_video->fps=force_fps;
      sh_video->frametime=1.0f/sh_video->fps;
      vo_fps = force_fps;
    }

    if(!sh_video->fps && !force_fps){
      MSG_ERR(MSGTR_FPSnotspecified);
      sh_video=d_video->sh=NULL;
    }
  }
if(sh_video->is_static) { vo_da_buffs=1; enable_xp=0; } /* supress video mode for static pictures */
}else enable_xp=0;

fflush(stdout);

if(!sh_video && !sh_audio){
    MSG_FATAL("No stream found\n");
    goto goto_next_file; // exit_player(MSGTR_Exit_error);
}

/* display clip info */
demux_info_print(demuxer);

//================== Read SUBTITLES (DVD & TEXT) ==========================
if(sh_video){

if (spudec_ifo) {
  unsigned int palette[16], width, height;
  pinfo[xp_id].current_module="spudec_init_vobsub";
  if (vobsub_parse_ifo(spudec_ifo, palette, &width, &height, 1) >= 0)
    vo_spudec=spudec_new_scaled(palette, sh_video->disp_w, sh_video->disp_h);
}

#ifdef USE_DVDNAV
if (vo_spudec==NULL && stream->type==STREAMTYPE_DVDNAV) {
  pinfo[xp_id].current_module="spudec_init_dvdnav";
  vo_spudec=spudec_new_scaled(dvdnav_stream_get_palette((dvdnav_priv_t*)(stream->priv)),
			    sh_video->disp_w, sh_video->disp_h);
}
#endif

#ifdef USE_DVDREAD
if (vo_spudec==NULL && stream->type==STREAMTYPE_DVD) {
  pinfo[xp_id].current_module="spudec_init_dvdread";
  vo_spudec=spudec_new_scaled(((dvd_priv_t *)(stream->priv))->cur_pgc->palette,
			    sh_video->disp_w, sh_video->disp_h);
}
#endif

if (vo_spudec==NULL) {
  pinfo[xp_id].current_module="spudec_init_normal";
  vo_spudec=spudec_new_scaled(NULL, sh_video->disp_w, sh_video->disp_h);
  spudec_set_font_factor(font_factor);
}

if (vo_spudec!=NULL)
  inited_flags|=INITED_SPUDEC;

#ifdef USE_SUB
// after reading video params we should load subtitles because
// we know fps so now we can adjust subtitles time to ~6 seconds AST
// check .sub
  pinfo[xp_id].current_module="read_subtitles_file";
  if(sub_name){
    subtitles=sub_read_file(sub_name, sh_video->fps);
    if(!subtitles) MSG_ERR(MSGTR_CantLoadSub,sub_name);
  } else
  if(sub_auto) { // auto load sub file ...
    subtitles=sub_read_file( filename ? sub_filename( get_path("sub/"), filename )
	                              : "default.sub", sh_video->fps );
  }
  if(subtitles && stream_dump_type==3) list_sub_file(subtitles);
  if(subtitles && stream_dump_type==4) dump_mpsub(subtitles, sh_video->fps);
#endif	

}
//================== Init AUDIO (codec) ==========================
pinfo[xp_id].current_module="init_audio_codec";

if(sh_audio){
  // Go through the codec.conf and find the best codec...
  sh_audio->codec=NULL;
  if(audio_family) MSG_INFO(MSGTR_TryForceAudioFmt,audio_family);
  while(1){
    sh_audio->codec=find_codec(sh_audio->format,NULL,sh_audio->codec,1);
    if(!sh_audio->codec){
      if(audio_family) {
        sh_audio->codec=NULL; /* re-search */
        MSG_ERR(MSGTR_CantFindAfmtFallback);
        audio_family=NULL;
        continue;      
      }
      MSG_ERR(MSGTR_CantFindAudioCodec,sh_audio->format);
      MSG_HINT( MSGTR_TryUpgradeCodecsConfOrRTFM,get_path("codecs.conf"));
      sh_audio=d_audio->sh=NULL;
      break;
    }
    if(audio_codec && strcmp(sh_audio->codec->codec_name,audio_codec)) continue;
    else if(audio_family && strcmp(sh_audio->codec->driver_name,audio_family)) continue;
    MSG_OK("%s audio codec: [%s] drv:%s (%s)\n",audio_codec?"Forcing":"Detected",sh_audio->codec->codec_name,sh_audio->codec->driver_name,sh_audio->codec->s_info);
    break;
  }
}

if(sh_audio){
  MSG_V("Initializing audio codec...\n");
  if(!init_audio(sh_audio)){
    MSG_ERR(MSGTR_CouldntInitAudioCodec);
    sh_audio=d_audio->sh=NULL;
  } else {
    MSG_INFO("AUDIO: srate=%d  chans=%d  bps=%d  sfmt=0x%X  ratio: %d->%d\n",sh_audio->samplerate,sh_audio->channels,sh_audio->samplesize,
        sh_audio->sample_format,sh_audio->i_bps,sh_audio->o_bps);
  }
}

/*================== Init VIDEO (codec & libvo) ==========================*/
if(!sh_video)
   goto main;
pinfo[xp_id].current_module="init_video_codec";

/* Go through the codec.conf and find the best codec...*/
sh_video->inited=0;
codecs_reset_selection(0);
if(video_codec){
    /* forced codec by name: */
    MSG_INFO("Forced video codec: %s\n",video_codec);
    init_video(sh_video,video_codec,NULL,-1);
} else {
    int status;
    /* try in stability order: UNTESTED, WORKING, BUGGY, BROKEN */
    if(video_family>=0) MSG_INFO(MSGTR_TryForceVideoFmt,video_family);
    for(status=CODECS_STATUS__MAX;status>=CODECS_STATUS__MIN;--status){
	if(video_family>=0) /* try first the preferred codec family:*/
	    if(init_video(sh_video,NULL,video_family,status)) break;
	if(init_video(sh_video,NULL,NULL,status)) break;
    }
}
if(!sh_video->inited){
    MSG_ERR(MSGTR_CantFindVideoCodec,sh_video->format);
    MSG_HINT( MSGTR_TryUpgradeCodecsConfOrRTFM,get_path("codecs.conf"));
    if(!sh_audio) goto goto_next_file;
    sh_video = d_video->sh = NULL;
    goto main;
}

MSG_OK("%s video codec: [%s] vfm:%s (%s)\n",
    video_codec?"Forcing":"Detected",sh_video->codec->codec_name,sh_video->codec->driver_name,sh_video->codec->s_info);


if(auto_quality>0){
    /* Auto quality option enabled*/
    output_quality=get_video_quality_max(sh_video);
    if(auto_quality>output_quality) auto_quality=output_quality;
    else output_quality=auto_quality;
    MSG_V("AutoQ: setting quality to %d\n",output_quality);
    set_video_quality(sh_video,output_quality);
}

// ========== Init display (sh_video->disp_w*sh_video->disp_h/out_fmt) ============

    inited_flags|=INITED_VO;
    MSG_V("INFO: Video OUT driver init OK!\n");
    pinfo[xp_id].current_module="init_libvo";
    init_video_vaa();
    fflush(stdout);

//================== MAIN: ==========================
   main:
if(!sh_video) osd_level = 0;


{
int grab_frames=0;

//================ SETUP AUDIO ==========================

if(sh_audio){
  const ao_info_t *info=audio_out->info;
  pinfo[xp_id].current_module="setup_audio";
  MSG_INFO("AO: [%s] %iHz %s %s\n",
      info->short_name,
      force_srate?force_srate:sh_audio->samplerate,
      sh_audio->channels>1?"Stereo":"Mono",
      audio_out_format_name(sh_audio->sample_format)
  );
  MSG_V("AO: Description: %s\nAO: Author: %s\n",
      info->name, info->author);
  if(strlen(info->comment) > 0)
      MSG_V("AO: Comment: %s\n", info->comment);

  if(!audio_out->init(force_srate?force_srate:sh_audio->samplerate,
      sh_audio->channels,sh_audio->sample_format,0)){
    MSG_ERR(MSGTR_CannotInitAO);
    sh_audio=d_audio->sh=NULL;
    if(sh_video == NULL)
      goto goto_next_file;
  } else {
    inited_flags|=INITED_AO;
  }
}

pinfo[xp_id].current_module="av_init";

if(sh_video) sh_video->timer=0;
if(sh_audio) sh_audio->timer=-audio_delay;

if(!sh_audio){
  MSG_INFO(MSGTR_NoSound);
  if(verbose) MSG_V("Freeing %d unused audio chunks\n",d_audio->packs);
  ds_free_packs(d_audio); // free buffered chunks
  d_audio->id=-2;         // do not read audio chunks
  if(audio_out) uninit_player(INITED_AO); // close device
}
if(!sh_video){
   MSG_INFO("Video: no video!!!\n");
   if(verbose) MSG_V("Freeing %d unused video chunks\n",d_video->packs);
   ds_free_packs(d_video);
   d_video->id=-2;
   if(vo_inited) uninit_player(INITED_VO);
}

if(demuxer->file_format!=DEMUXER_TYPE_AVI) pts_from_bps=0; // it must be 0 for mpeg/asf!
if(force_fps){
  vo_fps = sh_video->fps=force_fps;
  sh_video->frametime=1.0f/sh_video->fps;
  MSG_INFO(MSGTR_FPSforced,sh_video->fps,sh_video->frametime);
}
    if(vo_get_num_frames(&xp_num_frames)!=VO_TRUE)
    {
	if(enable_xp)
	    exit_player("Selected -vo driver doesn't support DECODING AHEAD! Try other.");
	xp_num_frames=0;
    }
    if(enable_xp)
    {
      pinfo[xp_id].current_module="init_xp";
      if(xp_num_frames > 4) /* we need at least 5 buffers to suppress screen judering */
      {
	if(init_dec_ahead(sh_video,sh_audio)!=0)
	{
	    enable_xp = 0;
	    exit_player("Can't initialize decoding ahead!\n");
	}
	if(run_dec_ahead()!=0)
	{
	    enable_xp = 0;
	    exit_player("Can't run decoding ahead!\n");
	}
	MSG_OK("Using DECODING AHEAD mplayer's core with %u buffers\n",xp_num_frames);
	xp_drop_frame=frame_dropping;
	decore_video_ptr = xp_decore_video;
	/* reset counters */
	dec_ahead_active_frame=0;
	abs_dec_ahead_active_frame=0;
	loc_dec_ahead_active_frame=0;
	xp_n_frame_to_drop=0;
	xp_drop_frame_cnt=0;
      }
      else
	exit_player("Not enough buffers for DECODING AHEAD! Try other '-vo' driver.\n");
    }

fflush(stdout);
fflush(stderr);
/* let thread will decode ahead*/
    if(enable_xp)
    {
	volatile unsigned ada_blitted_frame;
	do {
	usleep(0);
	LOCK_VDEC_LOCKED();
	ada_blitted_frame = abs_dec_ahead_blitted_frame;
	UNLOCK_VDEC_LOCKED();
	}while(ada_blitted_frame < xp_num_frames/2 && !xp_eof);
    }
//==================== START PLAYING =======================

MSG_OK(MSGTR_StartPlaying);fflush(stdout);

InitTimer();

#ifdef USE_DVDNAV
if (dvd_nav && stream->type==STREAMTYPE_DVDNAV) {
  dvdnav_stream_fullstart((dvdnav_priv_t *)stream->priv);
}
#endif

total_time_usage_start=GetTimer();
audio_time_usage=0; video_time_usage=0; vout_time_usage=0;
if(benchmark) 
{
  max_audio_time_usage=0; max_video_time_usage=0; max_vout_time_usage=0;
  min_audio_time_usage=HUGE_VAL; min_video_time_usage=HUGE_VAL; min_vout_time_usage=HUGE_VAL;
}
mp_msg_flush();
while(!eof){
    float v_pts=0;
//    unsigned int aq_total_time=GetTimer();
    float aq_sleep_time=0;

    if(play_n_frames>=0){
      --play_n_frames;
      if(play_n_frames<0) eof = PT_NEXT_ENTRY;
    }

    eof |= decore_audio();
/*========================== UPDATE TIMERS ============================*/
if(!sh_video) {
  if(benchmark && verbose) MSG_STATUS("A:%6.1f %4.1f%% %d%%   \r"
				,sh_audio->timer-audio_out->get_delay()
				,(sh_audio->timer>0.5)?100.0*audio_time_usage/(double)sh_audio->timer:0
				,cache_fill_status
				);
  goto read_input;
}else{
  int l_eof;

/*========================== PLAY VIDEO ============================*/

vo_pts=sh_video->timer*90000.0;
vo_fps=sh_video->fps;

    if(sh_video->is_static && our_n_frames)
    {
     /* don't decode if it's picture */
     usleep(0);
     if(vo_check_events()) goto repaint;
    }
    else
    {
	repaint:
	l_eof = (*decore_video_ptr)(rtc_fd,&vstat,&aq_sleep_time,&v_pts);
	eof |= l_eof;
	if(eof) goto do_loop;
    }
/*Output quality adjustments:*/
if(auto_quality>0){
  if(output_quality<auto_quality && aq_sleep_time>0)
      ++output_quality;
  else
  if(output_quality>1 && aq_sleep_time<0)
      --output_quality;
  else
  if(output_quality>0 && aq_sleep_time<-0.050f) // 50ms
      output_quality=0;
  set_video_quality(sh_video,output_quality);
}

read_input:

#ifdef USE_OSD
  if(osd_visible){
    if (!--osd_visible){
       vo_osd_progbar_type=-1; // disable
       vo_osd_changed(OSDTYPE_PROGBAR);
       if (osd_function != OSD_PAUSE)
	   osd_function = OSD_PLAY;
    }
  }
#endif

  if(osd_function==OSD_PAUSE){
    mp_cmd_t* cmd;
      if(verbose) {
	MSG_STATUS("\n------ PAUSED -------\r");
	fflush(stdout);
      }
      if (vo_inited && sh_video)
	 vo_pause();

      if (audio_out && sh_audio)
         audio_out->pause();	// pause audio, keep data if possible

      while( (cmd = mp_input_get_cmd(20,1)) == NULL) {
	     if(sh_video && vo_inited) vo_check_events();
             usleep(20000);
      }
      mp_cmd_free(cmd);
      osd_function=OSD_PLAY;
      if (audio_out && sh_audio)
        audio_out->resume();	// resume audio
      if (vo_inited && sh_video)
        vo_resume();	// resume video
      (void)GetRelativeTime();	// keep TF around FT in next cycle
  }
} /* else if(!sh_video) */
our_n_frames++;

/* skip some seconds... added by fly */

if(step_sec>0) {
	osd_function=OSD_FFW;
	rel_seek_secs+=step_sec;
}

#ifdef USE_DVDNAV
if (stream->type==STREAMTYPE_DVDNAV && dvd_nav_still)
    dvdnav_stream_sleeping((dvdnav_priv_t*)stream->priv);
#endif

//================= Keyboard events, SEEKing ====================

{
  mp_cmd_t* cmd;
  while( (cmd = mp_input_get_cmd(0,0)) != NULL) {
    switch(cmd->id) {
    case MP_CMD_SEEK : {
      int v,abs;
      v = cmd->args[0].v.i;
      abs = (cmd->nargs > 1) ? cmd->args[1].v.i : 0;
      if(abs) {
	abs_seek_pos = 3;
	if(sh_video) osd_function= (v > sh_video->timer) ? OSD_FFW : OSD_REW;
	rel_seek_secs = v/100.0;
      }
      else {
	rel_seek_secs+= v;
	osd_function= (v > 0) ? OSD_FFW : OSD_REW;
      }
    } break;
    case MP_CMD_AUDIO_DELAY : {
      float v = cmd->args[0].v.f;
      audio_delay += v;
      osd_show_av_delay = 9;
      if(sh_audio) sh_audio->timer+= v;
    } break;
    case MP_CMD_PAUSE : {
      osd_function=OSD_PAUSE;
    } break;
    case MP_CMD_QUIT : {
      exit_player(MSGTR_Exit_quit);
    }
    case MP_CMD_GRAB_FRAMES : {
      grab_frames=2;
    } break;
    case MP_CMD_PLAY_TREE_STEP : {
      int n = cmd->args[0].v.i > 0 ? 1 : -1;
      play_tree_iter_t* i = play_tree_iter_new_copy(playtree_iter);
      
      if(play_tree_iter_step(i,n,0) == PLAY_TREE_ITER_ENTRY)
	eof = (n > 0) ? PT_NEXT_ENTRY : PT_PREV_ENTRY;
      play_tree_iter_free(i);
    } break;
    case MP_CMD_PLAY_TREE_UP_STEP : {
      int n = cmd->args[0].v.i > 0 ? 1 : -1;
      play_tree_iter_t* i = play_tree_iter_new_copy(playtree_iter);
      if(play_tree_iter_up_step(i,n,0) == PLAY_TREE_ITER_ENTRY)
	eof = (n > 0) ? PT_UP_NEXT : PT_UP_PREV;
      play_tree_iter_free(i);
    } break;
    case MP_CMD_PLAY_ALT_SRC_STEP : {
      if(playtree_iter->num_files > 1) {
	int v = cmd->args[0].v.i;
	if(v > 0 && playtree_iter->file < playtree_iter->num_files)
	  eof = PT_NEXT_SRC;
	else if(v < 0 && playtree_iter->file > 1)
	  eof = PT_PREV_SRC;
      }
    } break;
    case MP_CMD_SUB_DELAY : {
      int abs= cmd->args[1].v.i;
      float v = cmd->args[0].v.f;
      if(abs)
	sub_delay = v;
      else
	sub_delay += v;
      osd_show_sub_delay = 9; // show the subdelay in OSD
    } break;
    case MP_CMD_OSD : 
      if(sh_video) {
	int v = cmd->args[0].v.i;
	if(v < 0)
	  osd_level=(osd_level+1)%3;
	else
	  osd_level= v > 2 ? 2 : v;
      } break;
    case MP_CMD_VOLUME :  {
      int v = cmd->args[0].v.i;
      if(v > 0)
	mixer_incvolume();
      else
	mixer_decvolume();
#ifdef USE_OSD
      if(osd_level){
	osd_visible=sh_video->fps; // 1 sec
	vo_osd_progbar_type=OSD_VOLUME;
	vo_osd_progbar_value=(mixer_getbothvolume()*256.0)/100.0;
	vo_osd_changed(OSDTYPE_PROGBAR);
      }
#endif
    } break;
    case MP_CMD_MIXER_USEMASTER :  {
    } break;
    case MP_CMD_CONTRAST :  {
      int v = cmd->args[0].v.i, abs = cmd->args[1].v.i;
      if(abs)
	v_cont = v > 100 ? 100 : v;
      else {
	if ( (v_cont += v) > 100 ) v_cont = 100;
      }
      if(v_hw_equ_cap & VEQ_CAP_BRIGHTNESS) {
	if(v_cont < -100) v_cont = -100;
      } else {
	if ( v_cont < 0 ) v_cont = 0;	    
      }
      if(set_video_colors(sh_video,"Contrast",v_cont)){
#ifdef USE_OSD
	if(osd_level){
	  osd_visible=sh_video->fps; // 1 sec
	  vo_osd_progbar_type=OSD_CONTRAST;
	  vo_osd_progbar_value=((v_cont)<<8)/100;
	  if(v_hw_equ_cap) vo_osd_progbar_value = ((v_cont+100)<<8)/200;
	  vo_osd_changed(OSDTYPE_PROGBAR);
	}
#endif
      }
    } break;
    case MP_CMD_BRIGHTNESS :  {
      int v = cmd->args[0].v.i, abs = cmd->args[1].v.i;
      if(abs)
	v_bright = v > 100 ? 100 : v;
      else {
	if ( (v_bright += v) > 100 ) v_bright = 100;
      }
      if(v_hw_equ_cap & VEQ_CAP_BRIGHTNESS) {
	if(v_bright < -100) v_bright = -100;
      } else {
	if ( v_bright < 0 ) v_bright = 0;	    
      }
      if(set_video_colors(sh_video,"Brightness",v_bright)){
#ifdef USE_OSD
	if(osd_level){
	  osd_visible=sh_video->fps; // 1 sec
	  vo_osd_progbar_type=OSD_BRIGHTNESS;
	  vo_osd_progbar_value=((v_bright)<<8)/100;
	  if(v_hw_equ_cap) vo_osd_progbar_value = ((v_bright+100)<<8)/200;
	  vo_osd_changed(OSDTYPE_PROGBAR);
	}
#endif
      }      
    } break;
    case MP_CMD_HUE :  {
      int v = cmd->args[0].v.i, abs = cmd->args[1].v.i;
      if(abs)
	v_hue = v > 100 ? 100 : v;
      else {
	if ( (v_hue += v) > 100 ) v_hue = 100;
      }
      if(v_hw_equ_cap & VEQ_CAP_HUE) {
	if(v_hue < -100) v_hue = -100;
      } else {
	if ( v_hue < 0 ) v_hue = 0;	    
      }
      if(set_video_colors(sh_video,"Hue",v_hue)){
#ifdef USE_OSD
	if(osd_level){
	  osd_visible=sh_video->fps; // 1 sec
	  vo_osd_progbar_type=OSD_HUE;
	  vo_osd_progbar_value=((v_hue)<<8)/100;
	  if(v_hw_equ_cap) vo_osd_progbar_value = ((v_hue+100)<<8)/200;
	  vo_osd_changed(OSDTYPE_PROGBAR);
	}
#endif
      }	
    } break;
    case MP_CMD_SATURATION :  {
      int v = cmd->args[0].v.i, abs = cmd->args[1].v.i;
      if(abs)
	v_saturation = v > 100 ? 100 : v;
      else {
	if ( (v_saturation += v) > 100 ) v_saturation = 100;
      }
      if(v_hw_equ_cap & VEQ_CAP_SATURATION) {
	if(v_saturation < -100) v_saturation = -100;
      } else {
	if ( v_saturation < 0 ) v_saturation = 0;	    
      }
      if(set_video_colors(sh_video,"Saturation",v_saturation)){
#ifdef USE_OSD
	if(osd_level){
	  osd_visible=sh_video->fps; // 1 sec
	  vo_osd_progbar_type=OSD_SATURATION;
	  vo_osd_progbar_value=((v_saturation)<<8)/100;
	  if(v_hw_equ_cap) vo_osd_progbar_value = ((v_saturation+100)<<8)/200;
	  vo_osd_changed(OSDTYPE_PROGBAR);
	}
#endif
      }
    } break;
    case MP_CMD_FRAMEDROPPING :  {
      int v = cmd->args[0].v.i;
      if(v < 0)
	frame_dropping = (frame_dropping+1)%3;
      else
	frame_dropping = v > 2 ? 2 : v;
    } break;
#ifdef USE_TV
    case MP_CMD_TV_STEP_CHANNEL :  {
      if (tv_param_on == 1) {
	int v = cmd->args[0].v.i;
	if(v > 0)
	  tv_step_channel(tv_handler, TV_CHANNEL_HIGHER);
	else
	  tv_step_channel(tv_handler, TV_CHANNEL_LOWER);
      }
    } break;
    case MP_CMD_TV_STEP_NORM :  {
      if (tv_param_on == 1)
	tv_step_norm(tv_handler);
    } break;
    case MP_CMD_TV_STEP_CHANNEL_LIST :  {
      if (tv_param_on == 1)
	tv_step_chanlist(tv_handler);
    } break;
#endif
    case MP_CMD_VO_FULLSCREEN:
	vo_fullscreen();
	break;
    case MP_CMD_VO_SCREENSHOT:
	vo_screenshot();
	break;
    case MP_CMD_SUB_POS:
    {
        int v;
	v = cmd->args[0].v.i;
    
	sub_pos+=v;
	if(sub_pos >100) sub_pos=100;
	if(sub_pos <0) sub_pos=0;
	vo_osd_changed(OSDTYPE_SUBTITLE);
    }	break;
#ifdef USE_DVDNAV
    case MP_CMD_DVDNAV_EVENT: {
      dvdnav_priv_t * dvdnav_priv = (dvdnav_priv_t*)(stream->priv);
      dvdnav_event_t * dvdnav_event = (dvdnav_event_t *)(cmd->args[0].v.v);

      if (!dvdnav_event) {
        MSG_DBG2("DVDNAV Event NULL?!\n");
        break;
      }

      if (stream->type!=STREAMTYPE_DVDNAV) {
        MSG_DBG2("Got DVDNAV event when not running a DVDNAV stream!?\n");
        break;
      }

      switch (dvdnav_event->event) {
      case DVDNAV_BLOCK_OK: {
          /* be silent about this one */
                break;
          }
      case DVDNAV_HIGHLIGHT: {
          dvdnav_highlight_event_t *hevent = (dvdnav_highlight_event_t*)(dvdnav_event->details);
          if (!hevent) {
                MSG_ERR("DVDNAV Event: Highlight event broken\n");
                break;
          }

          if (hevent->display && hevent->buttonN>0)
          {
                //dvdnav_priv->seen_root_menu=1; /* if we got a highlight, we're on a menu */
                sprintf( dvd_nav_text, "Highlight button %d (%u,%u)-(%u,%u) PTS %d (now is %5.2f)",
                     hevent->buttonN,
                     hevent->sx,hevent->sy,
                     hevent->ex,hevent->ey,
                     hevent->pts, d_video->pts);
                MSG_DBG2("DVDNAV Event: %s\n",dvd_nav_text);
                //osd_show_dvd_nav_delay = 60;

                osd_show_dvd_nav_highlight=1;
                osd_show_dvd_nav_sx=hevent->sx;
                osd_show_dvd_nav_ex=hevent->ex;
                osd_show_dvd_nav_sy=hevent->sy;
                osd_show_dvd_nav_ey=hevent->ey;
          }
          else {
                  osd_show_dvd_nav_highlight=0;
                  MSG_DBG2("DVDNAV Event: Highlight Hide\n");
          }
        break;
        }
      case DVDNAV_STILL_FRAME: {
          dvdnav_still_event_t *still_event = (dvdnav_still_event_t*)(dvdnav_event->details);

          MSG_DBG2( "######## DVDNAV Event: Still Frame: %d sec(s)\n", still_event->length );
          while (dvdnav_stream_sleeping(dvdnav_priv)) {
            usleep(1000); /* 1ms */
          }
          dvdnav_stream_sleep(dvdnav_priv,still_event->length);
        break;
        }
      case DVDNAV_STOP: {
          MSG_DBG2( "DVDNAV Event: Nav Stop\n" );
        break;
        }
      case DVDNAV_NOP: {
        MSG_DBG2("DVDNAV Event: Nav NOP\n");
        break;
        }
      case DVDNAV_SPU_STREAM_CHANGE: {
        dvdnav_stream_change_event_t * stream_change=(dvdnav_stream_change_event_t*)(dvdnav_event->details);

	MSG_DBG2("DVDNAV Event: Nav SPU Stream Change: phys: %d logical: %d\n",
                stream_change->physical,
                stream_change->logical);

        if (vo_spudec && dvdsub_id!=stream_change->physical) {
                MSG_DBG2("d_dvdsub->id change: was %d is now %d\n",
                        d_dvdsub->id,stream_change->physical);
                // FIXME: need a better way to change SPU id
                d_dvdsub->id=dvdsub_id=stream_change->physical;
                if (vo_spudec) spudec_reset(vo_spudec);
        }

        break;
        }
      case DVDNAV_AUDIO_STREAM_CHANGE: {
        int aid_temp;
        dvdnav_stream_change_event_t *stream_change = (dvdnav_stream_change_event_t*)(dvdnav_event->details);

        MSG_DBG2("DVDNAV Event: Nav Audio Stream Change: phys: %d logical: %d\n",
                stream_change->physical,
                stream_change->logical);

        aid_temp=stream_change->physical;
        if (aid_temp>=0) aid_temp+=128; // FIXME: is this sane?
        if (d_audio && audio_id!=aid_temp) {
                MSG_DBG2("d_audio->id change: was %d is now %d\n",
                        d_audio->id,aid_temp);
                // FIXME: need a bettery way to change audio stream id
                d_audio->id=dvdsub_id=aid_temp;
                resync_audio_stream(sh_audio);
        }

        break;
      }
      case DVDNAV_VTS_CHANGE: {
        MSG_DBG2("DVDNAV Event: Nav VTS Change\n");
        break;
        }
      case DVDNAV_CELL_CHANGE: {
        /*dvdnav_cell_change_event_t *cell_change = (dvdnav_cell_change_event_t*)(dvdnav_event->details);*/
        /*cell_playback_t * cell_playback = cell_change->new_cell;*/

        MSG_DBG2("DVDNAV Event: Nav Cell Change\n");
        osd_show_dvd_nav_highlight=0; /* screen changed, disable menu */
        break;
        }
      case DVDNAV_NAV_PACKET: {
        MSG_DBG2("DVDNAV Event: Nav Packet\n");
        break;
        }
      case DVDNAV_SPU_CLUT_CHANGE: {
        uint32_t * new_clut = (uint32_t *)(dvdnav_event->details);

        MSG_DBG2("DVDNAV Event: Nav SPU CLUT Change\n");
        // send new palette to SPU decoder
        if (vo_spudec) spudec_update_palette(vo_spudec,new_clut);

        break;
        }
      case DVDNAV_SEEK_DONE: {
        MSG_DBG2("DVDNAV Event: Nav Seek Done\n");
        break;
        }
      }

      // free the dvdnav event
      free(dvdnav_event->details);
      free(dvdnav_event);
      cmd->args[0].v.v=NULL;
    }
    case MP_CMD_DVDNAV: {
      dvdnav_priv_t * dvdnav_priv=(dvdnav_priv_t*)stream->priv;

      switch (cmd->args[0].v.i) {
        case MP_CMD_DVDNAV_UP:
          dvdnav_upper_button_select(dvdnav_priv->dvdnav);
          break;
        case MP_CMD_DVDNAV_DOWN:
          dvdnav_lower_button_select(dvdnav_priv->dvdnav);
          break;
        case MP_CMD_DVDNAV_LEFT:
          dvdnav_left_button_select(dvdnav_priv->dvdnav);
          break;
        case MP_CMD_DVDNAV_RIGHT:
          dvdnav_right_button_select(dvdnav_priv->dvdnav);
          break;
        case MP_CMD_DVDNAV_MENU:
          printf("Menu call\n");
          dvdnav_menu_call(dvdnav_priv->dvdnav,DVD_MENU_Root);
          break;
        case MP_CMD_DVDNAV_SELECT:
          dvdnav_button_activate(dvdnav_priv->dvdnav);
          break;
        default:
          MSG_V( "Weird DVD Nav cmd %d\n",cmd->args[0].v.i);
          break;
      }
      break;
    }
#endif /* DVD NAV */
    default : {
      MSG_ERR("Received unknow cmd %s\n",cmd->name);
    }
    }
    mp_cmd_free(cmd);
  }
}

  if (seek_to_sec) {
    int a,b; float d;
    
    if (sscanf(seek_to_sec, "%d:%d:%f", &a,&b,&d)==3)
	rel_seek_secs += 3600*a +60*b +d ;
    else if (sscanf(seek_to_sec, "%d:%f", &a, &d)==2)
	rel_seek_secs += 60*a +d;
    else if (sscanf(seek_to_sec, "%f", &d)==1)
	rel_seek_secs += d;

     seek_to_sec = NULL;
  }
do_loop:
  /* Looping. */
  if(eof && loop_times>=0) {

    MSG_V("loop_times = %d, eof = %d\n", loop_times,eof);

    if(loop_times>1) loop_times--; else
    if(loop_times==1) loop_times=-1;

    eof=0;
    abs_seek_pos=3; rel_seek_secs=0; // seek to start of movie (0%)

  }

if(rel_seek_secs || abs_seek_pos){
  pinfo[xp_id].current_module="seek";
  if(enable_xp)
  {
      // *(char*)0x100 = 1; // Testing crasch
    if(!xp_eof) {
        struct timeval now;
        struct timespec timeout;
        int retcode=0;

        if(dec_ahead_in_lseek!=NoSeek) {
            // Wait for thread to finish last seek
            LOCK_SEEK_COND();
            gettimeofday(&now,NULL);
            timeout.tv_sec = now.tv_sec + 1;
            timeout.tv_nsec = now.tv_usec * 1000;
            while(dec_ahead_in_lseek!=NoSeek && !xp_eof && retcode != ETIMEDOUT)
                retcode = pthread_cond_timedwait(&seek_cond,&seek_cond_mutex,&timeout);
            UNLOCK_SEEK_COND();
            if(retcode == ETIMEDOUT) MSG_V("Start seek timeout\n");
            retcode=0;
        }

        // Wait for thread to prepare seek
        LOCK_SEEK();
        dec_ahead_in_lseek=PreSeek;
        LOCK_SEEK_COND();
        gettimeofday(&now,NULL);
        timeout.tv_sec = now.tv_sec + 1;
        timeout.tv_nsec = now.tv_usec * 1000;
        while(dec_ahead_in_lseek==PreSeek && !xp_eof && retcode != ETIMEDOUT)
            retcode = pthread_cond_timedwait(&seek_cond,&seek_cond_mutex,&timeout);
        UNLOCK_SEEK_COND();
        if(retcode == ETIMEDOUT) MSG_V("Pre seek timeout\n");
    } else
        dec_ahead_in_lseek=Seek;
    LOCK_VREADING();
    UNLOCK_SEEK();
  }
  mpxp_seek(xp_id,&vstat,&osd_visible,v_pts);
  if(enable_xp) 
  {
    /* Ugly hack: but we should read audio packet before video after seeking.
       Else we'll get picture destortion on the screen */
    eof |= decore_audio();
    UNLOCK_VREADING();
    if(!xp_eof) {
        struct timeval now;
        struct timespec timeout;
        int retcode=0;

        // Wait for thread to finish seek
        LOCK_SEEK_COND();
        gettimeofday(&now,NULL);
        timeout.tv_sec = now.tv_sec + 1;
        timeout.tv_nsec = now.tv_usec * 1000;
        while(dec_ahead_in_lseek!=NoSeek && !xp_eof && retcode != ETIMEDOUT)
            retcode = pthread_cond_timedwait(&seek_cond,&seek_cond_mutex,&timeout);
        UNLOCK_SEEK_COND();
        pthread_mutex_unlock(&seek_cond_mutex);
        if(retcode == ETIMEDOUT) MSG_V("End seek timeout\n");
    } else
        dec_ahead_in_lseek=NoSeek;
  }
  rel_seek_secs=0;
  abs_seek_pos=0;
  pinfo[xp_id].current_module=NULL;
}
#ifdef USE_OSD
    if(!enable_xp) update_osd(d_video->pts);
#endif
//================= Update OSD ====================
  
#ifdef USE_SUB
  // find sub
  if(subtitles && v_pts>0){
      float pts=v_pts;
      if(sub_fps==0) sub_fps=sh_video->fps;
      pinfo[xp_id].current_module="find_sub";
      if (pts > sub_last_pts || pts < sub_last_pts-1.0 ) {
         find_sub(subtitles,sub_uses_time?(100*(pts+sub_delay)):((pts+sub_delay)*sub_fps)); // FIXME! frame counter...
         sub_last_pts = pts;
      }
      pinfo[xp_id].current_module=NULL;
  }
#endif
  

  // VobSub subtitles
  if(vo_vobsub){
    pinfo[xp_id].current_module="vobsub";
    if(d_video->pts+sub_delay>=0)
      vobsub_process(vo_vobsub,d_video->pts+sub_delay);
    
    /* Don't know how to detect wether the sub has changed or not */
    vo_osd_changed(OSDTYPE_VOBSUB);
    pinfo[xp_id].current_module=NULL;
  }

  // DVD sub:
  if(vo_flags & 0x08){
    static vo_mpegpes_t packet;
    static vo_mpegpes_t *pkg=&packet;
    packet.timestamp=sh_video->timer*90000.0;
    packet.id=0x20; /* Subpic */
    while((packet.size=ds_get_packet_sub(d_dvdsub,&packet.data))>0){
      MSG_V("\rDVD sub: len=%d  v_pts=%5.3f  s_pts=%5.3f  \n",packet.size,v_pts,d_dvdsub->pts);
      vo_draw_frame(&pkg);
    }
  }else if(vo_spudec){
    unsigned char* packet=NULL;
    int len;
    pinfo[xp_id].current_module="spudec";
    while((len=ds_get_packet_sub(d_dvdsub,&packet))>0){
      MSG_V("\rDVD sub: len=%d  v_pts=%5.3f  s_pts=%5.3f  \n",len,v_pts,d_dvdsub->pts);
      spudec_assemble(vo_spudec,packet,len,90000*d_dvdsub->pts);
    }
    spudec_heartbeat(vo_spudec,90000*v_pts);

    /* Don't know how to detect wether the sub has changed or not */
    vo_osd_changed(OSDTYPE_SPU);
    pinfo[xp_id].current_module=NULL;
  }
  
} // while(!eof)

MSG_V("EOF code: %d\n",eof);

}

goto_next_file:  // don't jump here after ao/vo/getch initialization!

if(benchmark){
  double tot=video_time_usage+vout_time_usage+audio_time_usage;
  double total_time_usage;
  if(!enable_xp)
  {
	max_video_time_usage *= our_n_frames;
	max_vout_time_usage *= our_n_frames;
	max_audio_time_usage *= our_n_frames;
	min_video_time_usage *= our_n_frames;
	min_vout_time_usage *= our_n_frames;
	min_audio_time_usage *= our_n_frames;
  }
  total_time_usage_start=GetTimer()-total_time_usage_start;
  total_time_usage = (float)total_time_usage_start*0.000001;
  if(enable_xp)
  {
    MSG_INFO("\nMIN BENCHMARKs: *** MEANINGLESS IN XP MODE *** \n");
    MSG_INFO("MIN BENCHMARK%%: *** MEANINGLESS IN XP MODE *** \n");
  }
  else
  {
  MSG_INFO("\nMIN BENCHMARKs: VC:%8.3fs VO:%8.3fs A:%8.3fs\n",
	 min_video_time_usage,min_vout_time_usage,
	 min_audio_time_usage);
  if(total_time_usage>0.0)
    MSG_INFO("MIN BENCHMARK%%: VC:%8.4f%% VO:%8.4f%% A:%8.4f %% = %8.4f%%\n",
	   100.0*min_video_time_usage/total_time_usage,
	   100.0*min_vout_time_usage/total_time_usage,
	   100.0*min_audio_time_usage/total_time_usage,
	   100.0*min_video_time_usage/total_time_usage+
	   100.0*min_vout_time_usage/total_time_usage+
	   100.0*min_audio_time_usage/total_time_usage
	   );
  }
  MSG_INFO("\nAVE BENCHMARKs: VC:%8.3fs VO:%8.3fs A:%8.3fs Sys:%8.3fs = %8.3fs\n",
	 video_time_usage,vout_time_usage,audio_time_usage,
	 total_time_usage-tot,total_time_usage);
  if(total_time_usage>0.0)
    MSG_INFO("AVE BENCHMARK%%: VC:%8.4f%% VO:%8.4f%% A:%8.4f%% Sys:%8.4f %% = %8.4f%%\n",
	   100.0*video_time_usage/total_time_usage,
	   100.0*vout_time_usage/total_time_usage,
	   100.0*audio_time_usage/total_time_usage,
	   100.0*(total_time_usage-tot)/total_time_usage,
	   100.0);
  if(enable_xp)
  {
    MSG_INFO("\nMAX BENCHMARKs: *** MEANINGLESS IN XP MODE *** \n");
    MSG_INFO("MAX BENCHMARK%%: *** MEANINGLESS IN XP MODE *** \n");
  }
  else
  {
  MSG_INFO("\nMAX BENCHMARKs: VC:%8.3fs VO:%8.3fs A:%8.3fs\n",
	 max_video_time_usage,max_vout_time_usage,
	 max_audio_time_usage);
  if(total_time_usage>0.0)
    MSG_INFO("MAX BENCHMARK%%: VC:%8.4f%% VO:%8.4f%% A:%8.4f %% = %8.4f%%\n",
	   100.0*max_video_time_usage/total_time_usage,
	   100.0*max_vout_time_usage/total_time_usage,
	   100.0*max_audio_time_usage/total_time_usage,
	   100.0*max_video_time_usage/total_time_usage+
	   100.0*max_vout_time_usage/total_time_usage+
	   100.0*max_audio_time_usage/total_time_usage
	   );
/* This code computes number of frame which should be dropped
   in ideal case (without SYSTIME); i.e. when file is located
   in RAM and kernel+other_tasks eat 0% of CPU. */
    MSG_INFO("TOTAL BENCHMARK: from %u frames should be dropped: %u (at least)\n"
	    ,our_n_frames,bench_dropped_frames);
  }
  if(enable_xp)
  MSG_INFO("\nREAL RESULTS: from %u was dropped=%u\n"
	,our_n_frames,xp_drop_frame_cnt);
  else
  MSG_INFO("\nREAL RESULTS: dropped=%u too slow=%u too fast=%u\n"
	,vstat.drop_frame_cnt,vstat.too_slow_frame_cnt,vstat.too_fast_frame_cnt);
}


if(eof == PT_NEXT_ENTRY || eof == PT_PREV_ENTRY) {
  eof = eof == PT_NEXT_ENTRY ? 1 : -1;
  if(play_tree_iter_step(playtree_iter,eof,0) == PLAY_TREE_ITER_ENTRY) {
    uninit_player(INITED_ALL-(INITED_LIRC+INITED_INPUT+INITED_VO));
    eof = 1;
  } else {
    play_tree_iter_free(playtree_iter);
    playtree_iter = NULL;
  }
} else if (eof == PT_UP_NEXT || eof == PT_UP_PREV) {
  eof = eof == PT_UP_NEXT ? 1 : -1;
  if(play_tree_iter_up_step(playtree_iter,eof,0) == PLAY_TREE_ITER_ENTRY) {
    uninit_player(INITED_ALL-(INITED_LIRC+INITED_INPUT+INITED_VO));
    eof = 1;
  } else {
    play_tree_iter_free(playtree_iter);
    playtree_iter = NULL;
  }
}else { // NEXT PREV SRC
     uninit_player(INITED_ALL-(INITED_LIRC+INITED_INPUT+INITED_VO));
     eof = eof == PT_PREV_SRC ? -1 : 1;
}

uninit_player(INITED_VO);

if(eof == 0) eof = 1;

while(playtree_iter != NULL) {
  filename = play_tree_iter_get_file(playtree_iter,eof);
  if(filename == NULL) {
    if( play_tree_iter_step(playtree_iter,eof,0) != PLAY_TREE_ITER_ENTRY) {
      play_tree_iter_free(playtree_iter);
      playtree_iter = NULL;
    };
  } else
    break;
} 

if( playtree_iter != NULL ){

  pinfo[xp_id].current_module="uninit_vcodec";
  if(sh_video) uninit_video(sh_video);
  sh_video=NULL;

  pinfo[xp_id].current_module="uninit_acodec";
  if(sh_video) uninit_audio(sh_audio);
  sh_audio=NULL;
  
  pinfo[xp_id].current_module="free_demuxer";
  if(demuxer) free_demuxer(demuxer);
  demuxer=NULL;

  pinfo[xp_id].current_module="free_stream";
  if(stream) free_stream(stream);
  stream=NULL;
  
  pinfo[xp_id].current_module="sub_free";
  if ( subtitles ) 
   {
    sub_free( subtitles );
    sub_name=NULL;
    vo_sub=NULL;
    subtitles=NULL;
   }

  vo_inited=0;
  audio_out=NULL;
  eof = 0;
  goto play_next_file;
}

exit_player(MSGTR_Exit_eof);

return 1;
}
