/*
  MplayerXP's Signal handling
*/

#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <sys/resource.h>
#include "sig_hand.h"
#include "mp_msg.h"
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include "dec_ahead.h"

pth_info_t pinfo[MAX_XPTHREADS];
static int xp_threads = 0;

extern pid_t mplayer_pid;
extern pid_t dec_ahead_pid;


static void my_callback(int signo)
{
    int i;
    for(i=0; i < MAX_XPTHREADS && pinfo[i].pid != getpid(); i++);
    if(i >= MAX_XPTHREADS || i >= xp_threads || pinfo[i].pid != getpid()) {
        i = 0; /* Use 0 as default handler */
    }

   mp_msg(MSGT_CPLAYER,MSGL_FATAL,"catching signal: %s in thread: %s (%i) in module: %s\n",strsignal(signo),pinfo[i].thread_name,i,pinfo[i].current_module);


   pinfo[i].sig_handler();

   signal(signo,SIG_DFL);
   /* try coredump*/
   return;
}

int init_signal_handling( void (*callback)( void ))
{
  if(xp_threads >= MAX_XPTHREADS)
    return MAX_XPTHREADS-1;
  pinfo[xp_threads].sig_handler = callback;
  pinfo[xp_threads].pid = getpid();
  pinfo[xp_threads].current_module = NULL;
  xp_threads++;

  /*========= Catch terminate signals: ================*/
  /* terminate requests:*/
  signal(SIGTERM,my_callback); /* kill*/
  signal(SIGHUP,my_callback);  /* kill -HUP  /  xterm closed*/

  signal(SIGINT,my_callback);  /* Interrupt from keyboard */

  signal(SIGQUIT,my_callback); /* Quit from keyboard */
  /* fatal errors: */
  signal(SIGBUS,my_callback);  /* bus error */
  signal(SIGSEGV,my_callback); /* segfault */
  signal(SIGILL,my_callback);  /* illegal instruction */
  signal(SIGFPE,my_callback);  /* floating point exc. */
  signal(SIGABRT,my_callback); /* abort() */   
#ifdef RLIMIT_CORE
  {
    /* on many systems default coresize is 0.
       Enable any coresize here. */
    struct rlimit rl;
    getrlimit(RLIMIT_CORE,&rl);
    rl.rlim_cur = rl.rlim_max;
    setrlimit(RLIMIT_CORE,&rl);
  }
#endif
  return xp_threads-1;
}
